# -*- test-case-name: quotient.test.test_filepile -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

"""Consecutive filesystem mashing.  See FilePile.__doc__ for details.
"""

import os
import socket
from time import time as _uniquefloat

def unique():
    return str(long(_uniquefloat() * 1000))

try:
    from os import symlink
    from os import readlink
    from os import remove as rmlink
    from os import rename as mvlink
except:
    import shutil
    def symlink(value, filename):
        newlinkname = filename+"."+unique()+'.newlink'
        newvalname = os.path.join(newlinkname,"symlink")
        os.mkdir(newlinkname)
        f = open(newvalname,'wb')
        f.write(value)
        f.flush()
        f.close()
        try:
            os.rename(newlinkname, filename)
        except:
            os.remove(newvalname)
            os.rmdir(newlinkname)
            raise

    def readlink(filename):
        return open(os.path.join(filename,'symlink'),'rb').read()

    def rmlink(filename):
        shutil.rmtree(filename)

    def mvlink(src, dest):
        try:
            shutil.rmtree(dest)
        except:
            pass
        os.rename(src,dest)


from os import makedirs

prochost = socket.gethostname()
procpid = os.getpid()

def _bisect_cmp(a, x, cmpfunc=cmp, lo=0, hi=None):
    """A left-bisect on a vector using a user-provided comparison function.
    """
    if hi is None:
        hi = len(a)
    while lo < hi:
        mid = (lo+hi) // 2
        if cmpfunc(a[mid], x) == -1:
            lo = mid+1
        else:
            hi = mid
    return lo

def nextFileName(dirname, ext, directory=False):
    """Get a sequential file-name from a directory in a NFS and SMP-safe way.

    When you use this filename, do not run file() or symlink() on it; you
    must create a file elsewhere (you can use a mutated version of the returned
    filename to achieve SMP/NFS safety) and use os.rename over the returned
    filename.

    The returned filename includes the passed-in dirname; if you want an
    absolute path, pass an absolute dirname - don't manipulate the returned
    path.

    This will allocate a file-name using the following strategy:

        - look for a symlink named ext.sequence
        - if it's found
          - read the body of this symlink and jump to that sequence number
        - otherwise
          - start from 0
        - keep incrementing the sequence number until you arrive at a
        [sequence].ext which does not exist. (this is verified by creating a
        symlink, which is should always be an atomic test-and-set for non-local
        filesystems.)

    This algorithm is optimistic and may block for a while if there are a lot
    of concurrent processes working on the same directory.  Also, if you delete
    a symlink created in this way, sequence numbers may repeat.

    The optional 'directory' parameter will create a directory rather than a
    symlink, which is also safe.  As it is an empty directory, it will support
    os.rename'ing another directory over it.  At least on Linux.  At least on
    version 2.4.  10+.  Maybe.
    """
    if not os.path.isdir(dirname):
        makedirs(dirname)
    fn = os.path.join(dirname, ext+".sequence")
    try:
        num = int(readlink(fn))
    except (OSError, IOError):
        num = 0
    while True:
        extname = os.path.join(dirname, str(num)+ext)
        try:
            if directory:
                os.mkdir(extname)
            else:
                symlink(str(num), extname)
            break
        except OSError:
            pass
        num += 1
    try:
        os.remove(fn)
    except OSError:
        pass
    try:
        symlink(str(num), fn)
    except OSError:
        pass
    return extname


class _FilePileStackEntry:
    """I represent a cursor in an open directory; nothing to see here, move
    along, move along.
    """
    
    def __init__(self, dirname, sorter, pilext='pile', itemext='item'):
        
        """Create a pointer into a directory given with dirname.  Pass in
        options from a parent FilePile.  NOTE THAT CMPFUNC WILL BE IGNORED ON
        DIRECTORIES THAT HAVE HAD nextFileName USED WITH THE SAME EXTENSION
        GIVEN IN ITEMEXT.
        """
        self.dirname = dirname
        self.pilext = pilext
        self.itemext = itemext
        self.sorter = sorter
        self.dirlist = None
##         if os.path.exists(os.path.join(self.dirname, '.'+pilext+'.sequence')):
##             self.cmpfunc = LenientIntCompare(pilext=pilext, itemext=itemext)
##         else:
##             self.cmpfunc = self.sorter.comparePathFragments

    def makeChild(self, path):
        """Create a stack entry with similar options for a subpath of the
        directory I am listing.
        """
        return _FilePileStackEntry(os.path.join(self.dirname, path),
                                   self.sorter, self.pilext, self.itemext)

    def listdir(self, forwards=True):
        """If I do not have a cached directory listing, load a directory
        listing and sort it.
        """
        if self.dirlist is None:
            self.dirlist = []
            try:
                files = os.listdir(self.dirname)
            except OSError:
                pass
            else:
                for fn in files:
                    e = self.nameToEnt(fn)
                    if e is not None:
                        self.dirlist.append(e)
            self.dirlist.sort()
            if forwards:
                self.pos = -1
            else:
                self.pos = len(self.dirlist) - 1
        return self.dirlist

    def bisectTo(self, name, adjust=0):
        """Jump my pointer position to point at a file who compares as closely
        as possible to name with self.cmpfunc.
        """
        self.pos = _bisect_cmp(self.listdir(), self.nameToEnt(name)) + adjust

    def nameToEnt(self, name):
        fragment, ext = os.path.splitext(name)
        xt = ext[1:]
        if xt == self.pilext:
            isPile = True
        elif xt == self.itemext:
            isPile = False
        else:
            return None
        if self.sorter.needsSortKey:
            sortKey = self.sorter.fragmentToSortKey(fragment)
        else:
            sortKey = fragment
        return (sortKey, isPile, name)


    def next(self, forwards=True):
        """I AM NOT AN ITERATOR THIS IS JUST A GOOD METHOD NAME.  Get the next
        entry (isSubDirectory, itemOr_FilePileStackEntry).  If forwards is
        False, go backwards.
        """
        dirlist = self.listdir(forwards)
        if forwards:
            if self.pos == len(dirlist) -1:
                return
            self.pos += 1
        else:
            if self.pos < 0:
                return
        sortKey, isPile, name = dirlist[self.pos]
        if not forwards:
            self.pos -= 1
        if isPile:
            val = self.makeChild(name)
        else:
            val = self.sorter.loadItem(
                os.path.join(self.dirname, name))
        return isPile, val

    def prev(self):
        """synonym for .next(False)
        """
        return self.next(False)

# XXX TODO: something like this - make os.mkdirs NFS safe.  probably just means
# we need to wrap each iteration in a try:except:

## def makedirs(self, fn):
##     sp = os.path.split(os.path.abspath(fn))
##     for seg in xrange(len(sp)):
##         cur = os.path.join(*sp[:seg+1])
##         while not os.path.isdir(cur):
##             try:
##                 os.symlink("mkdir", ".mkdir-lock-%s" % cur)
##             except:
##                 pass

class FilePileIterator:
    """I am a sorted list of files stored in a tree of directories.  I map
    filenames to Python objects using a 'loader' factory function.
    """

    def __init__(self, dirname, sorter, pilext='pile', itemext='item'):
        self.dirname = dirname
        self.itemext = itemext
        self.pilext = pilext
        self.stack = [
            _FilePileStackEntry(dirname, sorter, pilext, itemext)]

    def rewind(self):
        """Jump back to the beginning of the top level directory, as if I had
        just been instantiated.
        """
        self.stack = [self.stack[0]]
        self.stack[0].pos = -1
        self.stack[0].dirlist = None

    def jumpTo(self, *path):
        """'Seek' my internal cursor forward to be as close as possible to an
        object at a given path.  This object does not need to exist, but if it
        does, it will be the next object returned by self.next().
        """
        piles = [("%s.%s" % (x, self.pilext)) for x in path[:-1]]
        item = '%s.%s' % (path[-1],self.itemext)
        self.stack = [self.stack[0]]
        for pilename in piles:
            newent = self.stack[-1].makeChild(pilename)
            self.stack[-1].bisectTo(pilename)
            if os.path.isdir(newent.dirname):
                self.stack.append(newent)
        else:
            self.stack[-1].bisectTo(item, adjust=-1)

    def next(self, forwards=True):
        """Get the next item at my current cursor location.  This will always
        ben an object created by self.loader.  If 'forwards' is false, go
        backwards.

        @raise: StopIteration when there are no items left.
        """
        while True:
            result = self.stack[-1].next(forwards)
            if result is None:
                if len(self.stack) == 1:
                    raise StopIteration()
                else:
                    self.stack.pop()
            else:
                isFilePile, item = result
                if isFilePile:
                    self.stack.append(item)
                else:
                    return item

    def prev(self):
        """Synonym for self.next(False).
        """
        return self.next(False)

    def __iter__(self):
        return self

from twisted.python.components import Interface

class ISorter(Interface):
    """
    I am an interface which implements storage of a particular type of thing in
    a FilePile.  I can sort path-fragments and implement the various simple
    functions that a FilePile requires to keep its elements sorted.

    A Sorter locates *paths* to *items* by *keys*.

    An 'abstract path' is a sequence of strings that describes a location in
    the filesystem relative to the directory that the FilePile is in.  For a
    more expansive description, see the docstring for pathFromItem.

    A 'concrete path' is just an absolute path-name on the filesystem.

    An 'item' is a Python object which is loaded somehow.  This is the kind of
    thing which FilePiles are really designed to store.

    A 'key' is an object which is convertible to an 'abstract path'.  For
    example, although the path might be ['2003', '7', '8', '23'], the key may
    be 1057726555.86: a fragment of a localtime tuple and the actual date
    stored in local time.  Keys can be compared to one and other and to items.
    They are used for querying and retrieval.

    @cvar allowDuplicates: A boolean.  Do I allow duplicate entries for the
    same item path?  This affects the way that the result of pathFromItem is
    interpreted: see its doc for details.
    """

    def pathFromItem(self, item):
        
        """Create a list of strings that represent the 'abstract' path to an
        item.  For example, if the filesystem path you want is

            2003.pile/6.pile/4.pile/17.pile/(???).item

        this method should return

            ['2003', '6', '4', '17']

        If the class var 'allowDuplicates' is true, insertions will
        automatically create a monotonically increasing numbered link to hold
        the place of the item being inserted and queries will rely on my
        comparison methods (itemBeforeKey/itemAfterKey) to authoritatively
        compare elements within the same sub-pile.  (It is assumed that items
        will sort in insertion order by those comparisons.)

        Otherwise, this will result in the path:

            2003.pile/6.pile/4.pile/17.item

        The class variable is so named due to the effect on insertions.  If
        duplicates are allowed, insertions of items with matching paths will
        replace them.  Otherwise, they will simply add a new
        sequentially-numbered item in that pile.
        """

    def pathFromKey(self, key):
        """Create an 'abstract' path (described in C{pathFromItem}) from a key.
        """

    def compareItemToKey(self, item, key):
        """This method is similar to cmp().  It returns -1, 0, or 1 depending
        on whether the item is less than, equal to, or greater than the given
        key, respectively.
        """

    def loadItem(self, fullpath):
        """Load an item from the filesystem.
        """

    def saveItem(self, item, fullpath):
        """Save an item to a path on the filesystem.
        """

    def removeItem(self, item, fullpath):
        """Remove an item from the filesystem.
        """


class FilePile:
    """
    I manage trees of filenames and symlinks to help you write storage
    systems.
    """

    def __init__(self, dirname, sorter, pilext='pile', itemext='item'):
        """Create a FilePile.

        @param sorter: an implementor of C{ISorter}.

        @param pilext: The extension to use for subdirectories which represent
        collections, or 'subpiles'.

        @param itemext: The extension to use for files or symlinks in this
        directory whose names will be passed to the 'loader' function, and
        returned from my iteration methods.
        """
        self.dirname = dirname
        self.itemext = itemext
        self.pilext = pilext
        self.sorter = sorter

    def pilePath(self, *path):
        "pilePath('1','2','3') => '<self.dirname>/1.pile/2.pile/3.pile'"
        return os.path.join(self.dirname,*[("%s.%s" % (x, self.pilext)) for x in path])

    def itemPath(self, *path):
        "itemPath('1','2','3') => '<self.dirname>/1.pile/2.pile/3.item'"
        return os.path.join(self.pilePath(*path[:-1]),'%s.%s' % (path[-1],self.itemext))

    def insertLink(self, name, *path):
        """Create an item with proper extensions at the given itemPath.

        @param name: the value of the symlink created.
        @param path: a tuple of names passed to self.itemPath(*path).
        """
        pp = self.pilePath(*path[:-1])
        if not os.path.isdir(pp):
            makedirs(pp)
        symlink(name, self.itemPath(*path))

    def numberedLink(self, *path):
        """Create an item with proper extensions at the given path.

        I use nextFileName; see its documentation for details.

        @param path: a tuple of names passed to self.pilePath(*path).
        """
        pp = self.pilePath(*path)
        return nextFileName(pp, '.'+self.itemext)

    def numberedDirectory(self, *path):
        pp = self.pilePath(*path)
        return nextFileName(pp, '.'+self.itemext, directory=True)

    def backwards(self):
        """Return an iterator that goes backwards over me.  This iterator will
        change my cursor, so watch out for concurrent uses!
        """
        return _b(iter(self))

    def __iter__(self):
        """Return me.  I will act as an iterator that goes forwards over my
        items.  Iterating this will change my cursor, so watch out for
        concurrent uses!
        """
        return FilePileIterator(self.dirname, self.sorter,
                                self.pilext, self.itemext)

    def itemsBetween(self, startKey, endKey=None, stopAfter=None):
        startPath = self.sorter.pathFromKey(startKey)
        itemIter = iter(self)
        itemIter.jumpTo(*startPath)
        return _Betweener(itemIter, self.sorter.compareItemToKey,
                          startKey, endKey, stopAfter)

    def _fullPathFromItem(self, item):
        abstractPath = self.sorter.pathFromItem(item)
        if self.sorter.allowDuplicates:
            fullpath = self.numberedLink(*abstractPath)
        else:
            fullpath = self.itemPath(*abstractPath)
        return fullpath

    def add(self, item):
        fullpath = self._fullPathFromItem(item)
        self.sorter.saveItem(item, fullpath)

    def remove(self, item):
        if self.sorter.allowDuplicates:
            dirpath = self.pilePath(*self.sorter.pathFromItem(item))
            for pathseg in os.listdir(dirpath):
                fullpath = os.path.join(dirpath, pathseg)
                if fullpath.endswith(self.itemext):
                    if self.sorter.loadItem(fullpath) == item:
                        self.sorter.removeItem(item, fullpath)
        else:
            fullpath = self._fullPathFromItem(item)
            self.sorter.removeItem(item, fullpath)

    def atKey(self, key):
        try:
            item = self.itemsBetween(key).next()
        except StopIteration:
            raise KeyError(key)
        cmpr = self.sorter.compareItemToKey(item, key)
        if cmpr != 0:
            raise KeyError(key)
        return item

class _Betweener:
    def __init__(self, itemIter, compare, startKey, endKey, stopAfter):
        self.startKey = startKey
        self.endKey = endKey
        self.stopAfter = stopAfter
        self.itemIter = itemIter
        self.compare = compare

    def __iter__(self):
        return self

    def next(self):
        while True:
            item = self.itemIter.next()
            if self.compare(item, self.startKey) == -1:
                continue
            if self.stopAfter is not None:
                self.stopAfter -= 1
                if self.stopAfter < 0:
                    raise StopIteration
            if self.endKey is not None:
                if self.compare(item, self.endKey) == 1:
                    raise StopIteration
            return item

        

class _b:
    """Backwards iterator for FilePile."""
    def __init__(self, f):
        self.f = f

    def next(self):
        return self.f.prev()

    def __iter__(self):
        return self

from twisted.trial import unittest

def _lenieint(x):
    """Try to convert an object to an int, but if we can't, just give up and
    return the object itself.
    """
    try:
        return int(x)
    except:
        return x

class LenientIntCompare:
    """I am a comparator object which can be used for directories which has had
    objects created in it by nextFileName.
    """
    def __init__(self, itemext='item', pilext='pile', filt=_lenieint):
        self.itemext = itemext
        self.pilext = pilext
        self.filt = filt

    def __call__(self, fn1, fn2):
        fn1p, fn1e = os.path.splitext(fn1)
        fn2p, fn2e = os.path.splitext(fn2)
        
        # note - it might be a good idea to guarantee that piles sort before
        # items (or vice-versa) in order to make certain kinds of sorting
        # easier

        fn1i = fn1e[1:] in (self.itemext, self.pilext)
        fn2i = fn2e[1:] in (self.itemext, self.pilext)
        if fn1i and fn2i:
            return cmp(self.filt(fn1p),
                       self.filt(fn2p))
        elif fn1i:
            return -1
        elif fn2i:
            return 1
        else:
            return cmp(fn1,fn2)


class DefaultSorter:
    __implements__ = ISorter

    loadItem = str
    needsSortKey = False
    compareItemToKey = cmp
    pathFromItem = str
    pathFromKey = str

    def saveItem(self, item, fullpath):
        pass

    def removeItem(self, item, fullpath):
        pass

class DecimalSorter:
    """A simple example of a Sorter, which can be either subclassed or used
    directly to have items sorted by an integer key.  It will make relatively
    sparse (on UNIX, nearly optimal) use of directory spacing.
    """

    exp = 3

    allowDuplicates = False

    def pathFromKey(self, i):
        assert i >= 0, "Keys must not be negative (but got %r)!" % (i,)
        # the very *embodiment* simplicitiy
        exp = self.exp
        l = []
        if i == 0:
            return ['0']
        while i:
            l.append(str(i))
            i = (i // (10 ** exp)) * 10 ** exp
            exp += 1
        l.reverse()
        return l

    def pathFromItem(self, item):
        return self.pathFromKey(int(item))

    def compareItemToKey(self, item, key):
        return cmp(int(item), key)

    def loadItem(self, fullpath):
        return int(os.path.splitext(os.path.basename(fullpath))[0])

    def saveItem(self, item, fullpath):
        try:
            makedirs(os.path.dirname(fullpath))
        except OSError:
            pass
        try:
            symlink(str(int(item)), fullpath)
        except OSError:
            raise
            raise Exception("Couldn't link %r for %r" % (fullpath, item))

    needsSortKey = True
    fragmentToSortKey = int
