# -*- test-case-name: atop.test.test_sched -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import time

from twisted.application.service import Service
from twisted.python.components import Adapter, registerAdapter, Interface
from twisted.python import log

from atop.powerup import Powerup, Storeup, IPowerStation
from atop.store import INDEX_AND_ITEM, SubStore, Pool, DateIndex, registerIndex, ItemMixIn, Item, referenced, indexed
from atop.fwd import ItemForwardingComponent


class IScheduler(Interface):
    """Marker interface for Scheduler.
    """
    def callItemLater(self, item):
        pass
    def clock(self):
        pass

class IScheduledEvent(Interface):
    """Interface which scheduled events must implement.

    Subclass Schedulable and don't worry about it, okay?  Also, subclass item.

    I know, this isn't really a component.  Live with it.
    """

SCHEDULED = registerIndex(float, "scheduled")
SCHEDULED_ERRORS = registerIndex(float, "scheduled_errors")

class Schedulable(ItemMixIn):
    """I am a mix-in for items which wish to schedule timed tasks.
    """

    __implements__ = IScheduledEvent

    status = ''

    scheduled = indexed(SCHEDULED)
    scheduledErrors = indexed(SCHEDULED_ERRORS)

    def schedule(self, newTime):
        """Schedule my run() method to be executed at the given local time.
        (Values returned from time.time() are local times, so use that and
        adjust.)
        """
        if self.scheduled != newTime:
            # print 'Scheduling', self, 'for', time.ctime(newTime)
            self.scheduled = newTime
            assert self.scheduled == newTime
            IScheduler(self.store).callItemLater(self)

    def delay(self, secondsLater):
        """Delay the scheduled call.
        @param secondsLater: how many seconds from its current firing time to delay
        """
        self.schedule(self.getIndexKey(None, "scheduled") + secondsLater)

    def reset(self, secondsFromNow):
        """Reset the scheduled call's timer.
        @param secondsFromNow: how many seconds from now it should fire,
            equivalent to C{self.cancel()} and then doing another
            C{reactor.callLater(secondsLater, ...)}
        """
        self.schedule(IScheduler(self.store).clock() + secondsFromNow)

    def unschedule(self):
        self.setIndexKey("scheduled", None)
        p = IScheduler(self.store).pool
        if self in p:
            p.removeItem(self)

    def run(self):
        """Implement
        """
        raise NotImplementedError("")

class SchedulerBase(Schedulable):
    __implements__ = IScheduler

    clock = staticmethod(time.time)

    MAXIMUM_LOCKS = 50

    pool = referenced

    def setUpPools(self, store):
        p = Pool(store, "Scheduled tasks")
        p.invisible = True
        p.addIndex(DateIndex(SCHEDULED.index))
        p.addIndex(DateIndex(SCHEDULED_ERRORS.index))
        self.pool = p
        getattr(store, 'rootRef', getattr(store, 'rootPoolRef', None)).getItem().addItem(p)

    def getEvents(self):
        return self.pool.queryIndex(SCHEDULED.index, result=INDEX_AND_ITEM)

    def callItemLater(self, item):
        """Call an item later, according to its date index.

        You probably want to use Schedulable.schedule(newTime) rather than
        calling this directly.
        """
        p = self.pool
        if item not in p:
            # These are almost always rescheduled items, not new
            # items, so the check saves us a function call, not
            # to mention cleans up our log output.
            p.addItem(item)
        for localtime, item in self.getEvents():
            self.schedule(localtime)
            return

    def nextEvent(self):
        try:
            return self.getEvents().next()
        except StopIteration:
            return None

    def _oneEvent(self, now, alreadyRan):
        next = self.nextEvent()
        if next is not None and next[0] < now:
            localtime, item = next
            scheduled = item.getIndexKey(None, SCHEDULED.index)
            if int(localtime) != int(scheduled):
                log.msg("Non-fatal (maybe) database inconsistency in the scheduled items pool")
                log.msg("The item in question is %s" % (item,))
                log.msg("The pool believes it should run at %f (%s)" % (localtime, time.ctime(localtime)))
                log.msg("The item believes it should run at %f (%s)" % (scheduled, time.ctime(scheduled)))
                log.msg("I am erasing the pool's belief.")
                item.removeFromIndex(SCHEDULED.index, localtime)
                return
            elif item.storeID in alreadyRan:
                # We're done for this tick
                return True
            alreadyRan[item.storeID] = None
            item.setIndexKey(SCHEDULED.index, None)
            try:
                self.store.subtransact(IScheduledEvent(item).run)
            except:
                # TODO: later this should also actually catch the Failure
                # here and pickle it somewhere.
                log.err()
                item.setIndexKey(SCHEDULED.index, None)
                item.setIndexKey(SCHEDULED_ERRORS.index, localtime)
            else:
                if item.getIndexKey(None, SCHEDULED.index) is None:
                    if item in self.pool:
                        self.pool.removeItem(item)
        else:
            return True

    def tick(self):
        now = self.clock()
        alreadyRan = {}
        self.store.batch(self._oneEvent, now, alreadyRan)
        next = self.nextEvent()
        if next is not None:
            # print 'Next event is', next[1], 'and scheduled at', time.ctime(next[0])
            return next[0]
        # print 'No more events, returning None'
        return None

class SubTicker(Item, Schedulable):
    __implements__ = IScheduledEvent
    substore = referenced

    def __init__(self, substore):
        Item.__init__(self, substore.store)
        self.substore = substore

    def __str__(self):
        return '<SubTicker (%s)>' % (self.substore.name,)

    def run(self):
        when = IScheduler(self.substore).tick()
        if when is None:
            self.unschedule()
        else:
            self.schedule(when)

class SystemScheduler(SchedulerBase, Storeup):
    __implements__ = SchedulerBase.__implements__, Powerup.__implements__
    persistenceForgets = ['nextCall']

    def setUpPools(self, store):
        SchedulerBase.setUpPools(self, store)
        store.setComponent(IScheduler, self)

    nextCall = None

    def callLater(self, secs, f, *a, **b):
        from twisted.internet import reactor
        return reactor.callLater(secs, f, *a, **b)

    def schedule(self, newTime):
        try:
            self.nextCall.cancel()
        except:
            pass
        self.nextCall = None
        if newTime is not None:
            self.nextCall = self.callLater(max(newTime - self.clock(),0), self.callLaterEr)

    def callLaterEr(self):
        nexttime = self.store.transact(self.tick)
        # print 'System Scheduler callLaterer', time.ctime(nexttime)
        self.schedule(nexttime)

class SubScheduler(SchedulerBase, Powerup):
    __implements__ = SchedulerBase.__implements__, Powerup.__implements__
    powerupName = 'Scheduled Tasks'

    storeDependencies = [SystemScheduler]

    def setUpPools(self, substore):
        SchedulerBase.setUpPools(self, substore)
        substore.setComponent(IScheduledEvent, SubTicker(substore))
        substore.setComponent(IScheduler, self)

    def schedule(self, newTime):
        IScheduledEvent(self.store).schedule(newTime)

class SchedulerService(Service, ItemMixIn):

    sched = referenced

    def __init__(self, store):
        # TODO: for store resurrection, this needs to be able to suck it out of
        # the powerups pool.
        self.sched = IPowerStation(store).installPowerup(SystemScheduler)

    def startService(self):
        Service.startService(self)
        self.sched.callLaterEr()

    def stopService(self):
        Service.stopService(self)
        self.sched.schedule(None)

class CallingItem(Item, Schedulable):

    """Helper for transaction-safe reactor.callLater()

    Instead of reactor.callLater(n, f, *a, **kw) simply
    use CallingItem(store, n, f, *a, **kw) in a transaction.
    Note that f, a, and kw must all be serializable.
    """

    def __init__(self, store, n, f, *a, **kw):
        Item.__init__(self, store)
        self.f = f
        self.a = a
        self.kw = kw
        self.reset(n)

    def run(self):
        self.unschedule()
        try:
            self.f(*self.a, **self.kw)
        finally:
            self.delete()
