# -*- test-case-name: atop.test.test_service -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

"""
Twisted application service integration for persistence services.
"""

from twisted.internet import defer
from twisted.application import service

from atop import store
from atop import powerup

class ServicePool(store.Pool):
    def __init__(self, *a, **kw):
        store.Pool.__init__(self, *a, **kw)
        self.addIndexes()

    def addIndexes(self):
        self.addIndex(store.NameIndex())

class Serviceup(powerup.Storeup):
    __implements__ = (service.IService, service.IServiceCollection)

    running = False

    children = store.referenced
    serviceParent = store.referenced

    def setUpPools(self, st):
        self.serviceParent = None
        self.children = ServicePool(st, "children")

        # XXX - This shouldn't really work like this.  Ordering
        # sucks, etc.  Can't think of a better solution for now
        # though.  -jp
        if service.IService(st, default=None) is None:
            st.setComponent(service.IService, self)
            st.setComponent(service.IServiceCollection, self)

    # IServiceCollection implementation
    def __iter__(self):
        return self.children.queryIndex()

    def getServiceNamed(self, name):
        return self.children.getIndexItem('name', name)

    def addService(self, service):
        if service.name is not None:
            if self.children.getIndexItem('name', service.name, default=None) is not None:
                raise RuntimeError("Service named %s already exists" % (service.name,))
        self.children.addItem(service)
        if self.running:
            service.startService()

    def removeService(self, service):
        try:
            self.children.removeItem(service)
        except KeyError:
            raise ValueError(service)
        if self.running:
            return service.stopService()
        return None

    # IService implementation
    def setName(self, name):
        if self.serviceParent is not None:
            raise RuntimeError("Already have a parent, can't change name.")
        self.touch()
        self.name = name

    def setServiceParent(self, parent):
        if self.serviceParent is not None:
            self.disownServiceParent()
        self.serviceParent = parent
        service.IServiceCollection(parent, default=parent).addService(self)

    def disownServiceParent(self):
        parent = self.serviceParent
        d = service.IServiceCollection(parent, default=parent).removeService(self)
        self.serviceParent = None
        return d

    def privilegedStartService(self):
        raise AssertionError("Serviceup.privilegedStartService *cannot* run.")

    def startService(self):
        self.touch()
        self.running = True
        for child in self:
            child.startService()

    def stopService(self):
        self.touch()
        self.running = False
        stopDeferreds = []
        children = list(self)
        children.reverse()
        for child in children:
            stopDeferreds.append(defer.maybeDeferred(child.stopService))
        return defer.DeferredList(stopDeferreds)
