# -*- test-case-name: atop.test.test_store -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

"""
Storage classes for ATOP.
"""

__version__ = '$Revision: 1.135 $'

import warnings
import time
import os
import sys
import new
import shutil
import atexit
import inspect
import itertools
import operator

from cStringIO import StringIO

from os.path import join as opj

from bsddb.db import *

from twisted.python import log
from twisted.python import plugin
from twisted.python import context
from twisted.python import failure
from twisted.python import components
from twisted.python.util import sibpath
from twisted.python.failure import Failure
from twisted.application import service
from twisted.internet import defer
from twisted.internet import task

from twisted.python import reflect
from twisted.python.components import Interface, Componentized, implements, CannotAdapt

from weakref import ref

from twisted.persisted import styles
from twisted.persisted.styles import Versioned, Ephemeral, doUpgrade

from atop.tpython import Forgetter, mkgmtime, FilesystemLock
from atop.fwd import ItemForwardingComponent, makeForwarder
from atop.revert import pickleItem, unpickleItem, PICKLE_PROTOCOL, PROFILING

indexTypes = {}

class IndexHandle(object):
    __slots__ = 'index', 'type'
    def __init__(self, T, index):
        self.type = T
        self.index = index
    def toType(self):
        return indexTypes[self.index]
    def __str__(self):
        return '<IndexHandle %s (%r)>' % (self.index, self.type)

def registerIndex(indexType, indexName):
    """Register a new index: specify the type that one expects from an index.

    Returns an index handle, which is an opaque (non-persistable!)  value.

    If the indexName has already been registered with a different type, raise
    an exception.
    """
    if indexName in indexTypes:
        raise RuntimeError("Index %r already registered." % indexName)
    indexTypes[indexName] = indexType
    return IndexHandle(indexType, indexName)

class IPoolIndex(Interface):
    def getIndexName(self):
        """Return a qualified name for this index, used to look up index_*
        methods on Items."""

    def keyToSortableString(self, key):
        """Convert a given 'key' object (such as a date) to a fixed-length
        sortable string: e.g. the date represented by the float
        1066147720.700978 might be converted to '2003/10/14/12/08/40'
        """

    def sortableStringToKey(self, sortableString):
        """Convert a value produced by keyToSortableString back into its
        original format.
        """

# XXX - these constants really shouldn't exist at all but they should
# ESPECIALLY not exist here !!! they belong in quotient.mimemessage or
# such. -gl

CREATED = 'created'

# the time that this object was first created by the system; when it was
# initially instantiated

RECEIVED = 'received'

# The time that this object was first reliably 'received' by a trusted
# forwarder.  Corresponds to a heuristic based on the last-found rfc822
# 'received' header.

MODIFIED = 'modified'

# Last time this object was changed.

class DateIndex:
    __implements__ = IPoolIndex
    def __init__(self, dateType=RECEIVED):
        self.iname = dateType

    def getIndexName(self):
        return self.iname

    def sortableStringToKey(self, sstr):
        return mkgmtime(tuple(map(int, sstr.split(".")))+(0,0,0))

    def keyToSortableString(self, key):
        if isinstance(key, (list, tuple, time.struct_time)):
            gmt = key
        else:
            gmt = time.gmtime(key)
        return "%04d.%02d.%02d.%02d.%02d.%02d" % gmt[:6]

class IntIndex:
    __implements__ = IPoolIndex

    name = 'int'
    def __init__(self, name='int'):
        self.name = name

    def getIndexName(self):
        return self.name

    def keyToSortableString(self, key):
        return '%0.10d' % key

    def sortableStringToKey(self, sstr):
        return int(sstr)

class UIDIndex(IntIndex):
    __implements__ = IPoolIndex

    name = 'uid'
    def __init__(self):
        IntIndex.__init__(self, self.name)

class NameIndex:
    __implements__ = IPoolIndex
    name = 'name'

    def __init__(self, name='name'):
        self.name = name

    def getIndexName(self):
        return self.name

    def keyToSortableString(self, key):
        return key

    def sortableStringToKey(self, sstr):
        return sstr

class CaseInsensitiveIndex(NameIndex):
    def keyToSortableString(self, key):
        return key.strip().lower()

class PowerupIndex(NameIndex):
    __implements__ = IPoolIndex
    def getIndexName(self):
        return 'powerup'

class ItemReferenceIndex(Forgetter):
    '''Allow access to a "foreign key"
    '''

    persistenceForgets = ['store']
    contextRemembers = [('store','store')]

    def __init__(self, attrname, store):
        self.name = attrname
        self.store = store

    def getIndexName(self):
        return self.name

    def keyToSortableString(self, key):
        # key ought to be an ItemReference.
        return '%0.10d' % key.storeID

    def sortableStringToKey(self, sstr):
        return ItemReference(self.store, int(sstr))

NAME = registerIndex(NameIndex, 'name')

# This is a value that is greater than anything that might show up inside
# one part of a key string, but less than a special magic value that lets
# us know when a particular prefix has ended.  It cannot show up in any parts
# of any key names inserted into the database.
SEP = chr(0x00)
END = chr(0xff)

# We use this because index keys contain their storeIDs to allow multiple
# values per key.  When we do queries we need to make sure that they ignore the
# storeID.

MAGIC = chr(0x01)

def md(dirname):
    if os.path.isdir(dirname):
        return False
    os.makedirs(dirname)
    return True

MAX_LOCKS = 1000
MAX_LOCKERS = 1000
MAX_OBJECTS = 1000

LOCK_TIMEOUT = 1
TXN_TIMEOUT = 1

DB_LOCK_NAME = 'lock'

class DatabaseAlreadyOpen(RuntimeError):
    def __str__(self):
        return 'Database Already Open: %s' % RuntimeError.__str__(self)

# This flag is really new.  Not using it will slow things
# down a lot, but at least it will work.
try:
    DB_TXN_WRITE_NOSYNC
except NameError:
    log.msg("Warning: DB_TXN_WRITE_NOSYNC not available.")
    DB_TXN_WRITE_NOSYNC = 0

class _EnvAndDB:
    cleanedUp = False

    def __init__(self, env, db, lock = None):
        self.env = env
        self.db = db
        self.lock = lock
        atexit.register(self._cleanup)

    def __len__(self):
        if self.lock is not None:
            return 3
        return 2

    def __getitem__(self, index):
        if self.lock is not None:
            return (self.env, self.db, self.lock)[index]
        return (self.env, self.db)[index]

    def _cleanup(self):
        if self.cleanedUp:
            return
        log.msg("Warning: Cleaning up after dirty database/environment")
        self.cleanup()

    def cleanup(self):
        # XXX - In a transaction, this code path is a SEGV.
        assert context.get('txn') is None
        self.db.close()
        self.env.close()
        if self.lock is not None:
            self.lock.unlock()
        self.db = self.env = self.lock = None
        self.cleanedUp = True

class DatabaseEnvironment:
    # Name of the lock file to protect this environment with
    lockFileName = None

    maxLocks = None
    maxLockers = None
    maxLockObjects = None

    lockTimeout = None
    txnTimeout = None

    envJoinFlags = DB_JOINENV

    envOpenFlags = (DB_INIT_MPOOL   # Init shared memory subsystem
                    | DB_INIT_LOG   # Init logging subsystem
                    | DB_INIT_LOCK  # Initialize locking subsystem
                    | DB_INIT_TXN   # Make environment transactional
                    | DB_CREATE)    # Create files as necessary

    envFlags = (DB_TXN_WRITE_NOSYNC
                | DB_AUTO_COMMIT)

    normalRecoveryFlags = DB_RECOVER
    fatalRecoveryFlags = DB_RECOVER_FATAL

    def prepareEnvironment(self, env):
        env.set_lk_max_locks(self.maxLocks or MAX_LOCKS)
        env.set_lk_max_lockers(self.maxLockers or MAX_LOCKERS)
        env.set_lk_max_objects(self.maxLockObjects or MAX_OBJECTS)
        env.set_timeout(self.lockTimeout or LOCK_TIMEOUT, DB_SET_LOCK_TIMEOUT)
        env.set_timeout(self.txnTimeout or TXN_TIMEOUT, DB_SET_TXN_TIMEOUT)
        env.set_flags(self.envFlags, 1)

    def openEnvironment(self, envdir, lockClean):
        if lockClean:
            # This *could* be 0 (no flag), but it doesn't
            # hurt (except it's slow) to recover a consistent
            # database like this, so we'll play it safe for
            # the time being.
            extraFlags = self.normalRecoveryFlags
        else:
            log.msg("Lock acquired uncleanly.  Performing fatal recovery.")
            extraFlags = self.fatalRecoveryFlags

        env = DBEnv()
        try:
            self.prepareEnvironment(env)
            try:
                log.msg("Opening environment...")
                env.open(envdir, self.envOpenFlags | extraFlags)
            except DBRunRecoveryError:
                if lockClean:
                    log.msg("FATAL RECOVERY FAILED.  CALL SLEEPYCAT IMMEDIATELY.")
                else:
                    log.msg("Regular recovery failed.")
                raise
            log.msg("Environment opened.")
        except:
            env.close()
            raise
        return env

    def openDatabase(self, env, datafile):
        db = DB(env)
        try:
            txn = env.txn_begin()
            try:
                log.msg("Opening database...")
                db.open(datafile, "db", DB_BTREE, DB_CREATE, txn=txn)
            except:
                txn.abort()
                db.close()
                env.close()
                raise
            txn.commit()
            log.msg("Database opened.")
        except:
            db.close()
            raise
        return db

    def join(self, envdir, datafile='db'):
        envdir = os.path.abspath(envdir)
        env = DBEnv()
        try:
            env.open(envdir, self.envJoinFlags)
            db = self.openDatabase(env, datafile)
        except:
            env.close()
            raise
        return _EnvAndDB(env, db)

    def fixupOldEnvironment(self, datadir):
        """Backwards compatibility upgrader
        """
        envdir = opj(datadir, 'env')
        if os.path.isdir(envdir):
            for fname in os.listdir(envdir):
                os.rename(opj(envdir, fname), opj(datadir, fname))
            os.rmdir(envdir)
            log.msg("Re-arranged directory structure beneath " + datadir)

    def open(self, envdir, datafile='db'):
        """Open a BSDDB environment and a single database.
        """
        envdir = os.path.abspath(envdir)
        md(envdir)

        lockfile = opj(envdir, self.lockFileName or DB_LOCK_NAME)
        lockObj = FilesystemLock(lockfile)

        if not lockObj.lock():
            raise DatabaseAlreadyOpen(envdir, lockfile)
        try:
            self.fixupOldEnvironment(envdir)

            env = self.openEnvironment(envdir, lockObj.clean)
            try:
                db = self.openDatabase(env, datafile)
            except:
                env.close()
                raise
        except:
            if lockObj.clean:
                lockObj.unlock()
            raise
        return _EnvAndDB(env, db, lockObj)

_dbenv = DatabaseEnvironment()
envdblock = _dbenv.open
envdbjoin = _dbenv.join

def retrieveIDKey(db, keyName):
    return int(db.get(keyName, txn=context.get('txn')) or 0)

def updateIDKey(db, keyName):
    txn = context.get('txn')
    if txn is None:
        raise NotImplementedError("Updating ID outside of transaction")
    val = db.get(keyName, flags=DB_RMW, txn=txn)
    if val is None:
        val = 0
    ret = int(val)
    db.put(keyName, str(ret + 1), txn=txn)
    return ret

class LoginNotFound(KeyError):
    pass

class DuplicateLogin(KeyError):
    pass

class NoSuchIndex(KeyError):
    pass

class Inconsistency(Exception):
    """
    A slightly weaker warning than an assert.  Something bad happened, but it
    was recoverable, and it's the kind of thing we need to consistently check
    for at runtime (it can't be disabled because it's not really specified by
    any invariant, and could be exploited by attackers).
    """

class AtomicFile(file):
    """I am a file which is moved from temporary to permanent storage when it
    is closed.

    After I'm closed, I will have a 'finalpath' property saying where I went.
    """
    def __init__(self, tempname, destdir):
        self._destdir = destdir
        file.__init__(self, tempname, 'wb')

    def close(self, destpath):
        file.close(self)
        fullpath = os.path.join(self._destdir, destpath)
        md(os.path.dirname(fullpath))
        self.finalpath = fullpath
        os.rename(self.name, fullpath)
        return fullpath

    def abort(self):
        os.unlink(self.name)

tempCounter = itertools.count()

def noLogout():
    pass

import traceback
def _fcremover(w, k, f):
    def remove(o):
        # Weakref callbacks cannot raise exceptions
        # or DOOM ensues
        try:
            try:
                f()
            except:
                log.msg("Exception in finalizer cannot be propegated")
                log.err()
            self = w()
            if self is not None:
                del self.data[k]
        except:
            try:
                traceback.print_exc(file=file('_fcremover.log', 'a'))
            except:
                try:
                    fh = file('_fcremover.log.log', 'a')
                    fh.write('Doom doom doom\n')
                    t, i, tb = sys.exc_info()
                    fh.write(str(i) + '\n')
                    fh.flush()
                    fh.close()
                except:
                    try:
                        fh = file('_fcremover.log.log.log', 'a')
                        fh.write('Tried really hard and failed to log an exception')
                        fh.flush()
                        fh.close()
                    except:
                        # Jesus christ.
                        pass
    return remove

class StrongCache(service.Service):
    """Holds strong references to objects, for a while.
    """

    interval = 10
    index = 0

    maxBufferSize = 1000
    defaultRingSize = 100

    def __init__(self, size=None):
        self.ringSize = size

    def startService(self):
        service.Service.startService(self)
        self.data = [{} for x in range(self.ringSize or self.defaultRingSize)]
        self.call = task.LoopingCall(self.tick)
        self.dcall = self.call.start(self.interval)
        self.dcall.addErrback(self._ebLoop)

    def stopService(self):
        service.Service.stopService(self)
        self.data = None
        if self.call is not None:
            self.call.stop()

    def cache(self, path, item):
        self.data[self.index][path] = item
        if len(self.data[self.index]) >= self.maxBufferSize:
            self.tick()

    def grow(self, amount=1):
        self.data[self.index + 1:self.index + 1] = [{} for x in range(amount)]

    def tick(self):
        self.index += 1
        self.index %= len(self.data)
        self.data[self.index].clear()

    def _ebLoop(self, failure):
        self.call = None
        self.dcall = None
        log.err(failure)

class FinalizingCache:
    """Possibly useful for infrastructure?  This would be a nice addition (or
    perhaps even replacement) for twisted.python.finalize.
    """
    def __init__(self):
        self.data = {}
        if not PROFILING:
            # see docstring for 'has'
            self.has = self.data.has_key

    def cache(self, key, value):
        fin = value.__finalizer__()
        self.data[key] = ref(value, _fcremover(ref(self), key, fin))
        return value

    def uncache(self, key):
        del self.data[key]

    def has(self, key):
        """Does the cache have this key?

        (This implementation is only used if the system is being profiled, due
        to bugs in Python's old profiler and its interaction with weakrefs.
        Set the module attribute PROFILING to True at startup for this.)
        """
        if key in self.data:
            o = self.data[key]()
            if o is None:
                del self.data[key]
                return False
            return True
        return False

    def get(self, key):
        o = self.data[key]()
        if o is None:
            raise Inconsistency("FinalizingCache has %r but its value is no more." % (key,))
        log.msg(db_cache="hit", key=key)
        return o



class Store(Versioned):
    try:
        from twisted.python.components import IComponentized
        __implements__ = IComponentized
    except ImportError:
        log.msg('TODO: add IComponentized to Twisted so that we can do this without relying on getattr magic.')

    store = None
    persistenceForgets = ['cache', 'env', 'db']

    lock = None
    closed = False

    service = None

    HYGENE_PERIOD = 60 * 60

    def __init__(self, datadir, filedir, archdir=None, join=False):
        """
        @param datadir: The directory in which to instantiate database files.
        @param filedir: The directory into which to place message files.
        @param archdir: The directory into which environment log files which
        are no longer needed will be moved.
        """
        self.filedir = filedir
        self.datadir = datadir
        self.tempdir = os.path.join(self.filedir, 'tmp')

        if archdir is None:
            archdir = sibpath(self.filedir, 'archive')
        self.archdir = archdir

        md(self.tempdir)
        dataWasntThere = md(self.datadir)
        self.setup(join)
        try:
            if dataWasntThere:
                self.transact(self.live)
            else:
                self.transact(self.resurrect)
        except:
            self.close()
            raise

    def resurrect(self):
        """Re-instantiate a database.
        """
        self.rootRef = ItemReference(self, 0)
        pupool = self.rootRef.getItem().getIndexItem('name', 'powerups')
        self.powerupsRef = pupool.referenceTo()
        from atop.powerup import PowerStation
        self.powerstation = PowerStation(self)
        self.powerstation._powerupsPool = self.powerupsRef

        # really, this ought to be handled by a different implementation of
        # IPowerStation considering the extra specialness that Store has to do,
        # but this at least fixes incorrect behavior with a relatively small
        # change... -gl

        for pup in pupool.queryIndex():
            self.powerstation.powerups[reflect.qual(pup.__class__)] = pup.referenceTo()

        self._databaseHygeneLoop = task.LoopingCall(self._databaseHygene)
        self._databaseHygeneLoop.start(self.HYGENE_PERIOD)

    def live(self):
        """This is the first time this store has been initialized.  Perform
        in-database initialization to create various objects which we will
        need.
        """

        # This is being done in this apparently round-about way (with root
        # vs. powerup pool) because we _need_ to be able to resurrect these
        # items deterministically if we lose track of the original Store
        # object.  If we lose the .tap, we need to be able to re-create it

        root = Pool(self, "root")
        root.addIndex(NameIndex())
        powerups = Pool(self, "powerups")
        powerups.addIndex(PowerupIndex())
        root.addItem(powerups)
        from atop.powerup import PowerStation
        self.powerstation = PowerStation(self)
        self.rootRef = root.referenceTo()
        self.powerupsRef = powerups.referenceTo()
        self.powerstation._powerupsPool = self.powerupsRef

        self._databaseHygeneLoop = task.LoopingCall(self._databaseHygene)
        self._databaseHygeneLoop.start(self.HYGENE_PERIOD)

    def _databaseHygene(self):
        # XXX - I /think/ self.backup() belongs in the middle
        self.checkpoint()
        # self.backup()
        self.archive()

    def listArchivable(self):
        """Return a list of filenames of archivable log files.

        The returned names are relative to C{self.datadir}.
        """
        return self.env.log_archive()

    def archive(self):
        """Archive all archivable log files.
        """
        arch = self.listArchivable()
        if not os.path.exists(self.archdir):
            os.mkdir(self.archdir)
        for fname in arch:
            shutil.move(opj(self.datadir, fname), opj(self.archdir, fname))
        if arch:
            log.msg("Archived " + ', '.join(arch))

    def checkpoint(self):
        """Checkpoint the database environment.

        This allows old environment logfiles to be archived.
        """
        log.msg("Checkpointing database...")
        self.env.txn_checkpoint()
        log.msg("Database checkpointed.")

    def touch(self):
        """This is a no-op for convenience.
        """

    def getComponent(self, interface, registry=None, default=None):
        # I am my own registry, really.
        from atop.powerup import IPowerStation
        if interface == IPowerStation:
            return self.powerstation
        ppool = self.powerupsRef.getItem()
        if context.get('txn') is None:
            warnings.warn("Calling Store.getComponent() outside a transaction is deprecated!",
                          DeprecationWarning, stacklevel=3)
        pitem = ppool.getIndexItem("powerup", reflect.qual(interface), default=default)
        if isinstance(pitem, ItemForwardingComponent):
            return pitem.fwdTo.getItem()
        return pitem

    def setComponent(self, interface, newItemComponent):
        if newItemComponent.store is not self:
            raise Inconsistency("You may only store top-level items as top-level powerups.")
        pups = self.powerupsRef.getItem()
        if newItemComponent not in pups:
            pups.addItem(newItemComponent)
        newItemComponent.appendToIndex("powerup", reflect.qual(interface))

    def unsetComponent(self, interfaceClass):
        try:
            oldItemComponent = interfaceClass(self)
        except CannotAdapt:
            raise KeyError(interfaceClass)
        else:
            pups = self.powerupsRef.getItem()
            oldItemComponent.removeFromIndex("powerup", reflect.qual(interfaceClass))
            for (iface, storeID) in pups.queryIndex('powerup', result=INDEX_AND_STORE_ID):
                if oldItemComponent.storeID == storeID:
                    break
            else:
                pups.removeItem(oldItemComponent)

    def removeComponent(self, itemComponent):
        if itemComponent.store is not self:
            raise Inconsistency("You cannot remove a foreign component from the store.")
        pups = self.powerupsRef.getItem()
        ifaces = []
        for (ifaceName, storeID) in pups.queryIndex(result=INDEX_AND_STORE_ID):
            if itemComponent.storeID == storeID:
                iface = reflect.namedAny(ifaceName)
                self.unsetComponent(iface)
                ifaces.append(iface)
        return ifaces

    def newSubFileForItem(self, substore):
        tmpname = os.path.join(self.tempdir, str(tempCounter.next())+".tmp")
        destpath = os.path.join(self.filedir, str(substore.storeID))
        return AtomicFile(tmpname, destpath)

    def newSubDirForItem(self, item, name):
        thisPath = os.path.join(self.filedir, str(item.storeID), name)
        md(thisPath)
        return thisPath

    def __getstate__(self):
        if context.get('item') is not None:
            fmt = "Serializing the Store from an Item (%r) in the Store (Check out %s)."
            item = context.get('item')
            raise Inconsistency(fmt % (item, _refPath(item, self)))
        return Versioned.__getstate__(self)

    def __setstate__(self, state):
        self.__dict__ = state
        self.setup()

    def setup(self, join=False):
        self.cache = FinalizingCache()
        if join:
            dbAndEnv = envdbjoin(self.datadir)
        else:
            dbAndEnv = envdblock(self.datadir)
        self.env, self.db = dbAndEnv.env, dbAndEnv.db
        self.cleanup = dbAndEnv.cleanup
        self.closed = False

    def close(self):
        if self.closed:
            raise ValueError("Store already closed")

        self.cleanup()
        self._databaseHygeneLoop.stop()
        self._databaseHygeneLoop = self.db = self.env = self.cleanup = None
        self.closed = True

    def nextID(self, allocate=True):
        """Allocate and return the next available store ID.
        """
        f = allocate and updateIDKey or retrieveIDKey
        return f(self.db, "next" + SEP + "id")

    def nextSubID(self, item, allocate=True):
        """Allocate and return the next available store ID for the given item.

        These IDs are independent of any other IDs, including the ones for this
        Store instance.
        """
        f = allocate and updateIDKey or retrieveIDKey
        return f(self.db, self._formatIDPath(
            item.getItemPath()) + SEP+'next'+SEP+'id')

    def setSubKey(self, item, key, value):
        ipfx = self._formatIDPath(item.getItemPath())
        self.db.put(ipfx + key, value, txn=context.get('txn'))

    def delSubKey(self, item, key, value):
        ipfx = self._formatIDPath(item.getItemPath())
        self.db.delete(ipfx + key, txn=context.get('txn'))

    def _loginkey(self, path, method):
        return 'login'+SEP+'%s/%s' % (method, path)

    def _loadItemByPath(self, ctxstore, path, alreadyLoadedItem=None):
        """
        Retrieve an item given a sequence of integers.

        @param alreadyLoadedItem: None, or an already-loaded-already-cached
        version of this item.
        """
        stringKey = self._formatIDPath(path)
        stringValue = self.db.get(stringKey, txn=context.get('txn'))
        if stringValue is None:
            return None
        myObject = context.call(
            {"store": ctxstore,
             "item": alreadyLoadedItem}, unpickleItem, stringValue)

        if alreadyLoadedItem is None:
            self.cacheItem(myObject, path)
            ctx = context.get('NewlyLoadedContext')
            if ctx is None:
                warnings.warn("Loading items outside of a transaction is deprecated!",
                    DeprecationWarning, stacklevel=4)
            else:
                ctx[path] = None

        log.msg(path=path, db_get=None, item=myObject, pickle=stringValue)
        if styles.versionedsToUpgrade:
            self.transact(doUpgrade)
        return myObject

    def cacheItem(self, item, path):
        assert isinstance(path, tuple)
        if self.cache.has(path):
            assert self.cache.get(path) is item, "inconsistency!?!  the db manages this now.  Cached item %r at %r is not %r" % (self.cache.get(path), path, item)
        else:
            self.cache.cache(path, item)

    def uncacheItem(self, item, path):
        assert isinstance(path, tuple)
        assert self.cache.get(path) is item, "inconsistency!?!  the db manages this now.  Trying to uncache item %r at %r is not %r" % (self.cache.get(path), path, item)
        self.cache.uncache(path)


    def _saveItemToPath(self, item, path):
        self.cacheItem(item, path)
        stringKey = self._formatIDPath(path)
        stringValue = context.call({'item': item}, pickleItem, item, PICKLE_PROTOCOL)
        log.msg(path=path, db_put=None, item=item, pickle=stringValue)
        self.db.put(stringKey, stringValue, txn=context.get('txn'))

    def _removeItemFromPath(self, item, path):
        self.uncacheItem(item, path)
        stringKey = self._formatIDPath(path)
        log.msg(path=path, db_del=None, item=item)
        self.db.delete(stringKey, txn=context.get('txn'))

    def _getItemByPath(self, ctxstore, *path):
        if self.cache.has(path):
            return self.cache.get(path)
        else:
            obj = self._loadItemByPath(ctxstore, path)
            if obj is None:
                raise KeyError(path)
            # self.cacheItem(obj, path)
            obj.activate()
            return obj

    def getItemByID(self, itemID):
        """Return the item from this Store with the given ID.

        @type itemID: C{int}
        """
        return self._getItemByPath(self, itemID)

    def getSubItem(self, subStore, subItemID):
        p = subStore.getItemPath()
        p.append(subItemID)
        return self._getItemByPath(subStore, *p)

    def lockItem(self, subItem):
        self.lockSubItem(self, subItem)

    def lockSubItem(self, subStore, subItem):
        p = subItem.getItemPath()
        stringKey = self._formatIDPath(p)
        self.db.get(stringKey, flags=DB_RMW, dlen=0, doff=0, txn=context.get('txn'))
        log.msg(db_lock=stringKey, path=p, item=subItem)

    def _inMemoryRevert(self, collection, skip):
        """Revert a collection of objects to their state in the database.
        """
        for (path, item) in collection:
            if path in skip:
                if self.cache.has(path):
                    self.cache.uncache(path)
            else:
                dupItem = self._loadItemByPath(item.store, path, item)
                if dupItem is not None:
                    assert dupItem is item
                elif self.cache.has(path):
                    self.cache.uncache(path)
                else:
                    log.msg("Reversion peculiarity: %s" % (item,))

    def new_txn(self):
        return self.env.txn_begin(context.get('txn'))

    def subtransact(self, f, *a, **k):
        """Open a database transaction object and commit it.
        """
        txn = self.new_txn()
        storageContext = {}
        antiStorageContext = {}
        newlyLoadedContext = {}
        cursorFactory = CursorFactory(self, txn)
        try:
            retvalue = context.call({'store': self,
                                     'txn': txn,
                                     'StorageContext': storageContext,
                                     'AntiStorageContext': antiStorageContext,
                                     'NewlyLoadedContext': newlyLoadedContext,
                                     'CursorFactory': cursorFactory},
                                    f, *a, **k)
            context.call({'txn': txn},
                         self._inMemoryCommit,
                         storageContext.iteritems(),
                         antiStorageContext.iteritems())
        except:
            t, v, tb = sys.exc_info()
            # log.err(failure.Failure(v, t, tb))
            cursorFactory.abort()
            txn.abort()
            log.msg(db_txn="abort")
            # is there a way to make this transaction read-only?  -glyph
            revertReadTxn = self.new_txn()
            try:
                try:
                    context.call({'txn': revertReadTxn},
                                 self._inMemoryRevert,
                                 storageContext.iteritems(),
                                 newlyLoadedContext)
                finally:
                    revertReadTxn.abort()
            except:
                log.msg("In memory revert failed.  Things are going to get weird.")
                log.err()
            raise t, v, tb
        else:
            cursorFactory.commit()
            txn.commit()            # this can fail with a RUNRECOVERY - that would be bad.
            log.msg(db_txn="commit")
            return retvalue

    def _inMemoryCommit(self, storeCollection, deleteCollection):
        for (path, item) in storeCollection:
            # save the item by ID, either overwriting or adding
            self._saveItemToPath(item, path)
        for (path, item) in deleteCollection:
            self._removeItemFromPath(item, path)

    def transact(self, f, *a, **k):
        """
        Run a function or method within the context of the storage system.

        This will keep track of all items created or touched during the
        duration of this function and save or revert them based on whether the
        function completes successfully.
        """
        if context.get('txn'):
            return f(*a,**k)
        return self.subtransact(f, *a, **k)

    def transback(self, result, f, *a, **k):
        """Utility method for using 'transact' as a callback to a Deferred.

        All this does is reverse the order of the first two arguments to
        'transact'.  Its common usage is:

            d = doSomethingLater()
            d.addCallback(store.transback, F, 'hello')

        As you might imagine, this causes F to be called as

            F(deferred-result, 'hello')
        """
        return self.transact(f, result, *a, **k)

    def batch(self, f, *a, **kw):
        """Run a batch processing function as many times as is possible.

        f(*a, **kw) will be invoked until time or database limits are reached.
        Each invocation of f should perform a small amount of work towards a
        larger goal.  If f(*a, **kw) returns a True value, processing will cease.

        @return: The number of times the function was called.
        """
        count = 1
        stat = self.env.lock_stat
        def locksRemaining():
            lockInfo = stat()
            return lockInfo['maxlocks'] - lockInfo['nlocks']
        startLocks = locksRemaining()
        if f(*a, **kw):
            return count
        lockChange = startLocks - locksRemaining()
        minLocks = max(startLocks - 20, lockChange)
        while locksRemaining() > minLocks:
            count += 1
            if f(*a, **kw):
                return count
        return count

    def _formatIDPath(self, path):
        L = ['%0.10d'] * len(path)
        L.insert(0, 'id')
        return ((SEP.join(L)) % tuple(path))

class CursorFactory:
    def __init__(self, store, txn):
        self.store = store
        self.txn = txn
        self.cursors = []
        self.done = False

    def cursor(self):
        if self.done:
            raise Inconsistency("Re-use of a storage factory.  You suck!")
        self.cursors.append(self.store.db.cursor(self.txn))
        return self.cursors[-1]

    def closeCursors(self):
        self.done = True
        for cursor in self.cursors:
            cursor.close()

    def commit(self):
        self.closeCursors()

    def abort(self):
        self.closeCursors()

class ItemReference:
    def __init__(self, store, storeID, iface=None):
        self.store = store
        self.storeID = storeID
        self.iface = iface

    def getItem(self):
        if self.iface is not None:
            return self.iface(self.store.getItemByID(self.storeID))
        return self.store.getItemByID(self.storeID)

    def __str__(self):
        return '<ItemReference to %s, %d>' % (getattr(self.store, 'storeID', None), self.storeID)

    def __hash__(self):
        return hash((self.store, self.storeID))

    def __setstate__(self, state):
        self.__dict__ = state
        if not hasattr(self, 'store'):
            st = context.get('store')
            if st is None:
                raise Inconsistency("You're loading an item reference outside the context of a store.")
            self.store = st

    def __ne__(self, other):
        #PYTHON IS SO AWESOME
        return not self.__eq__(other)

    def __eq__(self, other):
        if not isinstance(other, ItemReference):
            return False
        return (other.store is self.store) and (other.storeID == self.storeID)

    def __getstate__(self):
        # get the store we're saving to.
        topitem = context.get('item')
        if topitem is None:
            return self.__dict__
##             raise Inconsistency("Why are you saving an item reference outside the scope of an item?  "
##                                 "Consider using RemoteItemReference (when it becomes available.  "
##                                 "Talk to glyph.")
        elif topitem.store is self.store:
            # If the store we're saving to is the same store as the one for the
            # item we're storing a reference to, remember it from context on
            # load.
            dct = self.__dict__.copy()
            del dct['store']
            return dct
        elif topitem is self.store:
            # Most likely, what's happening here is that we're saving a
            # reference to a sub-item of the avatar within the avatar itself.
            # This will only happen in substores, because 'topitem is
            # self.store' means it's a sub-item of the top item which we are
            # currently saving.

            # In this case it's perfectly safe to return a synchronous
            # reference to the store, because it's going to be remembered by
            # the pickler.
            return self.__dict__
        else:
            if self.getItem() is topitem.store:
                raise Inconsistency("Trying to store an ItemReference on %s to its own store! Don't *do* that." % (topitem,))

            fmt = "Ref to %s stored from %s in %s should be %s (Check out %s)."
            raise Inconsistency(fmt % (self.getItem(),
                                       topitem,
                                       topitem.store,
                                       self.getItem().store,
                                       _refPath(topitem, self)))

def _refPath(a, b):
    # Convenience la la
    from twisted.python.reflect import objgrep, isSame
    return objgrep(a, b, isSame)

class indexed(object):
    """Descriptive of a property which is indexed by a particular index.
    """
    __slots__ = ['handle']
    def __init__(self, handle):
        """Create an indexed descriptor.

        Usage is best described with an example::

            MY_DATE_THINGY = registerIndex(float, 'date')
            ...
            class X(Item):
                foo = indexed(MY_DATE_THINGY)
                def doIt(self):
                    self.touch()
                    self.foo = Y(self.store)
        """
        self.handle = handle

    def __get__(self, obj, type=None):
        if obj is None:
            # crazy, man.  crazy.
            return None
        val = obj.getIndexKey(None, self.handle.index)
        if issubclass(self.handle.type, Item) and val is not None:
            return val.getItem()
        else:
            return val

    def __set__(self, obj, value):
        if issubclass(self.handle.type, Item) and value is not None:
            value = value.getItem()
        obj.setIndexKey(self.handle.index, value)

class referenced(object):
    """Use this on an Item to automatically refer to other items.  For example::

    class Foo(Item):
        ...
    class Bar(Item):
        fooX = referenced
        fooY = referenced
        def setFoos(self, x, y):
            'x and y are Foo instances'
            self.fooX = x
            self.fooY = y
        def getFoos(self):
            'returns a tuple of 2 Foo instances'
            return self.fooX, self.fooY
    """

    def __init__(self, iface):
        self.iface = iface

    def setName(self, attrname):
        self.name = "_"+attrname+"Ref"

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        x = getattr(obj, self.name, None)
        if x:
            if self.iface:
                return self.iface(x.getItem())
            else:
                return x.getItem()

    def __set__(self, obj, value):
        if value is not None:
            if hasattr(obj, 'store') and value.store is not obj.store:
                raise Inconsistency("You can't reference foreign items directly.")
            if self.iface:
                value = IPersistable(value).referenceTo()
            else:
                value = value.referenceTo()
        setattr(obj, self.name, value)

def transacted(method):
    """Use this on an Item to automatically run a method in a transaction.
    For example::

    class Foo(Item):
        def bar(self, aPool):
            aPool.addItem(self)
        bar = transacted(bar)
    """
    def _(self, *a, **kw):
        return self.store.transact(method, self, *a, **kw)
    return new.function(_.func_code, _.func_globals, _.func_name,
                        inspect.getargspec(_), _.func_closure)

class MetaItem(type):
    """I automate the various indexed attributes of an Item.
    """
    def __new__(metaclass, name, bases, ns):
        for k, v in ns.items():
            if v is referenced:
                r = v(None)
                r.setName(k)
                ns[k] = r
            if isinstance(v, referenced):
                ns[k] = v
                v.setName(k)
        return type.__new__(metaclass, name, bases, ns)


class ItemMixIn(object):
    """I am a base class for mix-ins designed for Items in order to create appropriate properties.
    """
    __metaclass__ = MetaItem


class Item(object, Forgetter, Versioned):

    __metaclass__ = MetaItem

    persistenceForgets = ['store']
    contextRemembers = [('store', 'store')]

    name = indexed(NAME)

    def __init__(self, store, name=None):
        self.storeID = store.nextID()
        self.store = store
        self.store.cacheItem(self, tuple(self.getItemPath()))
        if name is not None:
            self.name = name
        self.dateCreated = time.gmtime(time.time())
        self.touch()

    def __str__(self):
        return '<%s %r at 0x%x>' % (self.__class__.__name__, self.name, id(self))

    def __statecheck__(self):
        if self is not context.get('item'):
            fmt = "You were saving %s but I found %s in it (Check out %s)."
            item = context.get('item')
            raise Inconsistency(fmt % (item, self, _refPath(item, self)))

    def __getstate__(self):
        self.__statecheck__()
        return Forgetter.__getstate__(self)

    def fwd(self, *interfaces):
        if not interfaces:
            interfaces = self.__implements__
        return makeForwarder(interfaces)(self.referenceTo())

    def _antitouch(self):
        usctx = context.get('AntiStorageContext')
        if usctx is None:
            raise Inconsistency("You're deleting an object outside a transaction.")
        ikey = tuple(self.getItemPath())
        usctx[ikey] = self

    def delete(self):
        """Remove this Item from the database.
        """
        # XXX - We should probably nuke our cache directories too
        # But we can't do that here, because the txn might be
        # reverted.
        self._antitouch()
        for p in self.iterpools():
            p.removeItem(self)

    def referenceTo(self, iface=None):
        return ItemReference(self.store, self.storeID, iface)
    getRef = referenceTo

    def getItemPath(self):
        """Return a list describing this item's location in the main store.
        """
        ids = [self.storeID]
        ss = self.store
        while ss.store is not None:
            ids.append(ss.storeID)
            ss = ss.store
        ids.reverse()
        return ids

    def setDataKey(self, key, value):
        return self.store.setSubKey(self, key, value)

    def delDataKey(self, key, value):
        return self.store.delSubKey(self, key, value)

    def __finalizer__(self):
        return lambda : None

    def activate(self):
        """This method is called when this item has been loaded from the DB.
        """

    def reactivate(self, oldAttributes):
        """This method is called when this item has been re-loaded from the DB
        during a transaction revert.  Try to avoid putting complex code here;
        bugs here will likely be very difficult to understand.

        @param oldAttributes: a dict of the old {names: values} of attributes
        specified in persistenceForgets.
        """
        self.__dict__.update(oldAttributes)

    def touch(self):
        """
        Mark this object for persistence soon.

        This must be run in the context of StorQ.transact
        """
        sctx = context.get('StorageContext')
        if sctx is None:
            raise Inconsistency("You're touching an object outside a transaction.")
        ikey = tuple(self.getItemPath())
        sctx[ikey] = self

    def iterpools(self):
        if not self.pools:
            return
        for sid in self.pools.keys():
            yield self.store.getItemByID(sid)

    def change(self, notificationData=None):
        """Send out a notification that something about this item has changed.
        This will notify observers on the pools which this item inhabits.

        @param notificationData: an object that represents the change.  XXX
        TODO: donovan, what kind of change object do you want here?
        """
        for value in self.iterpools():
            value._announceItem(self, "itemChanged", notificationData)

    def lock(self):
        """Prepare this object for database update.
        """
        self.touch()
        self.store.lockItem(self)

    def setIndexKey(self, indexName, newKey, notify=False):
        self.touch()
        if self.pools:
            for pool in self.iterpools():
                oldKey = self.getIndexKey(pool, indexName)
                changedKey = self._changeIndexKey(pool, indexName, newKey)
                pool.changeKeyForItem(self, indexName, oldKey, changedKey, notify=notify)
        self._setIndexKeyData(indexName, newKey)

    def appendToIndex(self, indexName, indexKey):
        """Be careful with this method.  The store will not automatically
        remove state added using this method from a pool.  removeFromIndex,
        similarly.
        """
        for pool in self.iterpools():
            pool.changeKeyForItem(self, indexName, None, indexKey)

    def removeFromIndex(self, indexName, indexKey):
        for pool in self.iterpools():
            pool.changeKeyForItem(self, indexName, indexKey, None)


    def getIndexKey(self, pool, indexName, default=None):
        """Get a key value for a particular index.  The key value must match
        what the index expects.
        """
        m = getattr(self, "index_" + indexName, None)
        if m:
            return m(pool)
        if self._indexes:
            return self._indexes.get(indexName, default)
        return default

    _indexes = None

    def _setIndexKeyData(self, indexName, newValue):
        m = getattr(self, "setindex_" + indexName, None)
        if m:
            return m(newValue)
        if not self._indexes:
            self._indexes = {}
        self._indexes[indexName] = newValue

    def _changeIndexKey(self, pool, indexName, newValue):
        m = getattr(self, "changeindex_" + indexName, None)
        if m:
            return m(pool, newValue)
        return newValue

    pools = None

    def _initpools(self):
        self.lock()
        self.pools = {}                 # {int poolID:int UID}

    def _addedToPool(self, pool):
        if pool.store != self.store:
            raise Inconsistency(
                "You can't add foreign items directly to pools.")
        if not self.pools:
            self._initpools()
        if pool.storeID not in self.pools:
            self.touch()
            uid = pool.nextUID()
            self.pools[pool.storeID] = uid
        # else: already in pool?  what to do?

    def _removedFromPool(self, pool):
        if pool.store != self.store:
            raise Inconsistency(
                "You can't add foreign items directly to pools.")
        self.touch()
        del self.pools[pool.storeID]

    def index_uid(self, pool):
        return self.pools.get(pool.storeID)

    def setindex_uid(self, newuid):
        raise NotImplementedError("You can't change an item's UID.")

    def index_received(self, pool):
        ### TODO: Change the name of this index, to make a uniform 'date created' index
        return self.dateCreated

    def newCacheDir(self, dirname):
        cur = self.store
        while cur.store is not None:
            cur = cur.store
        thisPath = os.path.join(cur.filedir, str(self.store.storeID), 'cache', str(self.storeID), dirname)
        md(thisPath)
        return thisPath


class IComponentForwarder(Interface):
    """A forwarder which can be usd with SubStore's implementation of
    getComponent.

    The use-case for this class is to provide a way for provisioned powerups to
    return a 'I'm not implemented' component in cases where the provisioning
    has run out.
    """
    def forwardComponent(self, interface, default):
        """Return a proxy for myself which may vary in certain situations.
        """


class SubStore(Item, Componentized):
    def __init__(self, *a, **b):
        Componentized.__init__(self)
        Item.__init__(self, *a, **b)
        self.idcounter = 0
        p = Pool(self, name='Root Pool')
        p.addIndex(NameIndex())
        p.suggestIndex('name', False)
        self.rootPoolRef = p.referenceTo()

    def delete(self):
        raise NotImplementedError("Crap what the hell :(")

    def getRootPool(self):
        """Return the root pool for this store.
        """
        return self.rootPoolRef.getItem()

    def __getstate__(self, dict=None):
        Item.__statecheck__(self)
        if dict is None:
            dict = self.__dict__.copy()
        x = dict['_adapterCache'] = dict['_adapterCache'].copy()
        # kill all ephemeral adapters.  Didn't Componentized do this at one
        # point!?!?
        for k,v in x.items():
            if isinstance(v, Ephemeral):
                del x[k]
        return Forgetter.__getstate__(self, dict)

    # make sure that all appropriate interfaces touch me before making a
    # potentially destructive change!!  TODO: add special-casing for ephemerals

    def setComponent(self, interfaceClass, component):
        self.touch()
        if isinstance(component, Item):
            # if it's going to be automatic, it might as well be symmetrical
            component = ItemForwardingComponent(component.referenceTo())
        return Componentized.setComponent(self, interfaceClass, component)

    def getComponent(self, interface, *a, **k):
        component = Componentized.getComponent(self, interface, *a, **k)
        if isinstance(component, ItemForwardingComponent):
            component = component.fwdTo.getItem()
        if implements(component, IComponentForwarder):
            component = component.forwardComponent(interface)
        return component

    def addComponent(self, *a, **k):
        self.touch()
        Componentized.addComponent(self, *a, **k)

    def unsetComponent(self, *a, **k):
        self.touch()
        Componentized.unsetComponent(self, *a, **k)

    def removeComponent(self, *a, **k):
        self.touch()
        Componentized.removeComponent(self, *a, **k)

    def newFile(self):
        return self.store.newSubFileForItem(self)

    def newDirectory(self, name):
        """Create a directory in the user's "address
        space" or return the existing one
        """
        return self.store.newSubDirForItem(self, name)

    def setSubKey(self, item, key, value):
        return self.store.setSubKey(item, key, value)

    def delSubKey(self, item, key, value):
        return self.store.delSubKey(item, key, value)

    def nextID(self, allocate=True):
        return self.store.nextSubID(self, allocate)

    def nextSubID(self, item, allocate=True):
        return self.store.nextSubID(item, allocate)

    def cacheItem(self, item, path):
        return self.store.cacheItem(item, path)

    def getItemByID(self, itemID):
        return self.store.getSubItem(self, itemID)

    def transact(self, f, *a, **k):
        return self.store.transact(f, *a, **k)

    def subtransact(self, f, *a, **k):
        return self.store.subtransact(f, *a, **k)

    def transback(self, result, f, *a, **kw):
        return self.store.transback(result, f, *a, **kw)

    def batch(self, f, *a, **kw):
        return self.store.batch(f, *a, **kw)

    def lockItem(self, item):
        return self.store.lockSubItem(self, item)

class StorageService(service.Service):
    def __init__(self, store):
        self.setName(reflect.qual(self.__class__))
        self.store = store
        self.store.service = self

    def startService(self):
        service.Service.startService(self)
        def _():
            svc = service.IService(self.store, default=None)
            if svc is not None:
                return svc.startService()
        return self.store.transact(_)

    def _closeLater(self, result):
        d = defer.Deferred()
        from twisted.internet import reactor
        reactor.callLater(0, self._reallyClose, d, result)
        return d

    def _reallyClose(self, d, result):
        self.store.close()
        d.callback(result)

    def stopService(self):
        service.Service.stopService(self)
        def _():
            svc = service.IService(self.store, default=None)
            if svc is not None:
                return svc.stopService()
            return defer.succeed(None)
        return self.store.transact(_).addCallback(self._closeLater)

# subsubstore.store = substore; substore.store = realstore;
# subsubstore.getItemByPath => context.get('store') == subsubstore

ITEM               = object()
INDEX              = object()
INDEX_AND_ITEM     = object()
STORE_ID           = object()
INDEX_AND_STORE_ID = object()

class IPipe(Interface):
    def itemAdded(self, item):
        pass

    def itemRemoved(self, item):
        pass

    def itemChanged(self, item, notificationData):
        pass

class IPool(Interface):
    def suggestIndex(self, index, backwards=False):
        pass

    def suggestView(self, view):
        pass

    def addPipe(self, pipe, ephemeral=False):
        pass

    def removePipe(self, pipe, ephemeral=False):
        pass

    def queryIndex(self, iname='uid', startKey=None, endKey=None, result=ITEM,
                   backwards=False, limit=None, startItemID=None, endItemID=None):
        pass

    def getSequence(self):
        pass

    def getIndexItem(self, indexName, indexKey=None, default=object()):
        pass

    def nextUID(self, allocate=True):
        pass

    def __contains__(self, item):
        pass

    def changeKeyForItem(self, item, ipref, oldKey, newKey, notify=False):
        pass

    def addItem(self, item):
        pass

    def removeItem(self, item):
        pass

    def addIndex(self, index, name=None):
        pass

    def removeIndex(self, prefix):
        pass


def key2str(whatKey, whatItemID, towardsKey, pfx, obj):
    if whatKey is not None:
        whatStr = pfx + obj.keyToSortableString(whatKey)
        if whatItemID:
            whatStr += SEP + ('%0.10d' % whatItemID)
        elif towardsKey:
            whatStr += MAGIC
    elif towardsKey:
        whatStr = pfx + END
    else:
        whatStr = pfx
    return whatStr


class PoolQuery:
    indexObj = None

    def __init__(self, pool, iname='uid', startKey=None, endKey=None, result=ITEM,
                 backwards=False, limit=None, startItemID=None, endItemID=None):
        """Retrieve information about the contents of this pool.

        An iterator which yields the requested information about the requested
        Items is returned.  This generator I{must} be iterated in a storage
        context.

        @type iname: C{str}
        @param iname: The index to query.

        @type endKey: A type appropriate for the specified index.
        @param endKey: If specified, the last key at which to return a result.

        @param result: If specified as C{ITEM}, return the L{Item} itself
        for each content of the pool returned.  If specified as C{INDEX},
        return the key by which each item is indexed in this pool.  If
        specified as C{STORE_ID}, return the store ID for each content of
        the pool returned.

        @raise NoSuchIndex: If the specified index does not exist for this pool.
        """

        self.pool = pool
        self.iname = iname
        self.startKey = startKey
        self.endKey = endKey
        self.result = result
        self.backwards = backwards
        self.limit = limit
        self.startItemID = startItemID
        self.endItemID = endItemID
        self.indexPfx = pool._getSuperStore()._formatIDPath(pool.getItemPath()) + SEP + 'index' + SEP + iname + SEP
        self.count = 1
        self.cursorMethod = None

    def __str__(self):
        return '<PoolQuery over %(pool)s/%(iname)s [%(startKey)s:%(endKey)s]>' % vars(self)

    def __repr__(self):
        return str(self)

    def _nextKey(self):
        if not self.cursorMethod:
            self._init()
        return self.cursorMethod()

    def _init(self):
        try:
            self.indexObj = self.pool.indexes[self.iname]
        except KeyError:
            raise NoSuchIndex("Pool %r has no index %r" % (self.pool.name, self.iname))
        self.resume(False)

    def resume(self, adjust=True):
        """Resume the iterator, returning the most-recently-returned value.

        If 'adjust' is False, that value will be returned again by the next
        call to 'next()'.
        """
        if self.indexObj is None:
            return
        self.startStr = key2str(self.startKey, self.startItemID, self.backwards, self.indexPfx, self.indexObj)
        self.endStr = key2str(self.endKey, self.endItemID, not self.backwards, self.indexPfx, self.indexObj)
        c = context.get("CursorFactory").cursor()
        dummy = c.set_range(self.startStr)
        if self.backwards:
            self.oper = operator.lt
            self.cursorMethod = c.prev
            if dummy[0] == self.startStr:
                c.next(dlen=0,doff=0)
        else:
            self.oper = operator.gt
            self.cursorMethod = c.next
            c.prev(dlen=0, doff=0)

        if self.oper(self.startStr, self.endStr):
            raise Inconsistency("%squery starting at %s ending at %s will have no items: specify them in order!." %
                                (self.backwards and 'backwards ' or '', self.startStr.split(SEP), self.endStr.split(SEP)))

        if adjust:
            return self.next()

    def __iter__(self):
        return self

    def stop(self):
        self.next = _stopped
        raise StopIteration()

    def next(self):
        xvalue = self._nextKey()
        if xvalue is None:
            self.stop()
        skey, svalue = xvalue

        if self.limit is not None:
            if self.count > self.limit:
                self.stop()
        self.count += 1
        if self.oper(skey, self.endStr): # I need an exit.
            self.stop()

        storeID = int(svalue)
        indexString = skey[len(self.indexPfx):].split(SEP)[0]
        indexKeyObj = self.indexObj.sortableStringToKey(indexString)
        self.startKey = indexKeyObj

        if self.result == STORE_ID:
            return storeID
        elif self.result == ITEM:
            return self.pool.store.getItemByID(storeID)
        elif self.result == INDEX:
            return indexKeyObj
        elif self.result == INDEX_AND_ITEM:
            item = self.pool.store.getItemByID(storeID)
            return (indexKeyObj, item)
        elif self.result == INDEX_AND_STORE_ID:
            return (indexKeyObj, storeID)
        else:
            raise Inconsistency(
                "Invalid result-mode specified for index query: %s" % (result,))

def _stopped():
    raise StopIteration()

class Pool(Item):
    __implements__ = (IPool,)

    persistenceForgets = ['sequence', 'ephemeralPipes']
    sequence = None

    # These attributes are for use by UI.  I won't change the default sort
    # order because the semantics of queryIndex make sense as they are.

    suggestedIndex = 'uid'
    suggestedBackwards = True
    suggestedView = 'Table'

    ephemeralPipes = ()

    persistenceVersion = 2

    def upgradeToVersion2(self):
        self.touch()
        self.indexRemover = {}

    def upgradeToVersion1(self):
        self.touch()
        self.indexCounter = {}
        for iname in self.indexes:
            self.indexCounter[iname] = 0
        from atop import upgrade
        upgrade.recountIndexes(self)

    def __init__(self, store, name=None):
        self.indexes = {}
        self.indexCounter = {}
        self.indexRemover = {}
        self.pipes = []
        self.activate()
        Item.__init__(self,store,name)
        self.addIndex(UIDIndex())

    def __contains__(self, item):
        if item.pools:
            return self.storeID in item.pools
        return False

    def __str__(self):
        kname = self.__class__.__name__
        indexes = ' '.join(self.indexes.keys())
        return '<%s %s (Indexes: %s)>' % (kname, self.name, indexes)

    def suggestIndex(self, index, backwards=False):
        self.suggestedIndex = index
        self.suggestedBackwards = backwards

    def suggestView(self, view):
        self.suggestedView = view

    def addPipe(self, pipe, ephemeral=False):
        """Add a 'pipe', an object that is notified of items arriving in this
        pool, leaving it, and changing while inside it.

        The 'ephemeral' flag is intended for use by UI elements and other
        objects which do not reside in the database.  You will need to keep a
        strong reference to the pool around somewhere if you want ephemeral
        notifications.  (Normally this will not be a big deal because you've
        already got the pool open for reading, writing, or both, somewhere else
        anyway.)

        DO NOT USE THE EPHEMERAL FLAG FOR OBJECTS INSIDE THE DATABASE.  Failure
        of a non-ephemeral observer will _not_ abort the transaction, and will
        therefore put your database into an inconsistent state if you use it
        incorrectly.
        """
        assert components.implements(pipe, IPipe)
        if isinstance(pipe, Item):
            pipe = pipe.referenceTo()
        if ephemeral:
            L = self.ephemeralPipes
        else:
            self.touch()
            L = self.pipes
        L.append(pipe)

    def removePipe(self, pipe, ephemeral=False):
        """See addPipe.

        Note that pipes added with the ephemeral flag _must_ be removed with
        the ephemeral flag.
        """
        if isinstance(pipe, Item):
            pipe = pipe.referenceTo()
        if ephemeral:
            L = self.ephemeralPipes
        else:
            self.touch()
            L = self.pipes
        L.remove(pipe)

    def activate(self):
        self.ephemeralPipes = []

    def _getSuperStore(self):
        ss = self.store
        while ss.store is not None:
            ss = ss.store
        return ss

    def queryIndex(self, iname='uid', startKey=None, endKey=None, result=ITEM,
                   backwards=False, limit=None, startItemID=None, endItemID=None):
        """Retrieve information about the contents of this pool.

        A generator which yields the requested information about the requested Items is
        returned.  This generator I{must} be iterated in a storage context.

        @type iname: C{str}
        @param iname: The index to query.

        @type endKey: A type appropriate for the specified index.
        @param endKey: If specified, the last key at which to return a result.

        @param result: If specified as C{ITEM}, return the L{Item} itself
        for each content of the pool returned.  If specified as C{INDEX},
        return the key by which each item is indexed in this pool.  If
        specified as C{STORE_ID}, return the store ID for each content of
        the pool returned.

        @raise NoSuchIndex: If the specified index does not exist for this pool.
        """
        return PoolQuery(self, iname, startKey, endKey, result, backwards, limit, startItemID, endItemID)

    def getSequence(self):
        if self.sequence is None:
            self.sequence = list(self.queryIndex('uid',result=INDEX))
        return self.sequence

    _noIndexItem = object()

    def getIndexItem(self, indexName, indexKey=None, default=_noIndexItem):

        """Retrieve an item from an index.  If no indexKey is specified, I will
        return the first item in the index.

        Upon failure, (if there are no items in the index, or the specified key
        is not found), I will raise KeyError, or return the 'default' argument,
        if it is supplied.
        """
        # XXX FIXME this should NOT be auto-transacted
        def getter():
            for k, v in self.queryIndex(indexName, startKey=indexKey, result=INDEX_AND_ITEM):
                # only iterates once, haha.
                if indexKey is None:
                    return v
                if k != indexKey:
                    if default is Pool._noIndexItem:
                        raise KeyError(indexKey)
                    return default
                return v
            if default is Pool._noIndexItem:
                raise KeyError(indexKey)
            return default
        if context.get('txn') is None:
            warnings.warn("Calling getIndexItem() outside a transaction is deprecated!",
                          DeprecationWarning, stacklevel=2)
        return self.store.transact(getter)

    def nextUID(self, allocate=True):
        return self.store.nextSubID(self, allocate)

    def _manipItem(self, item, adding):
        """
        Either add or remove an Item from this Pool. This method is used by
        addItem and removeItem.
        """
        if adding:
            announceMethodName = "itemAdded"
            item._addedToPool(self)
        else:
            announceMethodName = "itemRemoved"
        for ipref, index in self.indexes.items():
            iname = index.getIndexName()
            ikey = item.getIndexKey(self, iname)
            if adding:
                oldKey = None
                newKey = ikey
            else:
                oldKey = ikey
                newKey = None
            self.changeKeyForItem(item, ipref, oldKey, newKey)
        self._updateIndexedCounts(item, adding)
        if not adding:
            item._removedFromPool(self)
        self._announceItem(item, announceMethodName)

    def changeKeyForItem(self, item, ipref, oldKey, newKey, notify=False):
        if ipref not in self.indexes:
            if ipref in self.indexRemover:
                index = self.indexRemover[ipref]
                newKey = None
            else:
                return
        else:
            index = self.indexes[ipref]
        sisid = '%0.10d' % item.storeID
        if oldKey is not None:
            # delete the old key.
            istr = index.keyToSortableString(oldKey)
            try:
                self.delDataKey(SEP+'index'+SEP+ipref+SEP+istr+SEP+sisid, sisid)
            except DBNotFoundError:
                pass
        if newKey is not None:
            # add the new key.
            istr = index.keyToSortableString(newKey)
            self.setDataKey(SEP+'index'+SEP+ipref+SEP+istr+SEP+sisid, sisid)

    def _updateIndexedCounts(self, item, adding):
        delta = adding and 1 or -1
        for iname in self.indexes:
            if item.getIndexKey(self, iname) is not None:
                self.touch()
                self.indexCounter[iname] += delta

    def _announceItem(self, item, announceMethodName, *args):
        for pipe in self.pipes:
            getattr(pipe, announceMethodName)(item, *args)
        for epipe in self.ephemeralPipes:
            # UI elements are non-transactional anyway, so we will not honor
            # their failures.  DO NOT ADD EPHEMERAL PIPES FOR DATABASE OBJECTS.
            try:
                getattr(epipe, announceMethodName)(item, *args)
            except:
                log.err()

    def addItem(self, item):
        if item in self:
            import warnings
            warnings.warn("Adding %s to %s, but it is already there." % (item, self),
                          Warning, stacklevel=2)
            return
        self._manipItem(item, True)

    def removeItem(self, item):
        self._manipItem(item, False)

    def addIndex(self, index, name=None):
        self.lock()
        if name is None:
            name = index.getIndexName()
        self.indexes[name] = index
        self.indexCounter[name] = 0

    def removeIndex(self, prefix):
        self.lock()
        self.indexRemover[prefix] = self.indexes.pop(prefix)
        from atop.upgrade import DelItemIndex, UpgradeQueue
        return UpgradeQueue(self.store, self, DelItemIndex(prefix)).start().addCallback(
            _indexRemoveDone, self, prefix
            )

def _indexRemoveDone(result, pool, prefix):
    del pool.indexCounter[prefix]
    del pool.indexRemover[prefix]
    return result

class HardPoint(Item):
    """
    I almost called this a PoolPipe since
    it is part Pool and part Pipe. It is a pipe
    that can be attached to one pool (via setPoint()).
    Other pipes can then be added to this faux pool.
    This allows pipes to be insulated from the source
    being moved.
    """
    __implements__ = IPipe
    persistenceForgets = ['ephemeralPipes']

    def __init__(self, store, name=None):
        Item.__init__(self,store,name)
        self.pipes = []
        self.activate()
        self.sourcePoolRef = None
        self.sourcePipe = None

    def setPoint(self, pool):
        if self.sourcePoolRef:
            p = self.sourcePoolRef.getItem()
            p.removePipe(self.fwd(IPipe))
        self.sourcePoolRef = pool.referenceTo()
        pool.addPipe(self.fwd(IPipe))
        self.touch()

    def addPipe(self, pipe, ephemeral=False):
        if ephemeral:
            L = self.ephemeralPipes
        else:
            self.touch()
            L = self.pipes
        L.append(pipe)

    def removePipe(self, pipe, ephemeral=False):
        if ephemeral:
            L = self.ephemeralPipes
        else:
            self.touch()
            L = self.pipes
        L.remove(pipe)

    def activate(self):
        self.ephemeralPipes = []

    def _announceItem(self, item, announceMethodName, *args):
        for pipe in self.pipes:
            getattr(pipe, announceMethodName)(item, *args)
        for epipe in self.ephemeralPipes:
            # UI elements are non-transactional anyway, so we will not honor
            # their failures.  DO NOT ADD EPHEMERAL PIPES FOR DATABASE OBJECTS.
            try:
                getattr(epipe, announceMethodName)(item, *args)
            except:
                log.err()

    def itemAdded(self, item):
        self._announceItem(item, 'itemAdded')

    def itemRemoved(self, item):
        self._announceItem(item, 'itemRemoved')

    def itemChanged(self, item, notificationData):
        self._announceItem(item, 'itemChanged')

class IPersistable(components.Interface):

    def touch(self):
        "notify that this thingy is dirty and in need of persistament"

    def referenceTo(self):
        "produce a sturdy ref to this object"

class Persistability(Item, components.Adapter):
    __implements__ = (IPersistable,)
    def __init__(self, original, store=None):
        Item.__init__(self, store or context.get('store'))
        components.Adapter.__init__(self, original)

class StoredAdapter(object, components.Adapter):
    __metaclass__ = MetaItem

    def __init__(self, original):
        object.__init__(self)
        components.Adapter.__init__(self, original)
