# -*- test-case-name: atop.test.test_tpython -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.
# objects that should be in tpython

import errno
import time
import sys
import os

from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet.protocol import Protocol
from twisted.python import reflect, context, log

from twisted.persisted.styles import Versioned

def mkgmtime(gmtt):
    'convert GMT time-tuple to local time'
    if time.daylight and gmtt[-1]:
        zone = time.altzone
    else:
        zone = time.timezone
    return time.mktime(gmtt) - zone

def gmtForYearMonthDay(year, month, day):
    return time.gmtime(                          # 3 to a gmt tuple...
        time.mktime(                             # 2 to a localtime float...
            (year, month, day, 0, 0, 0, 0, 0, 0) # 1 go from a localtime tuple...
        )
    )

class Forgetter:
    """Utility class for forgetting particular attributes when persisted.

    @cvar persistenceForgets: A list of strings, naming attributes which should
    be deleted when I am persisted.

    @cvar contextRemembers: A list of 2-tuples of strings, (contextName,
    attributeName), that name attributes which should be retrieved from
    twisted.python.context upon loading me.
    """

    def __getstate__(self, state=None):
        if state is None:
            state = self.__dict__.copy()
        if isinstance(self, Versioned):
            d = Versioned.__getstate__(self, state)
        else:
            d = state
        l = []
        reflect.accumulateClassList(self.__class__, 'persistenceForgets', l)
        for ig in l:
            if d.has_key(ig):
                del d[ig]
        return d

    def __setstate__(self, state):
        if isinstance(self, Versioned):
            Versioned.__setstate__(self, state)
        else:
            self.__dict__ = state
        l = []
        reflect.accumulateClassList(self.__class__, 'contextRemembers', l)
        for contextName, attributeName in l:
            setattr(self, attributeName, context.get(contextName))


class DebugShimTransport(Protocol):
    counter = 0

    def __init__(self, p, f=sys.stdout, lengthLimit=None):
        self.p = p
        self.f = f
        DebugShimTransport.counter += 1
        self.counter = DebugShimTransport.counter
        self.lengthLimit = lengthLimit

    def _mungeData(self, data):
        if self.lengthLimit and len(data) > self.lengthLimit:
            data = data[:self.lengthLimit - 12] + '<... elided>'
        return data

    # Protocol
    def makeConnection(self, t):
        self.origWrite = t.write
        t.write = self.write
        self.origWriteSeq = t.writeSequence
        t.writeSequence = self.writeSequence
        self.origLoseConn = t.loseConnection
        t.loseConnection = self.loseConnection
        self.t = t
        self.f.write('*\n')
        self.f.flush()
        self.p.makeConnection(t)

    def dataReceived(self, data):
        self.f.write( 'C %d: %r\n' % (self.counter, self._mungeData(data)))
        self.p.dataReceived(data)
        self.f.flush()

    def connectionLost(self, reason):
        self.f.write( 'C %d : %r\n' % (self.counter, reason) )
        self.f.flush()
        self.p.connectionLost(reason)

    def factory():
        def get(self):
            return self.p.factory
        def set(self, value):
            self.p.factory = value
        def delete(self, value):
            del self.p.factory
        return get, set, delete
    factory = property(*factory())

    # Transport

    def getHost(self):
        return self.t.getHost()

    def getPeer(self):
        return self.t.getPeer()

    def write(self, data):
        self.f.write( 'S %d: %r\n' % (self.counter, self._mungeData(data)))
        self.f.flush()
        self.origWrite(data)

    def writeSequence(self, iovec):
        self.f.write( 'S %d: %r\n' % (self.counter, iovec) )
        self.f.flush()
        self.origWriteSeq(iovec)

    def getPeer(self):
        return self.t.getPeer()

    disconnecting = False

    def loseConnection(self):
        self.f.write( 'S %d: *\n' % self.counter)
        self.f.flush()
        self.origLoseConn()
        self.disconnecting = True

    def registerProducer(self, producer, streaming):
        self.t.registerProducer(producer, streaming)

    def unregisterProducer(self):
        self.t.unregisterProducer()

def iterateInReactor(i, delay=0.0001, holder=[None]):
    d = defer.Deferred()
    def callLaterer():
        try:
            maybe = i.next()
            if isinstance(maybe, defer.Deferred):
                def rslts(x):
                    #print 'rslts', x
                    holder[0] = reactor.callLater(delay, callLaterer)
                    return x
                maybe.addBoth(rslts)
                return
        except StopIteration:
            d.callback(i)
        except:
            d.errback()
        else:
            holder[0] = reactor.callLater(delay, callLaterer)
    callLaterer()
    return d


PERIODS = (
    ('week', 60 * 60 * 24 * 7),
    ('day', 60 * 60 * 24),
    ('hour', 60 * 60),
    ('minute', 60),
    ('second', 1),
)

def humanInterval(seconds):
    s = []
    for (p, d) in PERIODS:
        n = int(seconds / d)
        seconds %= d
        if n:
            s.append('%d %s%s' % (n, p, n > 1 and 's' or ''))
    if s:
        return ', '.join(s)
    return '0 seconds'


class _counter:
    persec = 0
    def accum(self):
        self.persec += 1
        reactor.callLater(0, self.accum)
    def display(self):
        log.msg("FPS: %d" % self.persec)
        self.persec = 0
        reactor.callLater(1, self.display)

def framecounter():
    c = _counter()
    c.display()
    c.accum()

from atop.filepile import readlink, symlink, rmlink
class FilesystemLock:
    """A mutex.

    This relies on the filesystem property that creating
    a symlink is an atomic operation and that it will
    fail if the symlink already exists.  Deleting the
    symlink will release the lock.

    @ivar name: The name of the file associated with this lock.
    @ivar clean: Indicates whether this lock was released cleanly by its
    last owner.  Only meaningful after C{lock} has been called and returns
    True.
    """

    clean = None
    locked = False

    def __init__(self, name):
        self.name = name

    def lock(self):
        """Acquire this lock.

        @rtype: C{bool}
        @return: True if the lock is acquired, false otherwise.

        @raise: Any exception os.symlink() may raise, other than
        EEXIST.
        """
        try:
            pid = readlink(self.name)
        except (OSError, IOError), e:
            if e.errno != errno.ENOENT:
                raise
            self.clean = True
        else:
            if not hasattr(os, 'kill'):
                return False
            try:
                os.kill(int(pid), 0)
            except (OSError, IOError), e:
                if e.errno != errno.ESRCH:
                    raise
                rmlink(self.name)
                self.clean = False
            else:
                return False

        symlink(str(os.getpid()), self.name)
        self.locked = True
        return True

    def unlock(self):
        """Release this lock.

        This deletes the directory with the given name.

        @raise: Any exception os.readlink() may raise, or
        ValueError if the lock is not owned by this process.
        """
        pid = readlink(self.name)
        if int(pid) != os.getpid():
            raise ValueError("Lock %r not owned by this process" % (self.name,))
        rmlink(self.name)
        self.locked = False

import sys
import shutil
import hotshot
import profile
import marshal
import tempfile

from twisted.python import components

class IProfiler(components.Interface):
    def start(self):
        """Begin recording profiling information.
        """

    def stop(self):
        """Stop recording profiling information.
        """

    def save(self, fileObj):
        """Save recorded profiling information to the given file.
        """

from twisted.application import service

class HotshotProfilerService(service.Service):
    __implements__ = service.Service.__implements__, IProfiler

    running = False

    def __init__(self, filename=None):
        self.setName("atop.tpython.HotshotProfilerService")
        if filename is None:
            foo, self.fn = tempfile.mkstemp('hotshot')
        else:
            self.fn = filename

    def start(self):
        self.profile = hotshot.Profile(self.fn)
        self.profile.start()
        self.running = True

    def stop(self):
        self.profile.stop()
        self.running = False

    def save(self, fileObj):
        shutil.copyfileobj(file(self.fn, 'rb'), fileObj)
