# -*- test-case-name: quotient.test.test_prefup -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from twisted.python.components import Interface, CannotAdapt

from atop.powerup import Powerup, IPowerStation
from atop.store import Item, Pool, ItemReference

from quotient.searchup import ISearch
from quotient.prefup import IPrefs, Prefup
from quotient.proto.smtpout import IOutgoingMail
from quotient.blogup import IBlog
from quotient.cleanup import IClean
from quotient.application import IDivmodApp, IMailApp, IVoiceApp, IChatApp
from quotient.application import IPeopleApp, IAlbumApp, ICalendarApp, ITodoApp
from quotient.application import  IFindApp, IAdminApp, ISettingsApp
from quotient.smarts.image import IThumbnail
from quotient.addressbook import IAddressBook

# TODO - Get rid of these 2 classes asap
# They are vestigial and needed for upgrading. Once upgraded
# it shoudl be safe to remove them
class IApplication(Interface):
    pass
class Application(Item):
    pass

class IConfigure(Interface):
    pass

class Configup(Powerup):
    __implements__ = IConfigure, Powerup.__implements__
    powerupName = 'Configure Powerups'

    dependencies = [Prefup]

    persistenceVersion = 1

    def upgradeToVersion1(self):
        self.touch()
        IPowerStation(self.store).installPowerup(Prefup)

    def setUpPools(self, avatar):
        avatar.setComponent(IConfigure, self)
        self.setupIndexing()
        self.setupPrimaryNav()

    def setupIndexing(self):
        srch = ISearch(self.store, None)
        if srch is None:
            return
        # Index sent mail

        for (iface, attr) in ((IOutgoingMail, 'sentRef'), (IBlog, 'blogRef'), (IClean, 'cleanPool')):
            try:
                o = iface(self.store)
            except CannotAdapt:
                pass
            else:
                p = getattr(o, attr)
                if isinstance(p, ItemReference):
                    p = p.getItem()
                srch.indexPool(p)

    def getPupsNPools(self):
        pups = list(IPowerStation(self.store).getPowerupsPool().queryIndex())
        pools = list(self.store.getRootPool().queryIndex())
        rp = self.store.getRootPool()
        rp.setIndexKey('name', 'Explore')
        pools.append(rp)
        return pups + pools

    def setupPrimaryNav(self):
        all = self.getPupsNPools()
        prefs = IPrefs(self.store)

        for appName, iface in (('Divmod', IDivmodApp), ('Mail', IMailApp),
                               ('Voice', IVoiceApp), ('Chat', IChatApp),
                               ('People', IPeopleApp), ('Album', IAlbumApp),
                               ('Calendar', ICalendarApp), ('Todo', ITodoApp),
                               ('Find', IFindApp), ('Admin', IAdminApp)):
            appFunc = getattr(self, 'setup%sApp' % appName, None)
            aspect = iface(self.store, default=None)
            if appFunc is not None and aspect is not None:
                app = appFunc(all, aspect)
                if app is not None:
                    prefs.addToPrimaryNav(app)

    def setupDivmodApp(self, all, app):
        nav = ['Summary', 'Control Panel']
        L = []
        for n in nav:
            for a in all:
                if a.name == n:
                    a.touch()
                    a.applicationRef = app.getRef()
                    L.append(a)
                    break
        app.setApplicationNav(L)
        return app

    def setupMailApp(self, all, app):
        nav = self.generateMailNav(all, app)
        app.setApplicationNav(nav)
        return app

    def generateMailNav(self, all, app):
        nav = ['Inbox', 'Compose', 'Filtered', 'Mailing Lists', 'Spam',
               'SpamBayes', 'External Accounts',]
        if all is None:
            all = self.getPupsNPools()
        all.extend(app.appRoot.queryIndex())
        L = []
        for n in nav:
            for a in all:
                if a.name == n:
                    a.touch()
                    a.applicationRef = app.getRef()
                    L.append(a)
                    break

        if len(L) < 3:
            return []

        # Replace Explore with 'Mail Folders'
        mailFolders = app.appRoot
        mailFolders.applicationRef = app.getRef()
        for a in all:
            if a.name == 'Trash':
                mailFolders.addItem(a)
                break

        L.insert(3, mailFolders)
        settings = Pool(self.store, 'Mail Settings')
        settings.suggestIndex('name', backwards=False)
        settings.applicationRef = app.getRef()
        # Add SpamBayes
        settings.addItem(L.pop())
        # Add External Accounts
        settings.addItem(L.pop())
        L.append(settings)
        return L

    def setupVoiceApp(self, all, app):
        nav = ['Phone', 'Voice and Telephony', 'Calls']
        L = []
        for n in nav:
            for a in all:
                if a.name == n:
                    a.touch()
                    a.applicationRef = app.getRef()
                    L.append(a)
                    break
        app.setApplicationNav(L)
        return app

    def setupChatApp(self, all, app):
        return app

    def setupPeopleApp(self, all, app):
        addBook = IAddressBook(self.store)
        subApps = [addBook]
        for a in list(app.appRoot.queryIndex()):
            if a.name == 'Everybody':
                subApps.append(a)
                break
        for a in subApps:
            a.touch()
            a.applicationRef = app.getRef()
        app.setApplicationNav(subApps)
        return app

    def setupAlbumApp(self, all, app):
        thumbnail = IThumbnail(self.store, default=None)
        if thumbnail is not None:
            gallery = thumbnail.extractPool()
            gallery.applicationRef = app.getRef()
            app.setApplicationNav([gallery])
        else:
            app.setApplicationNav([])
        return app

    def setupCalendarApp(self, all, app):
        return app

    def setupTodoApp(self, all, app):
        return app

    def setupFindApp(self, all, app):
        nav = ['Search', 'Extracts', 'Explore']
        L = []
        for n in nav:
            for a in all:
                if a.name == n:
                    a.touch()
                    a.applicationRef = app.getRef()
                    L.append(a)
                    break
        app.setApplicationNav(L)
        return app

    def setupAdminApp(self, all, app):
        for a in all:
            if a.name == 'Administrative Controls':
                a.touch()
                a.applicationRef = app.getRef()
                app.setApplicationNav([a])
                break
        return app

    def setupSettingsApp(self, all, app):
        for a in all:
            if a.name == 'Control Panel':
                a.touch()
                a.applicationRef = app.getRef()
                app.setApplicationNav([a])
                break
        return app
