# -*- test-case-name: quotient.test.test_webui -*-
# -@*- test-case-name: quotient.test.test_functional -*-

import os, sys, warnings
from os import mkdir, path as osp
from os.path import join as opj

from twisted.application.service import Application, IService
from twisted.runner import procutils
from twisted.trial import unittest 
from twisted.internet import defer, protocol, reactor
from twisted.python import log, util, components
from twisted.spread import pb

from quotient.functional.navtags import IDoMyThing
from quotient.functional import client

import quotient
import atop
from atop import powerup, store


# Conditional logging code --------------------

# TODO: move the debug logging code
# into __init__.py to centralize it (?)

_unittest_debug = False
_sekritHQ_debug = False
_childTimeout_debug = False

class _Observer:
    def __init__(self, k):
        self.k = k
    def emit(self, events):
        if self.k in events: print ''.join(events[self.k])

def addObservers():
    for n, k in [(_unittest_debug, 'unittest'),
                 (_sekritHQ_debug, 'sekritHQ'),
                 (_childTimeout_debug, 'ctimeout')]:
        if n: log.addObserver(_Observer(k).emit)

addObservers()

def _dbgmsg(msg):
    log.msg(unittest=msg)

# --------------------------------------------

from twisted.internet import base
base.DelayedCall.debug = True


SkipTest = unittest.SkipTest
FailTest = unittest.FailTest
deferredResult = unittest.deferredResult
deferredError = unittest.deferredError


class AuthData(object):
    """for the functional tests, represents user signup/login data"""
    def __init__(self, u='', p='', rName=''):
        def _(**kw):  # cookbook recipie 1.2! :)
            return kw
        if not rName:
            rName = u
        signin = self.signin = _(username=u, password=p)
        self.signup = _(password____2=p, realname=rName, **signin)


class AssertFailure(Exception):
    pass

class TheOneTrueBaseClass(unittest.TestCase):
    diagnostic = 'An error occurred'

    urltrailName = '_urltrail.html'
    HTTPPort = None      
    HTTPSPort = None
    skip = False

    pbroot, pbcnx, pbport = None, None, None

    def setUpClass(self):
        logPath = self.logPath = opj(os.getcwd(), 'test.log')
        print "test_log: %s" % logPath
        
        tdir = self.tdir = self.mktemp()
        os.mkdir(tdir)
        assert os.path.exists(tdir)

        # tempdir is needed later
        self.tempDir = self.mktemp()
        mkdir(self.tempDir)

        print "self.tempDir: %s\nself.tdir:%s" % (self.tempDir, self.tdir)

        client.startUrlTrail(self.tempDir, self.urltrailName)

        # XXX: do we really still need the getsFile?
        getsFile = TheOneTrueBaseClass.getsFile = osp.abspath(opj(self.tempDir, 'getsFile.log'))
        file(getsFile, 'w').write('all retrieved urls\n\n')
        log.debug("getsFile: %s" % getsFile)



class PBUtil(object):
    tu = None
    _avatarName, _avatar, _store, _searchup= None, None, None, None

    def __init__(self, avatarName=None):
        self.avatarName = avatarName

    def _reallyGetStore(self):
        import gc
        return [x for x in gc.get_referrers(atop.store.Store) if isinstance(x, atop.store.Store)][0]

    def _getStore(self):
        if not self._store:
            self._store = self._reallyGetStore()
        return self._store
    store = property(_getStore)

    def _getAvatarName(self):
        return self._avatarName

    def _setAvatarName(self, name):
        """note, this won't change self.avatar.name, it just changes what avatar we're dealing with"""
        if name != self._avatarName:
            self._avatar = atop.credup.preauth(self.store, name)    
            self._avatarName = name
    avatarName = property(_getAvatarName, _setAvatarName)

    def _getAvatar(self):
        return self._avatar

    def _setAvatar(self, a):
        if a != self._avatar:
            self._avatar = a
            self._avatarName = a.name
    avatar = property(_getAvatar, _setAvatar)

    def _getPups(self):
        for v in powerup.IPowerStation(self.avatar).powerups.itervalues():
            yield v.getItem() 

    def getPupOfClass(self, klass):
        # this is an awful hack to get around the fact there doesn't
        # appear to be an ISearchup
        for p in self._getPups():
            if isinstance(p, klass):
                return p
        return None

    def _getSearchup(self):
        if not self._searchup:
            from quotient.searchup import Searchup
            self._searchup = self.getPupOfClass(Searchup)
        return self._searchup
    searchup = property(_getSearchup)

    def indexingTestSettings(self):
        from quotient.sysup import IControlPanel
        def _():
            cp = IControlPanel(self.store)
            cp.allowAction(IControlPanel.MESSAGE_INDEXING, True)
            return cp
        cp = self.store.transact(_)
        self.searchup.indexer.setMergeFactor(2)

        # the key name should be 'enable-indexing'
        return cp.actions['disable-indexing'] == True 


class SekritAgent(pb.Root):
    """our man on the *inside*"""
    tbs = None
    unsafeTracebacks = 1
    
    def observe(self, events):
        if events['message']:
            return
        if events['isError'] and 'failure' in events:
            if not self.tbs: self.tbs = []
            self.tbs.append(pb.failure2Copyable(events['failure'], unsafeTracebacks=1))

    def remote_startObserving(self):
        log.addObserver(self.observe)
        return "started observing"

    def remote_stopObserving(self):
        log.removeObserver(self.observe)
        return "stopped observing"

    def remote_allYourTracebacks(self):
        tbs, self.tbs = self.tbs, None
        return tbs
    
    def remote_makeAdmin(self, username):
        log.debug('remote_makeAdmin')
        from quotient import sysup
        u = PBUtil(username)
        def _():
            return atop.powerup.IPowerStation(u.avatar).installPowerup(sysup.Adminup)
        p = u.store.transact(_)
        log.msg('added %s to %s' % (p, username))
        return "Success"

    def remote_loadMailDir(self, username, path):
        log.debug('remote_loadMailDir')
        util = PBUtil(username)
        from quotient.scripts.usercreate import importDirectories
        importDirectories(util.avatar, {'directories': [path]})
        return "Success"

    def remote_getAvatarStoreID(self, username):
        log.debug('remote_getAvatarStoreID')
        u = PBUtil(username)
        return u.avatar.storeID

    def remote_indexingTestSettings(self, username):
        log.debug('remote_allowIndexing')
        u = PBUtil(username)
        return u.indexingTestSettings()

    def remote_waitForIndexing(self, username):
        log.debug('remote_waitForIndexing')
        u = PBUtil(username)
        from quotient.searchup import Searchup
        sup = u.getPupOfClass(Searchup)
        return sup.finishedIndexing()

    def remote_forceIndexerFlush(self, username):
        u = PBUtil(username)
        u.searchup.indexer.flush()

    def remote_indexerSanityCheck(self, username, query):
        u = PBUtil(username)
        from lupy.search import hits

        d = u.searchup.search(query)
        d.addCallback(lambda result: isinstance(result, hits.Hits))
        return d
        
    def remote_getSearchupStoreID(self, username):
        u = PBUtil(username)
        return u.searchup.storeID
        
    def remote_createUser(self, username, password, domain, ticketed, admin):
        from quotient.scripts import usercreate
        from quotient.test import qmanhole

        def ezdict(**kw):
            return kw
        opts = ezdict(username=username,
                      password=password, domain=domain,
                      ticketed=ticketed, admin=admin)

        st = qmanhole.getStore()
        st.transact(usercreate.createUser, st, opts)        


class SekritHQError(Exception):
    pass

class SekritHQ(object):
    """a local interface to the SekritAgent, handles calls in a synchronous fashion"""
    observing = False
    pbcnx = None
    
    def __init__(self, pbport):
        self.pbport = pbport
        # XXX: this is kind of sucky, but effective for right now
        tie = 'twisted.internet.error'
        self.acceptableFailures = ['%s.TimeoutError' % tie, '%s.ConnectError' % tie]  

    def _dbgmsg(self, msg):
        log.msg(sekritHQ=msg)
    
    def connect(self):
        factory = pb.PBClientFactory()
        factory.unsafeTracebacks = 1
        self.pbcnx = reactor.connectTCP("localhost", self.pbport, factory)
        done = defer.Deferred()
            
        def _gotRoot(root):
            self.pbroot = root
            return root

        d = factory.getRootObject()
        d.addCallbacks(_gotRoot, self._pbFailure)
        deferredResult(d)

    def disconnect(self):
        self._dbgmsg('disconnecting')
        if self.observing:
            self.stopObserving()
        self.pbcnx.disconnect()
        while self.pbcnx.state != 'disconnected':
            reactor.iterate(0)

    def pbCmd(self, *a):
        """a convenience function to run remote_ methods on the SekritAgent"""
        # XXX: Should probably add some kind of timeout code here
        d = self.pbroot.callRemote(*a).addErrback(self._pbFailure)
        return deferredResult(d)

    def _pbFailure(self, error):
        # XXX: i'm not quite sure what the proper way to handle errors is

        print "pbCmd got error %s" % error
        from twisted import internet
        error.trap(pb.PBConnectionLost,
                   internet.error.ConnectionRefusedError,
                   internet.error.ConnectionLost)
        self.pbcnx.disconnect() # XXX: this should probably try to reconnect somehow
        # XXX: if not a connection failure this should re-raise the exception
        return None
       
    def checkForTracebacks(self):
        tracebacks = self.pbCmd('allYourTracebacks')
        if tracebacks:
            for tb in tracebacks:
                if tb.type not in self.acceptableFailures:
                    from cStringIO import StringIO
                    for tb in tracebacks:
                        sio = StringIO()
                        tb.printTraceback(sio)
                        msg = 'child process had a traceback:\n\n%s' % sio.getvalue()
                        self._dbgmsg('\nchecked for tracebacks: and found one')
                        print msg
                        raise FailTest, msg
            
    def startObserving(self):
        status = self.pbCmd('startObserving')
        if status != 'started observing':
            raise warnings.warn("SekritAgent is not watching for tracebacks!") # XXX: this is probably suck
            return
        self.observing = True
        self._dbgmsg('traceback observation status: %s' % status)
        return status

    def stopObserving(self):
        status = self.pbCmd('stopObserving')
        if status != 'stopped observing':
            print "stopObserving did not get expected status response"
        self.observing = False
        self._dbgmsg('traceback observation status: %s' % status)
        return status
        
    # XXX: This has pretty awful semantics, but is solid in concept. i need
    # to revisit this when i get the chance and make it a little neater
    def _resultUtil(self, cmd, result, expected=None): 
        """trivial method to handle result checking and reporting
        if expected = None (the default value) then the result is just debug logged and returned"""
        self._dbgmsg('%s returned %s' % (cmd, result))
        if expected and result != expected:
            raise SekritHQError, "SekritHQ.%s failed, expected %s got %s" % (cmd, expected, reesult)
        return result

    def makeAdmin(self, user):
        cmd, expect = 'makeAdmin', 'Success'
        self._dbgmsg('makeAdmin, creating admin user: %s' % user)
        return self._resultUtil(cmd, self.pbCmd(cmd,user), expect)
        
    def loadMailDir(self, user, path):
        cmd, expect = 'loadMailDir', 'Success'
        return self._resultUtil(cmd, self.pbCmd(cmd, user, path), expect)
        
    def getAvatarStoreID(self, user):
        cmd = 'getAvatarStoreID'
        result = self.pbCmd('getAvatarStoreID', user)
        if not result: 
            raise SekritHQError, "getAvatarStoreID failed"
        return self._resultUtil(cmd, result)

    def indexingTestSettings(self, user):
        cmd = 'indexingTestSettings'
        result = self.pbCmd(cmd, user)
        if not result: 
            raise SekritHQError, "indexingTestSettings failed"
        return self._resultUtil(cmd, result)
        
    def waitForIndexing(self, user):
        cmd = 'waitForIndexing'
        return self._resultUtil(cmd, self.pbCmd(cmd, user))

    def forceIndexerFlush(self, user):
        cmd = 'forceIndexerFlush'
        return self._resultUtil(cmd, self.pbCmd(cmd, user))

    def indexerSanityCheck(self, user, query):
        cmd = 'indexerSanityCheck'
        return self._resultUtil(cmd, self.pbCmd(cmd, user, query), True)
        
    def getSearchupStoreID(self, user):
        cmd = 'getSearchupStoreID'
        result = self.pbCmd(cmd, user)
        if not result:
            raise SekritHQError, "getSearchupStoreID failed to return a value"
        return self._resultUtil(cmd, result)


class ChildProcessProtocol(protocol.ProcessProtocol):
    delayed = None
    _chldDbg = lambda self, msg: log.msg(childTimeout=msg)
    
    def __init__(self, doneLoading):
        self.doneLoading = doneLoading 
        self.errBuff = []
        self._processExited = defer.Deferred()

    def outReceived(self, data):
        m = 'child process outRecieved: %s' % data
        log.msg(m)
        if 'set uid/gid' in data:
            d, self.doneLoading = self.doneLoading, None
            if d:
                d.callback("twistd ready")
        elif 'Server shut down' in data:
            d, self.shutDown = self.shutDown, None
            d.callback('server shut down')

    def errReceived(self, data):
        # something using twisted.internet.app is screwing this up
        m = 'child process errRecieved: %s' % data
        log.msg(m)
        if 'DeprecationWarning' in data:
            return
        self.errBuff.append(data)

    def getErrBuf(self):
        buff, self.errBuff = self.errBuff, []
        return ''.join(buff)

    def processEnded(self, status):
        log.msg(ctimeout='processEnded called with status: %s, self._processExited: %s' % (status, self._processExited))
        if self._processExited:
            log.msg(ctimeout='callback self._processExited')
            self._processExited.callback(status)

    def shutdownOrKillYourself(self):
        # first, we'll try a SIGTERM
        import signal
        log.msg(ctimeout='shutdownOrKillYourself called')
        pid = self.transport.pid
        log.msg(ctimeout='self.transport.pid: %s' % pid)

        def _timeout():
            # and if that fails we STAB!
            log.msg(ctimeout='ChildProcessProtocol.shutdownOrKillYourself._timeout')
            self.transport.signalProcess('KILL')
            
        log.msg(ctimeout='setting up timeout call')
        _delayed = reactor.callLater(0, _timeout)

        def _err(fail):
            # 'KILL' will cause an error to be raised, but that's okay because we
            # want it DEAD DEAD DEAD!
            log.msg(ctimeout='ChildProcessProtocol.shutdownOrKillYourself._err callled')
            from twisted.internet.error import ProcessTerminated
            fail.trap(ProcessTerminated)
            return None
        self._processExited.addErrback(_err)

        def _cleanup(ignore):
            log.msg(ctimeout='ChildProcessProtocol.shutdownOrKillYourself._cleanup called')
            if _delayed.active():
                log.msg(ctimeout='_delayed was active!')
                _delayed.cancel()
        self._processExited.addBoth(_cleanup)

        if pid: 
            log.msg(ctimeout="we're not already dead, kill -TERM %s" % pid)
            os.kill(pid, signal.SIGTERM)
            log.msg(ctimeout='called sigterm')
        else:
            log.msg(ctimeout="we're already dead")
            self._processExited = defer.succeed(None)

        log.msg(ctimeout='returning self._processExited')
        return self._processExited

class _WaitToStop(protocol.Factory):
    def __init__(self):
        self.deferred = defer.Deferred()
    def stopFactory(self):
        self.deferred.callback('foo')

class TestCase(TheOneTrueBaseClass):
    certificate = """
-----BEGIN RSA PRIVATE KEY-----
MIIEowIBAAKCAQEAyXXBZHuMWibUt7pvdEI6cCC5ma99ZgKo1X6NeWcwwwCuf2hQ
oNA/AYhkXbV+MqbMdLGym7AbCHFMaeHWDZ4/jz0SybU7R+9BnpvZLQg6nqb8PgJK
sb0cihlVoDw/chradMblmraPEWJIHyJ8MfO8J0sgsoAwGaa1DG3OlKs45/yrnx9P
JeWn9w5biZCCFgNcGwnf4afc0SMOJnJto+3aFtZ7Xyp+6ncjOkLZkmCax1ixA3KK
Nkzgc+xhAgq7brVBhssh4w5ORsLK3TZ80CYyGmSc/Aw0/vYp6czs0kiju/oiLtYE
B95kqPn89j74xog5vQ6KmdGmSxiqakRy1J3DCQIDAQABAoIBAQDCr1Zd6Xs/38iv
aK6udhYNJ8LUvx96xlUsRCkpykv4Fp+LJl7MmZJZw9niKTM0dbadit/kznHBXeQP
apw1+tjHbqjKXqZUJOWQBmIAv7QOf6bBsetHCk8bMdJqj+5sMYEtWsymD4gAr66s
z+GfUEBi5JpLhc916oGJ6k+5oiHTJMYvYYgIAEIkhq0XHRM3zXEzY5hfAVP1ecng
zUkVrPTQVxX80uoW4rDV5MIRno/WgT8Un3ykWcYC1bC8Qpy893XH07tWoAyJ6S3Z
L4B93aV3tnk97yE+OCoZ4PDGDVMTyHiIB6mdTagumlY7SzWBTsZSKnuRX4PjdWZj
rzimzmuVAoGBAOUSEyLVjWIuIQRjMv+QftnHJIl6EdJZXydz3MZWF4ju7cherK6t
rxtBr6PjMGC3glPKZCzWX/gvI/WmHR5Oh63ZM2Xki5NLf1xZ3H+JWAUvbmKIfRBy
Lf8E2APJt+RaQuoG7RZdELN3UuQHt4JMKrBOg2qzuIWQ80lBGFtfEUnjAoGBAOEk
u8khSgw62u0Rs4VbBzUK/i1nJY+PR4DXefYdEZnVBoyLpWcBaVhvbFxpEStXFWL5
sH9WWhuhKKfzZbrHysa0aeuTn4adrFNq0HxQqd8slPqe215UqdWKZ6CTBy93JAtU
quUy9yiLdm9VbYNeyvujZzQLnOA4DIuaUi3igQMjAoGAH4Gq7CGfPp/FnsreGtSO
/fksB1So2tu8ALso5Ia22fiy79pFugjdiswwEkSPRN44qLmvqZMaxBIHyiO84Yw/
kSPMs/kJUYsSy2r4rvn/B0eM0TSgrwBlXywVYf/+sQPRLEsjlO3b/DMvTXnaPauo
O3OxE54AhS54p9yyMu8Se60CgYBZhMuNu0i/+B7YKDOjEICbHJXX461RmXyqxAJg
U0qOSd9ivpuziyIh4ZvKWpRTETsVsYX1+5W0aq3JqcOHMq1zuIRtALNrSQId4Nio
tUH11LoftE4xAaTNnUx60jsHKm4Qe0e//jW7z0f2GJeEpoo658u/K+biWCbU0zVd
vTPi1QKBgD/05h5rdhLAltwrawTruHVh+U4NxCq5H+M1Dld8PHZIb+zwDMjMOlP9
JkHb2LK1ZvMe0h3EubncS60KwrybUUXqyRw4j1OW0iDoiUcWvfJr1pW8o1j52jm5
qeyQl38E+0HOvYxa4pTRkx/Kp2zVPW8ckTyHE5dueu+KDcJj+QQI
-----END RSA PRIVATE KEY-----
-----BEGIN CERTIFICATE-----
MIIEhTCCA22gAwIBAgIBADANBgkqhkiG9w0BAQQFADCBjTELMAkGA1UEBhMCVVMx
ETAPBgNVBAgTCE5ldyBZb3JrMREwDwYDVQQHEwhOZXcgWW9yazEPMA0GA1UEChMG
RGl2bW9kMREwDwYDVQQLEwhTZWN1cml0eTETMBEGA1UEAxMKZGl2bW9kLmNvbTEf
MB0GCSqGSIb3DQEJARYQYWRtaW5AZGl2bW9kLmNvbTAeFw0wNDAxMTcwMDA1MjVa
Fw0wNTAxMTYwMDA1MjVaMIGNMQswCQYDVQQGEwJVUzERMA8GA1UECBMITmV3IFlv
cmsxETAPBgNVBAcTCE5ldyBZb3JrMQ8wDQYDVQQKEwZEaXZtb2QxETAPBgNVBAsT
CFNlY3VyaXR5MRMwEQYDVQQDEwpkaXZtb2QuY29tMR8wHQYJKoZIhvcNAQkBFhBh
ZG1pbkBkaXZtb2QuY29tMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA
yXXBZHuMWibUt7pvdEI6cCC5ma99ZgKo1X6NeWcwwwCuf2hQoNA/AYhkXbV+MqbM
dLGym7AbCHFMaeHWDZ4/jz0SybU7R+9BnpvZLQg6nqb8PgJKsb0cihlVoDw/chra
dMblmraPEWJIHyJ8MfO8J0sgsoAwGaa1DG3OlKs45/yrnx9PJeWn9w5biZCCFgNc
Gwnf4afc0SMOJnJto+3aFtZ7Xyp+6ncjOkLZkmCax1ixA3KKNkzgc+xhAgq7brVB
hssh4w5ORsLK3TZ80CYyGmSc/Aw0/vYp6czs0kiju/oiLtYEB95kqPn89j74xog5
vQ6KmdGmSxiqakRy1J3DCQIDAQABo4HtMIHqMB0GA1UdDgQWBBQo3AN1b41OShqw
5rrozZYg1iTm/zCBugYDVR0jBIGyMIGvgBQo3AN1b41OShqw5rrozZYg1iTm/6GB
k6SBkDCBjTELMAkGA1UEBhMCVVMxETAPBgNVBAgTCE5ldyBZb3JrMREwDwYDVQQH
EwhOZXcgWW9yazEPMA0GA1UEChMGRGl2bW9kMREwDwYDVQQLEwhTZWN1cml0eTET
MBEGA1UEAxMKZGl2bW9kLmNvbTEfMB0GCSqGSIb3DQEJARYQYWRtaW5AZGl2bW9k
LmNvbYIBADAMBgNVHRMEBTADAQH/MA0GCSqGSIb3DQEBBAUAA4IBAQCw+6jVUrgc
wJSqkfXWRzB2a8FqUN/P0tUbqQjs/VNyRJfwQCde1p36s6Lmv1s/SHYr9YAg44XJ
ADbLPEiMSkhRkhfCb+cuA+9594aoQZRjknLGunPZ4WyQbe/Pr7ZAnSLWbA6drJIm
gbTHbI1gdWRvS1C7qvHKAT36JqT8C6NWzwY4aQIRvMy+KPvfD2Gufv430UMwQ4sQ
cP1vBtUwQw0QPdYcajGu2tIP6L70idJGEVhwpdPb39cPxN3Rs6tMwQ4TfGBm7BVl
st7n2hlHzFVhYhwhRh04CoQDNquYOy33vgR6hRErmxMWgXeUXu7ozfh3OflnuLQj
Hl5sckHv7WhU
-----END CERTIFICATE-----
"""
    _didSetUpClass = False
    includeTelnetPBInTac, authData = None, None
    childProcessSetupTimeout = 10 

    # convenience for logging child timeout debug messages
    _chldDbg = lambda self, msg: log.msg(childTimeout=msg)
    
    def _initPorts(self):
        # heuristic, but bleh
        portlist = ['httpPort', 'httpsPort', 
                    'pop3Port', 'pop3sPort', 
                    'smtpPort', 'smtpsPort', 
                    'imap4Port','imap4sPort',
                    'sipPort', 'pbPort']
       
        prts = dict(zip(portlist, [0 for n in xrange(len(portlist))]))            

        dL = []
        def _findAPort():
            dfac = _WaitToStop()
            dL.append(dfac.deferred)
            t = reactor.listenTCP(0, dfac)
            portno = t.getHost().port
            t.stopListening()
            return portno

        for key, attr in [('httpPort', self.HTTPPort),
                          ('httpsPort', self.HTTPSPort)]:

            if attr:
                prts[key] = attr
            else:
                prts[key] = _findAPort()

        self.deployConfig.update(prts)
        self.pbport = _findAPort()
        
        deferredResult(defer.DeferredList(dL))

    def _getDeploymentString(self, config):
        # this is because i need to be able to monkey patch this method in test_functional
        l = ["""
import os

from quotient.deployment import deploy

config = %s

application = deploy(**config)

from quotient.functional.unittest import SekritAgent
from twisted.spread import pb
from twisted.application import internet

pbfactory = pb.PBServerFactory(SekritAgent())
internet.TCPServer(%d, pbfactory).setServiceParent(application)

""" % (config, self.pbport)]

        if self.includeTelnetPBInTac:
            l.append("""
from twisted.manhole import telnet
telnet.Shell.checkUserAndPass = lambda s, u, p: True
myFactory = telnet.ShellFactory()
myFactory.password = 'guest'
myFactory.username = 'guest'
internet.TCPServer(7771, myFactory).setServiceParent(application)
""")
        return ''.join(l)

    def _spawnChildProcess(self):
        # Okay, let's run the child process!
        self._done = defer.Deferred()
        self.childProcess = ChildProcessProtocol(self._done)

        env = {}
        if os.environ.get('PATH'):
            env['PATH'] = os.environ['PATH']
        env['PYTHONPATH'] = os.path.split(quotient.__path__[0])[0]
        if os.environ.get('PYTHONPATH'):
            env['PYTHONPATH'] += ':' + os.environ['PYTHONPATH']
        log.msg("env: %s" % env)

        TWISTD = (procutils.which("twistd2.3") + procutils.which("twistd"))[0]
        PYTHON = sys.executable

        bin = PYTHON
        args = ['python', TWISTD, "-noy", self.tacfile]

        log.msg(ctimeout='about to spawn child process')
        reactor.spawnProcess(self.childProcess, bin, args, env, self.tdir)
        log.msg(ctimeout='spawned child process')

        def _timeout():
            log.msg(ctimeout='TestCase._spawnChildProcess._timeout called')
            self.fail = True

            # wait for child process to die
            d = self.childProcess.shutdownOrKillYourself()
            assert d is self.childProcess._processExited
            log.msg(ctimeout='waiting for child process to die')
            deferredResult(d)
            msg = """Timed out! could not set up child process!!!!
child processes stderr buffer:

%s
""" % self.childProcess.getErrBuf()
            self._done.errback(FailTest(msg))
        _delayed = reactor.callLater(self.childProcessSetupTimeout, _timeout)

        def _cleanup(ignore):
            log.msg(ctimeout='TestCase._spawnChildProcess._cleanup called')
            if _delayed.active():
                log.msg(ctimeout='_delayed was active!')
                _delayed.cancel()
            return ignore
        self._done.addBoth(_cleanup)
        return self._done


    def setUpClass(self):
        TheOneTrueBaseClass.setUpClass(self)
        
        if self.HTTPPort:
            import warnings
            warnings.warn("Debugging settings are in effect. Be sure to set HTTPPort to None before committing!")

        pemFilePath = opj(self.tdir, "server.pem")
        file(pemFilePath, 'w').write(self.certificate)

        self.deployConfig  = {
            'domainNames': ['localhost'],
            'webLogFile': "web.log",
            'asteriskPort': None,
            'webWrapper': None,
            'manholeUser': 'guest',
            'manholePassword': 'guest',
            'manholePort': 'tcp:48787',
            'initialTicketCount': 3,
            'uid': os.getuid(),
            'gid': os.getgid(),
            'dataDirectory': 'data/db/',
            'fileDirectory': 'data/files',
            'webLogFile': 'web.log',
            'certificateFile': 'server.pem',
            'privateKeyFile': 'server.pem',
        }

        self._initPorts()

        tacpath = opj(os.getcwd(),self.tdir)
        self.tacfile = tacfile = opj(tacpath, 'qtest.tac')
        print "tac_path: %s" % tacpath
        print "signin_url: https://localhost:%s/" % self.deployConfig['httpsPort'] 
        
        from pprint import PrettyPrinter
        pp = PrettyPrinter()

        dc = pp.pformat(self.deployConfig)
        log.msg("deployConfig: %s" % dc)
        assert dc
        
        fp = file(tacfile, 'w')
        ds = self._getDeploymentString(dc)
        log.msg(ctimeout=ds)
        fp.write(ds)
        fp.flush(); fp.close()

        def _eb(fail):
            log.msg(ctimeout='_eb called')
            print fail.getTraceback()
            return fail

        d = self._spawnChildProcess().addErrback(_eb)
        deferredResult(d)

        log.msg(ctimeout='about to set up sekritHQ')

        self.sekritHQ = SekritHQ(self.pbport)
        self.sekritHQ.connect()
        self.sekritHQ.startObserving()

        self._didSetUpClass = True


    def tearDownClass(self):
        if not self._didSetUpClass:
            return

        pidfile = opj(self.tdir, 'twistd.pid')
        log.msg('pidfile: %s' % pidfile)

        self.sekritHQ.disconnect()

        d = self.childProcess.shutdownOrKillYourself()
        deferredResult(d)
        
        client.urlTrail.write('</ol></body></html>')
        client.urlTrail.close()

    def setUp(self):
        client.port = self.deployConfig['httpPort']
        self.sekritHQ.checkForTracebacks()
        
    def tearDown(self):
        self.sekritHQ.checkForTracebacks()

    def doSignup(self, **kw):
        from quotient.functional.navtags import url, link, form
        if not kw:
            kw = self.authData.signup
        return [url('/'), link('signup'), form('signup', **kw)]
        
    def doSignin(self, **kw):
        # absolutely positvely guarantees you will get a fresh session
        # OR YOUR MONEY BACK!
        from quotient.functional.navtags import url, form, assertString
        if not kw:
            signin = self.authData.signin
        from nevow import guard
        return [url("/%s" % guard.LOGOUT_AVATAR), assertString('Claim a Ticket'), url('/'), form('signin', **signin)]

    def writeTestName(self):
        client.urlTrail.write("""</ol><p><b>%s</b></p><ol>""" % self.caseMethodName)

    def _setDiagnosticMessage(self):
        d = '\n\nAssertion failed in page file:/%s\n\nretrieved from url: %s\n\nurltrail is: file:///%s\n\n'
        self.diagnostic = d % (client.urlTrail.filename, client.page.url, client.urlTrail.path)

    def recordResult(self, value):
        try: 
            from quotient.functional.client import IPage
            IPage(value)      # if the value is a Page, this won't throw an exception
            # this is called after the client creates a page object for us
        except components.CannotAdapt:
            pass   # if value isn't a page, skip the rest
        else:
            self.result = client.page.domstr
            log.msg(TestCase="saved page and result")
            self._setDiagnosticMessage()
        return value 

    def runPbRootCommand(self, *a):
        # XXX: a temporary hack while refactoring
        return self.sekritHQ.pbCmd(*a)

    def runSeq(self, *seq):
        from quotient.functional.navtags import runSeq as ntRunSeq
        def cb(result):
            self.recordResult(result)
            self.sekritHQ.checkForTracebacks()
            return result

        def eb(failure):
            # we do this because a child's traceback is more likely to be the cause
            # of a given failure
            self.sekritHQ.checkForTracebacks()
            # what do we do about the original failure?
            return failure

        d = ntRunSeq(seq, perIterCallbacks=(cb,eb))
        deferredResult(d)


class TestCaseWithNoChild(TheOneTrueBaseClass):
    """for faster testing, this connects to a separate,
    constantly running instance of quotient"""

    def setUpClass(self):
        TheOneTrueBaseClass.setUpClass(self)

    def setUp(self):
        TheOneTrueBaseClass.setUp(self)
        client.port = self.HTTPPort
        self._curIdNum = 0
    
# XXX: context.key.split('.')[0] is a total hack
# <fzZzy> yeah, it's rendering the View name, but it should be rendering the instance.name attribute
# <fzZzy> er, it's using the View name as the key
# <fzZzy> instead of the instance.name
def getIdAttr(context, name):
    """generates the proper id attribute for use in href tags"""
    return '%s.%s' % (context.key.split('.')[0], name)


