# -*- test-case-name: quotient.test.test_mimemessage -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from tempfile import TemporaryFile
import random
import datetime
import types
import time

from atop import tpython

from twisted.persisted.styles import Versioned
from twisted.internet import defer, reactor, protocol, ssl as sslModule
from twisted.python.components import implements, Interface, Adapter
from twisted.python import log, failure
from twisted.application import service
from twisted.protocols import imap4
from twisted.python.failure import Failure
from twisted.internet.error import ConnectionDone

import formless
from formless import webform, iformless
from nevow import tags
from nevow import inevow

from atop import store
from atop import sched
from atop import powerup
from atop.service import Serviceup
from atop.fwd import ItemForwardingComponent

from quotient import stats
from quotient import mimemessage
from quotient import provisioner
from quotient import sysup
from quotient.proto import pop3client

class GrabStatus:
    """Formattable object that describes status of a grabber.
    """

    message = "***ERROR***"
    successful = 1

    def __init__(self, message='***ERROR***', successful=1):
        self.lastChanged = time.time()
        self.message = message
        self.successful = successful

    def __str__(self):
        return self.message

    def __repr__(self):
        return '<%s: %r>' % (self.__class__, self.message)

class NeverChecked(GrabStatus):
    def __init__(self, message="Never checked this account."):
        GrabStatus.__init__(self, message)

class CheckingPaused(GrabStatus):
    def __init__(self, message="Checking paused."):
        GrabStatus.__init__(self, message, 0)

class CheckSuccessful(GrabStatus):
    def __init__(self, messageCount):
        self.messageCount = messageCount
        self.curTime = time.localtime()
        GrabStatus.__init__(self, "Last check %s; Retrieved %s messages" % (
            time.asctime(self.curTime), messageCount
        ))

class CheckInProgress(GrabStatus):
    def __init__(self, message="Check in progress..."):
        GrabStatus.__init__(self, message)

class FetchingMessage(GrabStatus):
    def __init__(self, xOf, y):
        GrabStatus.__init__(self, "Retrieving message %s of %s" % (xOf + 1, y))

class AuthFailed(GrabStatus):
    def __init__(self, message):
        GrabStatus.__init__(self, "Server said: %s" % (message,), 0)

class RetrieveFailed(GrabStatus):
    def __init__(self, message):
        GrabStatus.__init__(self, "Retrieval Failed: %s" % (message,), 0)

class ExamineFailed(GrabStatus):
    def __init__(self, message):
        GrabStatus.__init__(self, "IMAP4 EXAMINE Failed: %s" % (message,), 0)

class LoggingIn(GrabStatus):
    def __init__(self, message="Logging in..."):
        GrabStatus.__init__(self, message)

class SelectingMailbox(GrabStatus):
    def __init__(self, message="Selecting mailbox"):
        GrabStatus.__init__(self, message)

class POP3MessageGrabber(pop3client.POP3Client):
    # Connect to server
    # Authenticate
    # Retrieve UID message listing
    # Find UIDs which do not exist locally
    # Prune any UIDs in excess of the limit imposed by maxMessages
    # Retrieve messages of remaining UIDs
    # Before each message, check the factory's paused attribute

    retrieveBackwards = False
    messagesReceived = 0

    def __init__(self, arrival, maxMessages=None):
        """
        @param arrival: The Arrival pool.
        @param maxMessages: The greatest number of messages to
        retrieve before disconnecting.
        """
        self.maxMessages = maxMessages
        self.arrival = arrival

    def serverGreeting(self, challenge):
        pop3client.POP3Client.serverGreeting(self, challenge)
        self.stats = stats.Statoscope(name="popgrabber")
        self.factory.setStatus(LoggingIn())
        self._plaintext(self.factory.username, self.factory.password
            ).addCallbacks(self._cbLogin, self._ebLogin
            ).addErrback(log.err
            )

    def inIndex(self, uid):
        p = self.arrival
        q = p.queryIndex('pop3uid', startKey=uid, endKey=uid, result=store.STORE_ID)
        for storeID in q:
            return True
        return False

    def markedBad(self, uid):
        return uid in self.factory.badUIDs

    def shouldRetrieve(self, uid):
        quid = '%d-%s' % (self.factory.storeID, uid)
        def _():
            return not (self.inIndex(quid) or self.markedBad(uid))
        return self.arrival.store.transact(_)

    def _errback(self, failure):
        if not failure.check(pop3client.LineTooLong):
            log.err(failure)
            self.transport.loseConnection()

    def _cbLogin(self, result):
        self.factory.setStatus(CheckInProgress())
        self.listUID(
            ).addCallbacks(self._cbListUID, self._ebListUID
            ).addErrback(self._errback
            )

    def _ebLogin(self, failure):
        failure.trap(pop3client.ServerErrorResponse)
        self.factory.setStatus(AuthFailed(failure.value.args[0]))
        self.quit().addErrback(self._errback)

    def _cbListUID(self, uidList):
        keep = []
        for (idx, uid) in enumerate(uidList):
            if uid is None:
                continue
            if self.shouldRetrieve(uid):
                keep.append((idx, uid))

        if not self.retrieveBackwards:
            keep.reverse()

        # Let the factory limit how much work we do
        if self.maxMessages:
            keep = keep[-self.maxMessages:]
        self.totalMessages = len(keep)

        if keep:
            self.retrieveMessages(keep)
        else:
            self.quit().addErrback(self._errback)

    def _ebListUID(self, failure):
        failure.trap(pop3client.ServerErrorResponse)
        self.factory.setStatus(RetrieveFailed(failure.value.args[0]))
        self.quit().addErrback(self._errback)

    def retrieveMessages(self, msgs):
        # Let the factory tell us to stop at any time
        if self.factory.paused:
            return self._cbFinishedRetrieve(None)

        # Keep the user informed as to our progress
        soFar = self.messagesReceived + 1
        status = FetchingMessage(soFar, self.totalMessages)
        self.factory.setStatus(status)

        # Grab the next message
        idx, uid = msgs.pop()
        receiver = self.factory.createMIMEReceiver()
        receiver.makeConnection(None)
        d = self.retrieve(idx, lambda line: receiver.lineReceived(line)
            ).addCallback(self._cbRetrieve, receiver, uid, time.time()
            ).addErrback(self._ebRetrieve, receiver, uid
            )
        if msgs:
            d.addCallback(lambda r: self.retrieveMessages(msgs))
        else:
            d.addCallback(self._cbFinishedRetrieve)
        d.addErrback(self._errback)

    def _cbRetrieve(self, result, receiver, uid, started):
        duration = time.time() - started
        receiver.messageDone()
        receiver.connectionLost(Failure(ConnectionDone()))
        msg = receiver.message
        self.factory.store.transact(self.factory.messageReceived, msg, uid)
        self.stats.record(msg.size, duration)
        self.messagesReceived += 1

    def _ebRetrieve(self, failure, receiver, uid):
        failure.trap(pop3client.LineTooLong)
        receiver.connectionLost(failure)
        self.factory.markBad(uid)
        return failure

    def _cbFinishedRetrieve(self, result):
        if self.stats.bytes:
            log.msg('%s: %s'% (self, self.stats))
        self.quit().addErrback(self._errback)

    def connectionLost(self, reason):
        if self.messagesReceived > 0:
            self.factory.setStatus(CheckSuccessful(self.messagesReceived))
            self.factory.successfulCheck = True
        elif self.factory.paused:
            self.factory.setStatus(CheckingPaused("Checking paused: User requested"))
            self.factory.successfulCheck = True

class FileWrapper:
    # XXX - This is ghetto, clean it up
    def __init__(self, messageReceiver, stats = None):
        self.mr = messageReceiver
        self.mr.makeConnection(None)
        self.buffer = ''
        self.stats = stats
        if stats:
            stats.start()
        self.bytes = 0

    def write(self, bytes):
        self.bytes += len(bytes)
        self.buffer += bytes
        lines = self.buffer.splitlines(True)
        for l in lines:
            if l.endswith('\n') or l.endswith('\r\n'):
                l = l.rstrip('\r\n')
                self.mr.lineReceived(l)
            else:
                self.buffer = l
                break

    def seek(self, off, from_):
        pass

    def tell(self):
        return self.bytes

    def close(self):
        self.mr.messageDone()
        self.mr.connectionLost(Failure(ConnectionDone()))
        if self.stats:
            self.stats.finish(self.bytes)

    messageDone = close

MONTHS = ["Jan", "Feb", "Mar", "Apr", "May", "Jun",
          "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]

class IMAP4MessageGrabber(imap4.IMAP4Client):

    allowInsecureLogin = False

    ###
    ### Begin Monkey Patching
    ###
    def rawDataReceived(self, data):
        self._pendingSize -= len(data)
        if self._pendingSize > 0:
            self._pendingBuffer.write(data)
        else:
            passon = ''
            if self._pendingSize < 0:
                data, passon = data[:self._pendingSize], data[self._pendingSize:]
            self._pendingBuffer.write(data)
            rest = self._pendingBuffer
            self._pendingBuffer = None
            self._pendingSize = None
            size = rest.tell()
            rest.seek(0, 0)
            self._parts.append('ELIDED')
            self.elided = rest

            # XXX TRANSACT ME together with the transaction in _cbFetch
            def _():
                self.elided.messageDone()
                self.factory.messageReceived(self.elided.mr.message)
            self.factory.avatar.transact(_)

            self.setLineMode(passon.lstrip('\r\n'))

    def _setupForLiteral(self, rest, octets):
        self._pendingBuffer = self.messageFile(octets)
        self._pendingSize = octets

        # Chop of the {..} bit
        rest = rest[:rest.find('{')]

        self._parts = [rest, '\r\n']
        self.setRawMode()

    def _IMAP4Client__cbFetch(self, (lines, last)):
        flags = {}
        for line in lines:
            parts = line.split(None, 2)
            if len(parts) == 3:
                if parts[1] == 'FETCH':
                    try:
                        id = int(parts[0])
                    except ValueError:
                        raise imap4.IllegalServerResponse, line
                    else:
                        data = imap4.parseNestedParens(parts[2])
                        while len(data) == 1 and isinstance(data, types.ListType):
                            data = data[0]
                        while data:
                            if len(data) < 2:
                                raise imap4.IllegalServerResponse("Not enough arguments", data)
                            if data[1] == 'ELIDED':
                                data[1] = self.elided
                                self.elided = None
                            flags.setdefault(id, {})[data[0]] = data[1]
                            del data[:2]
        return flags

    ###
    ### End Monkey Patching
    ###

    ###
    ### Callbacks
    ###

    def messageFile(self, octets):
        mr = self.factory.createMIMEReceiver()
        f = FileWrapper(mr, self.stats)
        self.onMessage += 1
        self.setFetching()
        return f

    def serverGreeting(self, caps):
        self.factory.setStatus(LoggingIn())
        self.stats = stats.Statoscope(name="imapgrabber")
        self.registerAuthenticator(imap4.CramMD5ClientAuthenticator(self.factory.username))
        self.registerAuthenticator(imap4.LOGINAuthenticator(self.factory.username))
        d = self.authenticate(self.factory.password)
        d.addCallbacks(self._cbLogin, self._ebLogin).addErrback(self._ebDoDisconnect)

        # reactor.callLater(60, self.nooploop)

    def newMessages(self, exists, recent):
        if self.factory.paused:
            return

        if recent:
            if self.lastUid is None:
                msgs = imap4.MessageSet(1, None)
            else:
                msgs = imap4.MessageSet(self.lastUid, None)
            d = self.fetchUID(msgs, uid=True)
            d.addCallbacks(self._cbFetchUids, self._ebDoDisconnect)

    ###
    ### End callbacks
    ###

    def _loggedOut(self, something):
        if isinstance(something, failure.Failure):
            log.err(something)
        if self.stats.bytes:
            log.msg('%s: %s' % (self, self.stats))

    def _ebDoDisconnect(self, failure):
        log.err(failure)
        self.factory.setStatus(AuthFailed(str(failure.value)))
        self.logout().addBoth(self._loggedOut)

    def _cbLogin(self, result):
        self.factory.setStatus(SelectingMailbox())
        self.select(self.factory.mailbox).addCallbacks(self._cbSelect, self._ebSelect)

    def _ebLogin(self, failure):
        failure.trap(imap4.NoSupportedAuthentication)
        if self.startedTLS or self.allowInsecureLogin or self.transport.getHost()[0] == 'SSL':
            return self.login(self.factory.username, self.factory.password
                ).addCallback(self._cbLogin
                )
        else:
            self.factory.setStatus(AuthFailed("No secure authentication available"))
            self.factory.paused = True
            self.logout().addBoth(self._loggedOut)

    def _ebSelect(self, failure):
        self.examine(self.factory.mailbox).addCallbacks(self._cbSelect, self._ebDoDisconnect)

    def _cbSelect(self, details):
        self.factory.setStatus(CheckInProgress())
        try:
            uidvalidity = str(details['UIDVALIDITY'])
        except KeyError:
            # This is totally bunk, the server is probably broken
            self.factory.setStatus(ExamineFailed("Server is probably broken - missing UIDVALIDITY"))
            self.logout().addBoth(self._loggedOut)
        else:
            if not self.factory.uidvalidity:
                self.factory.setUIDValidity(uidvalidity)      # First time we've checked this box
            elif self.factory.uidvalidity != uidvalidity:
                self.factory.setStatus(RetrieveFailed("POD PEOPLE: Mailbox was replaced"))
                self.logout().addBoth(self._loggedOut)
                return

            if details.get('EXISTS', 0):
                # Fetch everything newer than the last UID we saw.
                if self.lastUid is None:
                    msgs = imap4.MessageSet(1, None)
                else:
                    msgs = imap4.MessageSet(self.lastUid, None)
                d = self.fetchUID(msgs, uid=True)
                d.addCallbacks(self._cbFetchUids, self._ebDoDisconnect)
            else:
                self.factory.setStatus(CheckSuccessful(0))

    def _cbFetchUids(self, flags):
        v = [int(v['UID']) for v in flags.itervalues()]
        low = min(v)
        high = max(v)
        if high <= self.lastUid:
            self.factory.setStatus(CheckingPaused("No new messages"))
        else:
            if self.lastUid is None:
                msgs = imap4.MessageSet(low, high)
            else:
                msgs = imap4.MessageSet(max(low, self.lastUid + 1), high)
            self.onMessage = -1
            self.totMessages = len([x for x in flags.itervalues() if int(x['UID']) in msgs])
            self.setFetching()
            d = self.fetchMessage(msgs, uid=True)
            d.addCallbacks(self._cbFetch, self._ebDoDisconnect)

    def setFetching(self):
        self.factory.setStatus(FetchingMessage(self.onMessage, self.totMessages))

    def _cbFetch(self, messages):
        filterUID = self.lastUid
        def _():
            # XXX TRANSACT ME with the transaction in rawDataReceived
            for result in messages.values():
                if int(result['UID']) > self.lastUid:
                    self.factory.setLastUid(int(result['UID']))
        self.factory.avatar.transact(_)

        self.factory.setStatus(CheckSuccessful(len(messages)), True)
        self.cleanupOldMessages()

    def cleanupOldMessages(self):
        if self.factory.age is not None:
            # Retrieve the dates of all the messages in this folder
            self.fetchInternalDate('1:*', True
                                   ).addCallbacks(self._cbCleanup, self._ebCleanup
                                                  )
        else:
            self.logout().addBoth(self._loggedOut)

    def _ebCleanup(self, failure):
        log.err(failure)
        return self.logout().addBoth(self._loggedOut)

    def _cbCleanup(self, result):
        delete = self.filterOld(result, self.factory.age)
        if delete:
            return self.setFlags(','.join(map(str, delete)), ['\\Deleted'], uid=True
                                 ).addCallbacks(self._cbDelete, self._ebDelete
                                               )

    def _ebDelete(self, failure):
        self.log(failure)
        return self.logout().addBoth(self._loggedOut)

    def _cbDelete(self, result):
        return self.logout().addBoth(self._loggedOut)

    def filterOld(self, result, age, now=None):
        if now is None:
            now = datetime.date.today()
        delete = []
        for (uid, date) in result.iteritems():
            try:
                day, month, year = date['INTERNALDATE'].split(None, 1)[0].split('-', 2)
                day = int(day)
                year = int(year)
                month = MONTHS.index(month)
                date = datetime.date(year, month + 1, day)
            except:
                log.msg("Junk response from FETCH INTERNALDATE query")
                log.err()
            else:
                if now.toordinal() - date.toordinal() > age:
                    delete.append(uid)
        return delete

    _pendingCall = None
    def nooploop(self, call=None):
        def setNoop(call):
            self._pendingCall = call
        self.noop(
            ).addCallback(lambda _: reactor.callLater(60, self.nooploop)
            ).addCallback(setNoop
            )

    def connectionLost(self, reason):
        if self._pendingCall and self._pendingCall.active():
            self._pendingCall.cancel()
        self._pendingCall = None


def check(self):
    """Check

    Check mail right now.
    """
    pass
check = formless.autocallable(check)


class MailPort(formless.Integer):
    pass


class MailPortValidator(Adapter):
    __implements__ = iformless.IInputProcessor,

    def process(self, context, boundTo, data):
        return self.original.coerce(data[0], None), bool(context.locate(inevow.IRequest).args.get('ssl', [''])[0])


class MailPortRenderer(webform.BaseInputRenderer):
    def input(self, context, slot, data, name, value):
        try:
            port, checked = value
        except:
            port = value
            checked = None
        ssl = self.original.getAttribute('ssl')
        return [
            tags.script(language="javascript")[tags.xml("""
function on%(name)s(cb) {
    var val = document.getElementById('%(key)s');
    if (cb.checked) {
        val.value = %(ssl)s
    } else {
        val.value = %(default)s
    }
}
                """ % {'key': context.key, 'ssl': ssl, 'default': port, 'name': context.key.replace('.', '_')})
            ],
            tags.input(name=name, type="text", value=port, id=context.key), " SSL: ", tags.input(
                type="checkbox", name="ssl",
                onclick="on%s(this)" % context.key.replace('.', '_'),
                checked = checked and 'checked' or None)
        ]


username = formless.String(label="Username", description="Your Username.", required=True)
password = formless.Password(label="Password", description="Your Password.", required=True)
# Don't actually check to see if the host exists until freeform supports deferreds
#host = formless.Hostname(label="Host", description="The address of the Host, eg mail.mac.com")
host = formless.String(label="Host", description="The address of the Host, eg mail.mac.com", required=True)
port = formless.Integer(label="Port", description="The Port to connect to", required=True)

msgsPer = formless.Integer(
    label = "Messages Per Grab",
    description = "Maximum number of messages to retrieve per connection (lets you train SpamBayes a little so your messages are classified properly when they do arrive). Set to 0 to grab all messages from the server at once.",
    default = 0,
)
paused = formless.Boolean(
    label = "Pause Grabber",
    description = "Suspend checking of this account",
)

age = formless.Integer(
    label = "Deletion Age",
    description = "Minimum age of messages to delete from this account's server.  Set to 0 to disable this feature.",
    default = 0)

popport = MailPort(label="Port", description="The POP3 Port to connect to", default=110, ssl=995)
imapport = MailPort(label="Port", description="The IMAP Port to connect to", default=143, ssl=993)
#popport = formless.Integer(label="Port", description="The POP3 Port to connect to", default=110)
#imapport = formless.Integer(label="Port", description="The IMAP Port to connect to", default=143)

ssl = formless.Boolean(
    label = "SSL",
    description = "Use a secure (encrypted) socket layer for connections",
)

address = formless.String(
    label="Email Address",
    description = "The email address associated with this account"
)


class IGrabberFactory(formless.TypedInterface):
    check = check
    password = password
    class GrabberAttributes(formless.TypedInterface):
        """A grouping of all the grabber attributes, so that
        they can all be changed in one transaction.

        Except for password. That has to be changed
        in a different way.
        """
        interval = formless.Integer(
            label = "Interval",
            description = "The amount of time to wait between checks for new mail.")
        username = username
        host = host
        port = port
        address = address
        ssl = ssl
        msgsPer = msgsPer
        paused = paused


class GrabberFactory(store.Item, protocol.ClientFactory, sched.Schedulable, object):
    __implements__ = sched.Schedulable.__implements__, IGrabberFactory,

    persistenceForgets = ['currentProtocol', 'doneDeferred', 'successfulCheck']

    _lastGrabTime = 0
    interval = 60 * 5

    jitter = 0.11962656492 # Molar Planck constant times c, Joule meters / mole.

    age = None
    debug = False
    _paused = False
    _initialized = False

    persistenceVersion = 1

    def upgradeToVersion1(self):
        self.touch()
        d = vars(self)
        self.username = d.pop('username')
        self.host = d.pop('host')
        self.port = d.pop('port')
        self._initialized = True
        self.name = self._calcName()

    def __init__(self, avatar, username, password, host,
                 port, address, msgsPer=0, paused=False,
                 interval=None, ssl=False, age=None):
        """
        @param username: The username with which to log into this account.
        @param password: The password with which to authenticate for this account.

        @param host: The hostname of the server to which to connect.
        @param port: The port number of the server to which to connect.

        @param address: The email address to associated with messages received from this account.

        @param msgsPer: The maximum number of messages to retrieve per connection.
        @param interval: The number of seconds between checks.
        @param ssl: Indicate whether or not to use SSL for this connection.
        @param age: The age, in days, of messages to delete from the server.
        """
        store.Item.__init__(self, avatar)
        self.username = username
        self.password = password
        self.host = host
        self.port = port
        self.address = address
        if interval is not None:
            self.interval = interval
        self.msgsPer = msgsPer
        self.paused = paused
        self.ssl = ssl
        self.age = age
        self.setStatus("Initializing...")

        self._initialized = True
        self.name = self._calcName()

    def __repr__(self):
        return "%s://%s@%s (%s)" % (self.protoName, self.username, self.host, self.status)

    def getInterval(self):
        return random.normalvariate(self.interval, self.interval * self.jitter)

    def delete(self):
        IGrabberAdder(self.store).delete([self])

    def _calcName(self):
        if not self._initialized:
            return '<uninitialized %s>' % (self.__class__.__name__,)
        return '%s://%s@%s:%s/' % (self.protoName, self.username,
                                   self.host, self.port)

    def makeNameSensitiveToProperty(propname):
        def get(self):
            return getattr(self, '_nameProp_' + propname)
        def set(self, value):
            setattr(self, '_nameProp_' + propname, value)
            self.name = self._calcName()
        return property(get, set)

    username = makeNameSensitiveToProperty('username')
    host = makeNameSensitiveToProperty('host')
    port = makeNameSensitiveToProperty('port')

    del makeNameSensitiveToProperty

    def ssl():
        def get(self):
            return self._ssl
        def set(self, value):
            self.touch()
            self._ssl = value
            if self._ssl:
                if self.port == self.DEFAULT_PORT:
                    self.port = self.SSL_PORT
                if 'protoName' not in vars(self):
                    self.protoName = self.protoName + 's'
            else:
                if self.port == self.SSL_PORT:
                    self.port = self.DEFAULT_PORT
                if 'protoName' in vars(self):
                    del self.protoName
            self.name = self._calcName()
        return get, set
    ssl = property(*ssl())

    def paused():
        def get(self):
            def getPaused():
                if self._paused:
                    return True
                cp = sysup.IControlPanel(self.store.store)
                perm = sysup.IControlPanel.EXTERNAL_GRABBING
                return not cp.actionAllowed(perm)
            return self.store.transact(getPaused)
        def set(self, value):
            def setPaused():
                self.touch()
                self._paused = value
                if self._paused:
                    self.unschedule()
                else:
                    self.reset(self.getInterval())
            self.store.transact(setPaused)
        return get, set
    paused = property(*paused())

    def avatar():
        def _getAvatar(self):
            return self.store
        def _setAvatar(self):
            raise NotImplementedError("WHAT THE HECK, YO")
        return _getAvatar, _setAvatar
    avatar = property(*avatar())

    def createMIMEReceiver(self, trustDate=True):
        md = mimemessage.IMIMEDelivery(self.store)
        args = (self.protoName, self.username, self.host, self.port)
        origin = '%s://%s@%s:%s/' % args
        return md.createMIMEReceiver(trustDate, origin)

    doneDeferred = None
    currentProtocol = None
    lastUid = None
    successfulCheck = False
    statusSet = False

    def setLastUid(self, lastUid):
        """Call me from within a transaction.
        """
        self.touch()
        self.lastUid = lastUid

    def setStatus(self, stat, success=False):
        def _():
            # TODO: move this elsewhere
            self.status = stat
            self.statusSet = True
            self.successfulCheck = success
        self.avatar.transact(_)

    def messageReceived(self, message):
        self.touch()

        # self.avatar.realm.service.stats.processedBytes += bytes
        # self.avatar.realm.service.stats.receivedMessages += 1

        message.setMessageAddress(self.address or ('@'.join((self.username, self.host))))

        try:
            provisioner.IProvisioner(self.avatar).deplete(self.tankID, message.size)
        except provisioner.NotEnoughFuel:
            self.paused = True
            self.setStatus(CheckingPaused("Quota exceeded"))

    def startedConnecting(self, connector):
        args = (self.protoName, self.username, self.host, self.port)
        log.msg('%s://%s@%s:%s/ now connecting' % args)

    def clientConnectionFailed(self, connector, reason):
        log.err(reason)
        def _():
            self.reset(self.getInterval())

            self.currentProtocol = None
            self.setStatus(CheckingPaused(str(reason.value)))
            if self.doneDeferred:
                self.doneDeferred.errback(reason)
                self.doneDeferred = None
        self.avatar.transact(_)

    def clientConnectionLost(self, connector, reason):
        def _():
            self.currentProtocol = None
            if not self.successfulCheck and not self.statusSet:
                self.setStatus(CheckingPaused(str(reason.value)))
            if self.doneDeferred:
                self.doneDeferred.callback(None)
                self.doneDeferred = None
        self.avatar.transact(_)

    def buildProtocol(self, addr):
        assert not self.currentProtocol, \
               "GrabberFactory can manage only one connection at a time"
        p = self.protocol()
        p.factory = self
        p.lastUid = self.lastUid
        p.maxMsgs = self.msgsPer or None
        self.currentProtocol = p
        self.successfulCheck = False
        self.statusSet = False
        return self.debug and tpython.DebugShimTransport(p) or p

    _webUI_nextRun = 0
    _webUI_forceCount = 0

    def check(self):
        self.touch()

        # This restriction is a hack to be replaced at a later date
        # with a real provisioning-based solution.
        if self._webUI_nextRun < time.time():
            self._webUI_forceCount += 1
            self._webUI_nextRun = time.time() + (2 ** self._webUI_forceCount)
            self.run()
            return "Checking"
        else:
            return "Working!  Be patient."

    def run(self):
        """This function must be run in a storage context.
        """
        self.reset(self.getInterval())

        if self.paused or self.doneDeferred is not None:
            return self.doneDeferred

        self.doneDeferred = defer.Deferred()
        if self.ssl:
            if not sslModule.supported:
                log.err(failure.Failure(Exception("SSL connection requested, but SSL not available")))
                return None
            else:
                reactor.connectSSL(self.host, self.port, self, sslModule.ClientContextFactory())
        else:
            reactor.connectTCP(self.host, self.port, self)

        return self.doneDeferred

    def format(self):
        """Return a summary of this grabber as a nicely formatted string.
        """
        ## This should be removed and replaced with the IPreview adapter for grabber factories.
        return "%s@%s" % (self.username, self.host)


class POP3GrabberFactory(GrabberFactory):
    protoName = 'pop3'
    protocol = POP3MessageGrabber
    status = NeverChecked()

    MAX_BAD_UIDS = 10
    DEFAULT_PORT = 110
    SSL_PORT = 995

    messagePool = store.referenced

    persistenceVersion = 3

    def upgradeToVersion3(self):
        self.touch()
        self.badUIDs = []

    def upgradeToVersion2(self):
        self.touch()
        self.messagePool = self.poolRef.getItem()
        del self.poolRef

    def upgradeToVersion1(self):
        self.touch()
        from quotient.application import IMailApp
        from quotient.mimemessage import POP3UIDIndex
        arrival = IMailApp(self.store).appRoot.getIndexItem('name', 'Arrival')
        arrival.addIndex(POP3UIDIndex())

    def __init__(self, *a, **kw):
        GrabberFactory.__init__(self, *a, **kw)
        p = store.Pool(self.store)
        p.addIndex(store.DateIndex())
        self.messagePool = p
        self.badUIDs = []

    def getArrivalPool(self):
        def _():
            md = mimemessage.IMIMEDelivery(self.store)
            return md.getArrivalRef().getItem()
        return self.store.transact(_)

    def markBad(self, uid):
        def _():
            self.touch()
            self.badUIDs.append(uid)
            if len(self.badUIDs) > self.MAX_BAD_UIDS:
                self.setStatus(CheckingPaused("Too many unretrievable messages encountered"))
                self.paused = True
        self.store.transact(_)

    def buildProtocol(self, addr):
        assert not self.currentProtocol, \
               "GrabberFactory can manage only one connection at a time"
        p = self.protocol(self.getArrivalPool(), self.msgsPer)
        p.factory = self
        self.currentProtocol = p
        self.successfulCheck = False
        self.statusSet = False
        return self.debug and tpython.DebugShimTransport(p) or p

    def messageReceived(self, msg, uid):
        GrabberFactory.messageReceived(self, msg)
        msg.touch()
        msg.setIndexKey('pop3uid', '%d-%s' % (self.storeID, uid))
        if self.age:
            self.messagePool.addItem(msg)

    def getDeletionList(self, now=None):
        """Returns a list of UIDs which should be deleted

        UIDs returned will have their corresponding messages
        removed from the pop3-uid pool.
        """
        if now is None:
            now = time.gmtime()
        if not self.age:
            return []
        p = self.messagePool
        r = []
        then = time.mktime(now) - (60 * 60 * 24 * self.age)
        def _():
            for m in p.queryIndex('received', endKey=then):
                r.append(m.getIndexKey(None, 'pop3uid').split('-', 1)[1])
                p.removeItem(m)
            return r
        return self.avatar.transact(_)

class IMAP4GrabberFactory(GrabberFactory):
    protoName = 'imap4'

    # To be subsumed by, say, IMAP4MessageSynchronizer
    protocol = IMAP4MessageGrabber
    status = NeverChecked()

    DEFAULT_PORT = 143
    SSL_PORT = 993

    mailbox = 'inbox'
    uidvalidity = None

    def setUIDValidity(self, validity):
        def _():
            self.touch()
            self.uidvalidity = validity
        self.avatar.transact(_)

class IGrabberAdder(formless.TypedInterface):
    def addPopGrabber(self, username = username, password = password,
                      host = host, port = popport, address = address,
                      msgsPer = msgsPer):
        """Add Pop Grabber

        Add a POP3 email account to be checked by this Quotient account.
        """
    addPopGrabber = formless.autocallable(addPopGrabber)

    def addIMAPGrabber(self, username = username, password = password,
                       host = host, port = imapport,
                       address = address, msgsPer = msgsPer):
        """Add IMAP Grabber

        Add an IMAP email account to be checked by this Quotient account.
        """
    addIMAPGrabber = formless.autocallable(addIMAPGrabber)

    def delete(self, grabbers=formless.List()):
        """Delete a list of grabbers
        """

    grabbers = formless.List(actions=[delete], label="Existing accounts")


class IGrabberManager(Interface):
    def add(self, avatar):
        pass

class AvatarPool(store.Pool):
    pass

class GrabberManager(powerup.Storeup):
    avatars = store.referenced

    dependencies = [Serviceup]
    powerupName = 'Grabber Manager'

    def setUpPools(self, st):
        st.setComponent(IGrabberManager, self)
        self.avatars = AvatarPool(st, 'Grabber Adders')
        service.IServiceCollection(st).addService(self)

    def add(self, avatar):
        self.avatars.addItem(avatar)

    def allFactories(self):
        for a in self.avatars.queryIndex():
            ga = IGrabberAdder(a)
            for gf in ga.grabbersRef.getItem().queryIndex():
                yield gf

    def connectedProtocols(self):
        for gf in self.allFactories():
            if gf.currentProtocol is not None:
                yield gf.currentProtocol

    # IService implementation
    def privilegedStartService(self):
        raise AssertionError("privilegedStartService *cannot* be called!")

    def startService(self):
        pass

    def stopService(self):
        cp = None
        for cp in self.connectedProtocols():
            cp.transport.loseConnection()
        if cp is not None:
            # XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
            # loseConnection sucks XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
            # XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
            d = defer.Deferred()
            reactor.callLater(5, d.callback, None)
            return d

class GrabberPowerup(provisioner.ProvisionedPowerup):
    __implements__ = (IGrabberAdder, provisioner.IProvisionable,
                      provisioner.ProvisionedPowerup.__implements__)

    dependencies = [mimemessage.EmailPowerup, sched.SubScheduler]
    storeDependencies = [sysup.ControlPanel, GrabberManager]
    powerupName = "External Accounts"

    tankID = provisioner.ETank.INCOMING_MAIL
    startCost = 1

    persistenceVersion = 1

    def upgradeToVersion1(self):
        self.touch()
        gm = powerup.IPowerStation(self.store.store).installPowerup(GrabberManager)
        gm.add(self.store)

    def setUpPools(self, avatar):
        provisioner.ProvisionedPowerup.setUpPools(self, avatar)
        g = store.Pool(self.store, 'grabbers')
        self.grabbersRef = g.referenceTo()
        avatar.setComponent(IGrabberAdder, self.fwd())
        IGrabberManager(avatar.store).add(avatar)

    def delete(self, grabbers):
        for grabber in grabbers:
            grabber.paused = True
            grabber.unschedule()
            self.grabbersRef.getItem().removeItem(grabber)

    def getGrabbers(self):
        return list(self.grabbersRef.getItem().queryIndex())
    grabbers = property(getGrabbers)

    def addGrabber(self, grabberFactory, username, password,
                   host, port, address, msgsPer, ssl):
        if not msgsPer:
            msgsPer = 50
        gf = grabberFactory(self.store, username, password,
                            host, port, address, msgsPer,
                            ssl=ssl, age=0)
        gf.reset(1)
        self.grabbersRef.getItem().addItem(gf)
        gf.tankID = self.tankID
        return gf

    def addPopGrabber(self, username, password, host, port,
                      address, msgsPer):
        port, ssl = port
        return self.addGrabber(POP3GrabberFactory, username,
                               password, host, port, address,
                               msgsPer, ssl)

    def addIMAPGrabber(self, username, password, host, port,
                       address, msgsPer):
        port, ssl = port
        return self.addGrabber(IMAP4GrabberFactory, username,
                               password, host, port, address,
                               msgsPer, ssl)
