
"""
Functions for listing mailboxes in an asynchronous,
database-friendly way.
"""

import itertools

from twisted.python import context, log
from twisted.protocols.imap4 import iterateInReactor

def isAnEmailPool(pool):
    from atop.store import Pool
    x = lambda p: isinstance(p, Pool) and 'imapname' in pool.indexes
    return x(pool)

def listmbox(rootPool, continuator):
    """Iterator of mailboxes.

    Generates tuples of fully qualified mailbox names and
    corresponding pools.

    @param rootPool: The pool from which to begin listing
    mailboxes.

    @param continuator: A list which is monitored for
    per-reactor-iteration sensitive objects.  After the
    reactor is allowed to iterate, resume will be called
    on all these objects before the generator is called
    upon again.
    """
    name = None
    q = rootPool.queryIndex('name')
    continuator.append(q)
    for item in q:
        thisName = item.name
        if name is not None:
            thisName = name + '/' + thisName
        yield thisName, item
        if isAnEmailPool(item):
            for subname, subitem in listmbox(item, continuator):
                yield thisName + '/' + subname, subitem
    continuator.pop()

def locksRemaining(store):
    stat = store.env.lock_stat()
    return stat['maxlocks'] - stat['nlocks']

def dirtyItems(store):
    return len(context.get('StorageContext'))

def makeIterCountCondition(count):
    return itertools.count(-count).next

class LockWatchdog:
    """
    A class which keeps track of the ongoing usage of database
    locks and signals when work should be suspended, committed,
    and locks released.
    """

    def __init__(self, store, factor=2):
        self.store = store
        self.factor = factor
        self.beginLocks = self.lastLocks = locksRemaining(store)

    def condition(self):
        """Determine if another unit of work should be undertaken.

        Returns True if there will probably be enough locks for
        another unit of work similar in scope to those previously
        undertaken, False otherwise.
        """
        locksNow = locksRemaining(self.store)
        lockDiff = self.lastLocks - locksNow
        locksUsed = self.beginLocks - locksNow

        if lockDiff > (locksNow / self.factor):
            return False
        if locksUsed * 3 >= locksNow:
            return False
        if dirtyItems(self.store) > locksNow:
            return False

        self.lastLocks = locksNow
        return True

def makeMultiCondition(*conditions):
    """Join one or more conditions into a single.

    If any condition fails, the joint condition fails,
    otherwise it succeeds.
    """
    def multicondition():
        for c in conditions:
            if not c():
                return False
        return True
    return multicondition

def iterateSome(continuatorList, iterator, condition):
    """Consume a chunk of an iterator all in one go.

    The iterator is consumed until C{condition()} is False
    or the iterator is exhausted.

    @return: A list of the objects produced by the iterator.
    """
    for c in continuatorList:
        c.resume()

    L = []
    for i in iter(condition, False):
        try:
            L.append(iterator.next())
        except StopIteration:
            break
    if not L:
        raise StopIteration
    return L

def transactIterationSome(store, continuatorList, iterable, maxPerTransaction=1000):
    """Consume groups of objects from an iterator, each group in a new transaction.

    At most maxPerTransaction objects will be consumed per transaction.  Fewer
    may be consumed if the number of available locks falls too low.
    """
    iterator = iter(iterable)
    while True:
        c1 = LockWatchdog(store).condition
        c2 = makeIterCountCondition(maxPerTransaction)
        condition = makeMultiCondition(c1, c2)
        for e in store.transact(iterateSome, continuatorList, iterator, condition):
            yield e

def listMailboxes(store, rootPool):
    """Return an iterator over the mailboxes in the given pool.
    """
    L = []
    return transactIterationSome(store, L, listmbox(rootPool, L))

from twisted.internet import defer
def iterateInto(i, d, rate):
    l = []
    c = rate
    for e in i:
        l.append(e)
        c -= 1
        if c <= 0:
            yield None
            c = rate
    d.callback(l)

def deferredListify(iterable, rate=1):
    d = defer.Deferred()
    iterateInReactor(iterateInto(iterable, d, rate))
    return d

def deferredPoolVisit(store, query, visitor, numIterations=1):
    txnIter = transactIterationSome(store, [query], query)
    visitIter = itertools.imap(visitor, txnIter)
    cleanupIter = itertools.ifilter(lambda o: None, visitIter)
    return iterateInReactor(cleanupIter).addCallback(lambda r: None)
