# -*- test-case-name: quotient.test.test_sip -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from twisted.application import internet
from twisted.internet import defer
from twisted.names import client
from twisted.internet import error
from twisted.internet import abstract
from twisted.python import failure
from twisted.python import log
from twisted import cred
from atop import credup
from quotient.proto import twisted_sip as sip
from quotient import voiceup

def lookupService(domain):
    return client.lookupService('_sip._udp.' + domain)

class SIPServer(internet.UDPServer):
    def __init__(self, port, protocol, registry, locator):
        internet.UDPServer.__init__(self, port, protocol)
        self.registry = protocol.registry = registry
        self.locator = protocol.locator = locator

RegisterProxy = sip.RegisterProxy

# class RegisterProxy(sip.RegisterProxy):
#     def _cantForwardRequest(self, failure, message):
#         if failure.check(LookupError) and not self.locator.domainExists(message.uri.host):
#             return lookupService(message.uri.host
#                 ).addCallback(self._cbSRVLookup, message
#                 )
#         return sip.RegisterProxy._cantForwardRequest(self, failure, message)

#     def _cbSRVLookup(self, (ans, auth, add), message):
#         if not ans:
#             f = failure.Failure(LookupError("No SIP service at that domain"))
#             return sip.RegisterProxy._cantForwardRequest(self, f, message)
#         f = lambda s: (s.payload.priority, -s.payload.weight, s.payload.name, s.payload.port)
#         srvs = [f(s) for s in ans if s.type == dns.SRV]
#         srvs.sort()
#         srv = str(srvs[0][2])
#         port = srvs[0][3]

#         if abstract.isIPAddress(srv):
#             d = defer.succeed(srv)
#         else:
#             addr = [n.payload.dottedQuad() for n in (ans + auth + add) if n.type == dns.A and n.name == srv]
#             if addr:
#                 d = defer.succeed(addr[0])
#             else:
#                 d = client.getHostByName(srv)
#         d.addCallback(self._cbSRVResolve, port, message)

#     def _cbSRVResolve(self, address, port, message):
#         self.sendMessage((address, port), message)

class MultiDomainRegistry:
    __implements__ = sip.IRegistry, sip.ILocator

    def __init__(self, domains=None):
        if domains is None:
            domains = ()
        self.domains = dict([(d, sip.InMemoryRegistry(d)) for d in domains])

    def addDomain(self, name):
        self.domains[name] = sip.InMemoryRegistry(name)

    def removeDomain(self, name):
        del self.domains[name]

    def domainExists(self, name):
        return name in self.domains

    ###
    ### Begin IRegistry implementation
    ### 
    def registerAddress(self, avatar, domainURL, logicalURL, physicalURL, expires):
        """Register the physical address of a logical URL.

        @return: Deferred of (secondsToExpiry, contact URL) or failure with RegistrationError.
        """
        try:
            d = self.domains[domainURL.host.lower()]
        except KeyError:
            s = "Registration for unhandled domain: " + domainURL.host
            log.msg(s)
            return defer.fail(LookupError(s))
        else:
            return d.registerAddress(avatar, domainURL, logicalURL, physicalURL, expires)

    def unregisterAddress(self, domainURL, logicalURL, physicalURL):
        d = self.domains[domainURL.host.lower()]
        return d.unregisterAddress(domainURL, logicalURL, physicalURL)


    def getRegistrationInfo(self, logicalURL):
        """Get registration info for logical URL.

        @return: Deferred of (secondsToExpiry, contact URL) or failure with LookupError.
        """
        try:
            d = self.domains[logicalURL.host.lower()]
        except KeyError:
            s = "Unknown domain in getRegistrationInfo: " + logicalURL.host
            log.msg(s)
            return defer.fail(LookupError(s))
        else:
            x = d.getRegistrationInfo(logicalURL)
	    print "Registration info for %s is %s" % (logicalURL, x)
	    return x
    ###
    ### End IRegistry implementation
    ###

    ###
    ### Begin ILocator implementation
    ###

    def getAddress(self, logicalURL, askingURL=None):
        """Return physical URL of server for logical URL of user.

        @param userURL: a logical C{URL}.
        @return: Deferred which becomes URL or fails with LookupError.
        """
        d = self.domains.get(logicalURL.host.lower(), None)
        if d:
            return d.getAddress(logicalURL)
        else:
            # maybe that commented-out RegisterProxy is a better
            # approach? i dunno.

            return self._lookupURI(logicalURL)

    def _lookupURI(self, userURI):
        if abstract.isIPAddress(userURI.host):
            # it is an IP not a hostname
            if not userURI.port:
                userURI.port = 5060
            return defer.succeed(userURI)
        else:
            d = client.lookupService('_sip._udp.' + userURI.host)
            d.addCallback(self._resolveSRV, userURI)
            return d

    def _resolveSRV(self, (answers, _, __), userURI):
        if answers:
            answers = [(a.payload.priority, str(a.payload.target), a.payload.port) for a in answers]
            answers.sort()
            #always return the first one 'cause we don't do stateful proxying
            #(yes this is sorta bad)
            return sip.URL(host=answers[0][1], port=answers[0][2])
        else:
            #just do an A lookup
            if not userURI.port:
                userURI.port = 5060
            return userURI


    ###
    ### End ILocator implementation
    ###

class ForwardingRegistry(MultiDomainRegistry):
    def __init__(self, domains, portal, asteriskhost, asteriskport):
        if domains is None:
            domains = ()
        self.domains = dict([(d, sip.InMemoryRegistry(d)) for d in domains])
        self.asteriskhost = asteriskhost
        self.asteriskport = int(asteriskport)
        self.portal = portal

    def getAddress(self, logicalURL, askingURL=None):
        def forwardLookup(err):
            D = self.portal.login(credup.Preauthenticated(logicalURL.username),
                                  None, sip.IContact)
            def exceptor(innererr):
                innererr.trap(cred.error.UnauthorizedLogin)
                d = self.domains.get(askingURL.host, None)
                if d and d.users.has_key(askingURL.username):
                    return defer.succeed(sip.URL(self.asteriskhost, port=self.asteriskport, username=askingURL.username))
                else:
                    raise sip.SIPError(401)

            def successor((interface, ava, logout)):
                # user exists, is not logged in:
                # if altcontact set to a SIP url, return that
                c = ava.altcontact
                if c and c.startswith('sip:'):
                    return twisted_sip.parseURL(c)
                # otherwise for now, just return Temporarily Unavailable
                raise sip.SIPError(480)

            return D.addCallbacks(successor, exceptor)

        domain = self.domains.get(logicalURL.host.lower(), None)
        if domain:
            return domain.getAddress(logicalURL).addErrback(forwardLookup)
        else:
            return self._lookupURI(logicalURL)
