# -*- test-case-name: quotient.test.test_smtp -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

"""Incoming ESMTP server for delivery to users."""

import weakref

from twisted.protocols import smtp
from twisted.internet import protocol, defer
from twisted.python import log
from twisted.python import context
from twisted.cred.checkers import ICredentialsChecker
from twisted.internet.interfaces import ISSLTransport, ITLSTransport

from atop import tpython
from atop.store import MetaItem, Item, referenced
from atop.upgrade import DeleteMe
from atop.credup import Preauthenticated

from quotient import provisioner
from quotient.copyright import longversion
from quotient.mimemessage import IMIMEDelivery

from twisted.python.failure import Failure

def validateTo(portal, user):
    """Verify email is valid Quotient user."""
    user, dom = user.dest.local, user.dest.domain
    route = user.split('.')
    user = route.pop()
    fullname = user + '@' + dom
    origin = "smtp:" + fullname
    return portal.login(Preauthenticated(fullname), None, IMIMEDelivery
        ).addErrback(lambda F: Failure(smtp.SMTPBadRcpt(user))
        ).addCallback(lambda (I,A,L): lambda : Message(A.createMIMEReceiver(False,
                                                                            origin,
                                                                            route)))

QuotientMessageDeliveryFactory = DeleteMe

class QuotientMessageDelivery:
    """Message delivery into Quotient."""

    __implements__ = (smtp.IMessageDelivery,)

    tankID = provisioner.ETank.QUOTA

    def __init__(self, portal):
        """Create a QuotientMessageDelivery object.

        @param portal: a portal which can be logged into as a pre-authenticated
        user for an IMIMEDelivery avatar.
        """
        self.portal = portal

    def validateFrom(self, helo, origin):
        """Accept all Froms by default."""
        return origin

    def validateTo(self, user):
        return validateTo(self.portal, user)

    def receivedHeader(self, helo, origin, recipients):
        # XXX Hack?
        domain = str(recipients[0]).split('@', 1)[1]

        heloStr = ""
        if helo[0]:
            heloStr = " helo=%s" % (helo[0],)
        auth = ''
        from_ = "from %s ([%s]%s)%s" % (helo[0], helo[1], heloStr, auth)
        by = "by %s with ESMTP (%s)" % (domain, longversion)
        for_ = "for %s; %s" % (' '.join(map(str, recipients)), smtp.rfc822date())
        return "Received: %s\n\t%s\n\t%s" % (from_, by, for_)

_internalServerError = smtp.SMTPServerError(550, "Internal server error")

class Message:
    """A message being delivered by the ESMTP server."""

    __implements__ = smtp.IMessage,

    def __init__(self, receiver):
        self.receiver = receiver
        try:
            self.receiver.makeConnection(None)
        except:
            log.err()
            raise _internalServerError

    def lineReceived(self, line):
        try:
            return self.receiver.lineReceived(line)
        except:
            log.err()
            raise _internalServerError

    def eomReceived(self):
        try:
            self.receiver.messageDone()
        except:
            log.err()
            return defer.fail(_internalServerError)
        return defer.succeed("Message delivered")

    def connectionLost(self):
        self.receiver.connectionLost(None)

class ESMTP(smtp.ESMTP):
    def connectionMade(self):
        smtp.ESMTP.connectionMade(self)
        if ISSLTransport(self.transport, default=None) is not None or self.startedTLS:
            from twisted.protocols.imap4 import PLAINCredentials, LOGINCredentials
            self.challengers = self.challengers.copy()
            self.challengers['PLAIN'] = PLAINCredentials
            self.challengers['LOGIN'] = LOGINCredentials

    # XXX Override and improve on the poor error handling
    def connectionLost(self, reason):
        # self.sendCode(421, 'Dropping connection.') # This does nothing...
        # Ideally, if we (rather than the other side) lose the connection,
        # we should be able to tell the other side that we are going away.
        # RFC-2821 requires that we try.
        self.setTimeout(None)
        if self.mode is smtp.DATA:
            try:
                for message in self._SMTP__messages:
                    try:
                        message.connectionLost()
                    except:
                        log.err()
                del self._SMTP__messages
            except AttributeError:
                pass
        if self._onLogout:
            self._onLogout()
            self._onLogout = None

    def do_HELO(self, rest):
        peer = self.transport.getPeer()
        try:
            host = peer.host
        except AttributeError:
            host = str(peer)
        self._helo = (rest, host)
        self._from = None
        self._to = []
        self.sendCode(250, '%s Hello %s, nice to meet you' % (self.host, host))

    def do_DATA(self, rest):
        if self._from is None or (not self._to):
            self.sendCode(503, 'Must have valid receiver and originator')
            return
        assert self.delivery
        self.mode = smtp.DATA
        helo, origin = self._helo, self._from
        recipients = self._to

        self._from = None
        self._to = []
        self.datafailed = None

        try:
            self._SMTP__messages = [f() for (u, f) in recipients]
        except smtp.SMTPServerError, e:
            self.sendCode(e.code, e.resp)
            self.mode = smtp.COMMAND
            return
        except:
            log.err()
            self.sendCode(550, "Internal server error")
            self.mode = smtp.COMMAND
            return

        rcvdhdr = self.delivery.receivedHeader(
            helo, origin, [u for (u, f) in recipients])

        self._SMTP__inheader = self._SMTP__inbody = 0
        if rcvdhdr:
            try:
                for message in self._SMTP__messages:
                    message.lineReceived(rcvdhdr)
            except smtp.SMTPServerError, e:
                self.sendCode(e.code, e.resp)
                self.mode = smtp.COMMAND
                return
        self.sendCode(354, 'Continue')

    def dataLineReceived(self, line):
        if line[:1] == '.':
            if line == '.':
                self.mode = smtp.COMMAND
                if self.datafailed:
                    self.sendCode(self.datafailed.code,
                                  self.datafailed.resp)
                    return
                if not self._SMTP__messages:
                    self._messageHandled("thrown away")
                    return
                defer.DeferredList([
                    m.eomReceived() for m in self._SMTP__messages
                ], consumeErrors=True).addCallback(self._messageHandled
                                                   )
                del self._SMTP__messages
                return
            line = line[1:]

        if self.datafailed:
            return

        try:
            # Add a blank line between the generated Received:-header
            # and the message body if the message comes in without any
            # headers
            if not self._SMTP__inheader and not self._SMTP__inbody:
                if ':' in line:
                    self._SMTP__inheader = 1
                elif line:
                    for message in self._SMTP__messages:
                        message.lineReceived('')
                    self._SMTP__inbody = 1

            if not line:
                self._SMTP__inbody = 1

            for message in self._SMTP__messages:
                message.lineReceived(line)
        except smtp.SMTPServerError, e:
            self.datafailed = e
            for message in self._SMTP__messages:
                message.connectionLost()
    state_DATA = dataLineReceived

    # XXX Twisted handles eomReceived errors wrong
    def _messageHandled(self, resultList):
        failures = 0
        for (success, result) in resultList:
            if not success:
                failures += 1
                log.err(result)
        if failures:
            msg = 'Could not send e-mail'
            L = len(resultList)
            if L > 1:
                msg += ' (%d failures out of %d recipients)' % (failures, L)
            self.sendCode(550, msg)
        else:
            self.sendCode(250, 'Delivery succeeded')


class ESMTPInFactory(object, protocol.ServerFactory):
    """Server protocol for ESMTP.
    """
    __implements__ = protocol.ServerFactory.__implements__, smtp.IMessageDeliveryFactory

    protocol = ESMTP
    debug = False
    logfile = None

    connectedProtocols = None

    def __init__(self, portal, auths = None, ctxFactory = None):
        self.portal = portal
        self.auths = auths
        self.ctxFactory = ctxFactory

    def getMessageDelivery(self):
        return QuotientMessageDelivery(self.portal)

    def startFactory(self):
        self.connectedProtocols = weakref.WeakKeyDictionary()

    def stopFactory(self):
        for p in self.connectedProtocols:
            try:
                p.transport.loseConnection()
            except:
                log.err()
        self.connectedProtocols = None

    def buildProtocol(self, addr):
        p = self.protocol(self.auths, self.ctxFactory)
        p.portal = self.portal
        p.factory = self
        p.deliveryFactory = self
        self.connectedProtocols[p] = True
        if self.debug:
            if not self.logfile:
                self.logfile = file('smtpin.debug', 'w')
            p = tpython.DebugShimTransport(p, self.logfile)
        return p
