# -*- test-case-name: quotient.test.test_provisioner -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from twisted.python import log
from twisted.python.components import Interface
from twisted.python.components import registerAdapter
from twisted.cred.error import UnauthorizedLogin

from atop.fwd import ItemForwardingComponent
from atop.store import IComponentForwarder
from atop.powerup import Powerup

from quotient import sysup

class NotEnoughFuel(Exception):
    """I could not perform the task you asked for!
    """

class ProvisionerDisabled(NotEnoughFuel):
    """I couldn't perform the task you asked for because your account has been
    disabled.
    """

class FuelTanker:
    """I am a tanker full of fuel.
    """
    # TODO: overdraft mode?

    def __init__(self):
        self.fuel = {}
        self.tankSizes = {}

    def addTanks(self, newTanks):
        """Add a dict of new 'full tank' values to this tanker.
        """
        self.tankSizes.update(newTanks)
        self.fuel.update(newTanks)

    def refuel(self):
        """This should be called during the monthly provision update tick.
        """
        # refuel
        self.fuel.clear()
        self.fuel.update(self.tankSizes)

    def deplete(self, tankName, amount, dontRaise=False):
        val = self.fuel[tankName] - amount
        if val <= 0:
            if dontRaise:
                return False
            else:
                raise NotEnoughFuel(tankName, amount)
        else:
            self.fuel[tankName] = val
            return True

class Permissioner:

    gracePeriod = 1
    administrativeSuspension = 0

    def __init__(self):
        self.permissions = {}

    def tick(self):
        for k in self.permissions.keys():
            self.permissions[k] -= 1
            if not self.permissions[k]:
                del self.permissions[k]

    def check(self, permissionName, dontRaise=False):
        if self.permissions.get(permissionName) and not self.administrativeSuspension:
            return True
        elif dontRaise:
            return False
        raise NotEnoughFuel(permissionName)

    def addPermission(self, newPermission, months):
        if self.permissions.has_key(newPermission):
            self.permissions[newPermission] += months
        else:
            self.permissions[newPermission] = months + self.gracePeriod

class FeaturePackage:
    """I am a collection of services that have been pre-paid over some amount of time.
    """
    def __init__(self, months):
        self.months = months
        self.permissions = []
        self.fueltanks = {}

    def addFeature(self, name, amount=0.):
        self.permissions.append(name)
        if amount:
            self.fueltanks[name] = amount

    def grantTo(self, prov):
        """Grant this feature package to a provisioner.
        """
        prov.touch()
        for permission in self.permissions:
            prov.permissioner.addPermission(permission, self.months)
        prov.fueltanker.addTanks(self.fueltanks)

class IProvisionable(Interface):
    property("tankID", doc="""
    Return an identifier for the tank which this object consumes from.
    """)

    property("startCost", doc="""
    Return the amount of fuel consumed from the tank identified by tankID
    before this provisionable will allow itself to be run, instantiated,
    installed, etc.
    """)

class IProvisioner(Interface):
    """Interface spec for Provisioner class.
    """

class Provisioner(Powerup):
    """I am a collection of knowledge about what services the user has paid
    for.

    TODO: alert the user when things are about to run out.
    """

    __implements__ = (Powerup.__implements__, IProvisioner)
    powerupName = "Provisioner"
    invisible = True

    storeDependencies = [sysup.ControlPanel]

    persistenceVersion = 3

    def upgradeToVersion3(self):
        self.touch()
        self.permissioner.addPermission(EPermission.SIP_LOGIN, 12)

    def upgradeToVersion2(self):
        from atop import powerup
        self.touch()
        ps = powerup.IPowerStation(self.store.store)
        ps.installPowerup(sysup.ControlPanel)

    def upgradeToVersion1(self):
        self.touch()
        from quotient.application import Avatar
        avatar = self.store
        avatar.touch()
        avatar.__class__ = Avatar
        avatar.initialize()

    def __init__(self, store):
        Powerup.__init__(self, store)
        self.permissioner = Permissioner()
        self.fueltanker = FuelTanker()

    def setUpPools(self, avatar):
        avatar.setComponent(IProvisioner, self)

    def check(self, featureName, dontRaise=False):
        if not self.enabled:
            if dontRaise:
                return False
            else:
                raise ProvisionerDisabled(featureName)
        result = self.permissioner.check(featureName, dontRaise)
        if result:
            log.msg(**{'permissioner': 'check', 'name': featureName})
        return result

    def deplete(self, featureName, amount, dontRaise=False):
        self.touch()
        cp = sysup.IControlPanel(self.store.store)
        if not cp.actionAllowed(sysup.IControlPanel.PROVISIONING):
            amount = -amount
        if self.check(featureName, dontRaise):
            result = self.fueltanker.deplete(featureName, amount, dontRaise)
            log.msg(provisioner='deplete', amount=amount, name=featureName, avatar=self.store)
            return result
        elif dontRaise:
            return False
        raise NotEnoughFuel(featureName, amount)

    def tick(self):
        """This should be called once per month.
        """
        self.touch()
        self.fueltanker.refuel()
        self.permissioner.tick()
        log.msg(**{'provisioner': 'refuel', 'avatar': self.store})

    enabled = True

    def enable(self):
        self.touch()
        self.enabled = True

    def disable(self):
        self.touch()
        self.enabled = False

    def provision(self, provisionable):
        self.touch()
        return self.deplete(provisionable.tankID, provisionable.startCost, False)

from twisted.protocols import imap4, pop3, smtp
from nevow import inevow
from quotient.proto.twisted_sip import IContact
from twisted.protocols import smtp

interfaceToPermissions = {
    imap4.IAccount: "imap4-login",
    pop3.IMailbox: "pop3-login",
    inevow.IResource: "web-login",
    IContact: "sip-login",
    smtp.IMessageDeliveryFactory: "smtp-send",
}

def provisionRelation(avatar, interface):
    p = IProvisioner(avatar, default=None)
    i = interfaceToPermissions.get(interface, None)
    if p is not None:
        if i is not None:
            p.check(i)
        else:
            log.msg("Warning: allowing unprovisioned interface: %s" % interface)

# Some canonical names.

class EPermission:
    """Collection of names of permissions (features which are set on an on/off
    basis).
    """

    IMAP4_LOGIN = "imap4-login"
    POP3_LOGIN = "pop3-login"
    WEB_LOGIN = "web-login"
    SIP_LOGIN = "sip-login"
    CALENDAR = "calendar"
    ADDRESSBOOK = "addressbook"
    SMTP_SEND = "smtp-send"

class ETank:
    """Collection of names of tanks (features which have a certain amount of
    monthly charge).
    """

    EXTRACTION = "extraction"
    CLASSIFICATION = "classification"
    INDEXING = "indexing"

    QUOTA = "quota"
    OUTGOING_MAIL = "outgoing-mail"
    INCOMING_MAIL = "incoming-mail"

    TELEPHONY_SEND = "outgoing-calls"
    TELEPHONY_RECEIVE = "incoming-calls"

    BLOG_POST = "blog-post"


oneYearService = FeaturePackage(12)
for permissionFeature in (EPermission.WEB_LOGIN,
                          EPermission.IMAP4_LOGIN,
                          EPermission.POP3_LOGIN,
                          EPermission.SIP_LOGIN,
                          EPermission.ADDRESSBOOK,
                          EPermission.SMTP_SEND):
    oneYearService.addFeature(permissionFeature)

for tankFeature in (ETank.EXTRACTION,
                    ETank.CLASSIFICATION,
                    ETank.INDEXING):
    # 20 CPU minutes per month for each feature
    oneYearService.addFeature(tankFeature, 20. * 60.)

for tankFeature in (ETank.QUOTA,
                    ETank.OUTGOING_MAIL,
                    ETank.INCOMING_MAIL):

    # 50 MB plaintext storage per user per month - this is what SMTP should use (it
    # is a soft limit)
    oneYearService.addFeature(tankFeature, 1024 * 1024 * 50)

# 100 outgoing calls, unlimited incoming calls
oneYearService.addFeature(ETank.TELEPHONY_SEND, 100)
oneYearService.addFeature(ETank.TELEPHONY_RECEIVE, 1e100)

# 600 blog posts
oneYearService.addFeature(ETank.BLOG_POST, 600)

class ProvisionedPowerup(Powerup):
    __implements__ = Powerup.__implements__, IComponentForwarder

    dependencies = [Provisioner]

    def _getProvisioner(self):
        return IProvisioner(self.store)

    provisioner = property(_getProvisioner)

    def setUpPools(self, avatar):
        self.provisioner.provision(IProvisionable(self))

    def forwardComponent(self, interface):
        permission = interfaceToPermissions.get(interface)
        if permission is not None:
            if not self.provisioner.check(permission, True):
                failer = self.failedLogin(interface)
                return failer
        return self

    def failedLogin(self, interface):
        raise UnauthorizedLogin("Your account has been disabled.")
