# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import re
import os
import sys
import mailbox

from twisted.python import log
from twisted.python import usage
from twisted.python import util
from twisted.cred import checkers

from atop import store

from quotient import sysup
from quotient import searchup
from quotient import grabbers
from quotient import mimemessage

class Options(usage.Options):
    optParameters = [['datadir', 'd', 'data/db', 'Database directory of service to modify.'],
                     ['filedir', 'f', 'data/files', 'File directory of service to modify.'],
                     ['username', 'u', None, 'The user to create.'],
                     ['password', 'p', None, 'The password to use.'],
                     ['domain', '', 'localhost', 'Domain to add user to.']]

    optFlags = [['ticketed', 'T', 'Create this user as though they had redeemed a ticket.'],
                ['admin', 'A', 'Create this user as an administrator.']]

    def __init__(self, *a, **kw):
        usage.Options.__init__(self, *a, **kw)
        self['directories'] = []
        self['mailboxes'] = []
        self['imap4gs'] = []
        self['pop3gs'] = []

    def opt_directory(self, arg):
        """Import all message files in the given directory.
        """
        self['directories'].append(arg)

    def opt_mailbox(self, arg):
        """Import all messages in the given mbox-formatted file.
        """
        self['mailboxes'].append(arg)

    def opt_imap4g(self, arg):
        """Add an IMAP4 Grabber to this account.  (username:password:host:port)
        """
        self['imap4gs'].append(arg)

    def opt_pop3g(self, arg):
        """Add a POP3 Grabber to this account.  (username:password:host:port)
        """
        self['pop3gs'].append(arg)

    def postOptions(self):
        if not self['username']:
            raise usage.UsageError("No username specified")
        if not self['password']:
            self['password'] = util.getPassword('User password: ', True)

def createUser(st, options):
    u = options['username']
    p = options['password']
    d = options['domain']
    f = not options['ticketed']
    a = options['admin']

    ch = checkers.ICredentialsChecker(st)
    cp = sysup.IControlPanel(st)
    return sysup.makeUser(st, u, p, cp.POWERUPS + ([(sysup.Adminup, True)] * a), f, d)

def importDirectories(avatar, options, skipExt = ('py', 'pyc', 'pyo')):
    md = mimemessage.IMIMEDelivery(avatar)
    for directory in options['directories']:
        directory = os.path.abspath(directory)
        for fn in os.listdir(directory):
            df = os.path.join(directory, fn)
            if not os.path.isfile(df):
                continue
            if fn[:1] == '.' or os.path.splitext(fn)[1][1:] in skipExt:
                continue

            log.msg("Importing %s" % (df,))

            f = open(df)
            #this little hack should get us over any bumps
            #when loading from an MH folder:
            if os.linesep != '\n':
                for line in iter(f.readline,''):
                    if re.match(r'^(\w|-)+: ', line):
                        f.seek(-len(line),1)
                    break

            mp = md.createMIMEReceiver(True, "file://" + directory)
            try:
                m = mp.feedFileNow(f)
            except:
                log.err()

def importMailboxes(avatar, options):
    md = mimemessage.IMIMEDelivery(avatar)
    for m in options['mailboxes']:
        m = os.path.abspath(m)
        log.msg("Importing %s" % (m,))
        mbox = mailbox.PortableUnixMailbox(file(m))
        for (n, msg) in enumerate(mbox):
            mp = md.createMIMEReceiver(True, "mbox://%s#%d" % (m, n))
            mp.feedStringNow(str(msg) + '\r\n' + msg.fp.read())

def addPOP3Grabbers(avatar, options):
    ga = grabbers.IGrabberAdder(avatar)
    for pop3g in options['pop3gs']:
        u, p, h, P = pop3g.split(':')
        ga.addPopGrabber(u, p, h, (int(P), False), '', 0)

def addIMAP4Grabbers(avatar, options):
    ga = grabbers.IGrabberAdder(avatar)
    for imap4g in options['imap4gs']:
        u, p, h, P = imap4g.split(':')
        ga.addIMAPGrabber(u, p, h, (int(P), False), '', 0)

def makeAndImport(st, options):
    avatar = st.transact(createUser, st, options)

    lup = searchup.ISearch(avatar)
    lup.indexer.setMergeFactor(500)

    for f in (importDirectories, importMailboxes):
        f(avatar, options)

    lup.indexer.optimize()
    lup.indexer.close()

    for f in (addPOP3Grabbers, addIMAP4Grabbers):
        st.transact(f, avatar, options)

def main():
    o = Options()
    try:
        o.parseOptions(sys.argv[1:])
    except usage.UsageError, e:
        raise SystemExit(str(e))

    log.startLogging(sys.stdout)

    # We could join here, and some day we will.  For now, the
    # concurrency issues are too numerous for it to be worthwhile.
    st = store.Store(o['datadir'], o['filedir']) # , join=True)
    try:
        makeAndImport(st, o)
    finally:
        st.close()
