# -*- test-case-name: quotient.test.test_stats -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import time
import array

from twisted.python import context

class Statoscope(object):

    def __init__(self, bytes=0, time=0.0, msgs=0, name=""):
        self.name = name
        self.bytes = bytes
        self.time = time
        self.msgs = msgs
        self._mark = None

    def start(self):
        # assert self._mark is None
        self._mark = time.time()

    def finish(self, bytes=0):
        assert self._mark is not None
        self.record(bytes, time.time() - self._mark)
        self._mark = None

    def record(self, bytes=0, time=0):
        # Record the stats for 1 msg
        self.msgs += 1
        self.bytes += bytes
        self.time += time
        s = context.get("stats")
        if s is not None:
            s.update({self.name: (self.bps(), "Bytes / second")})
        self.lastElapsed = time

    def bps(self):
        # Bytes per second
        if self.time == 0:
            return 0.0
        return self.bytes/self.time

    def mps(self):
        # Messages per second
        if self.time == 0:
            return 0.0
        return self.msgs/self.time

    def __add__(self, other):
        self.msgs += other.msgs
        self.time += other.time
        self.bytes += other.bytes
        return self

    def __repr__(self):
        if self.msgs == 0:
            return '<Statoscope: no stats yet>'
        if self.bytes == 0:
            return '<msgs:%d t:%fs mps:%f Bps:-->'%(self.msgs,self.time,self.mps())

        return '<msgs:%d t:%fs mps:%f Bps:%f>'%(self.msgs,self.time,self.mps(),self.bps())

class AveragingResourceMeter:
    def __init__(self, name, maxTime):
        """Create a new resource meter.

        @param name: The name of the resource this meter is monitoring.
        @param times: The maximum age, in seconds, of samples to keep.
        """
        self.name = name
        self.lastTime = int(self.time())
        self.maxTime = maxTime
        self.setupRecords()

    def __getstate__(self):
        d = vars(self).copy()
        del d['records'], d['lastTime']
        return d

    def __setstate__(self, state):
        self.__dict__ = state
        self.setupRecords()

    def setupRecords(self):
        self.lastTime = int(self.time())
        self.records = array.array('i', '\0' * 4 * self.maxTime)

    time = staticmethod(time.time)

    def index(self, when = None):
        if when == None:
            when = self.lastTime
        return when % self.maxTime

    def zeroBetween(self, start, end):
        for p in xrange(start, min(end, start + self.maxTime)):
            i = self.index(p)
            self.records[i] = 0

    def record(self, weight=1):
        lastTime = int(self.time())
        if lastTime != self.lastTime:
            self.zeroBetween(self.lastTime + 1, lastTime)
            self.lastTime = lastTime
            self.records[self.index()] = weight
        else:
            self.records[self.index()] += weight

    def average(self, period):
        """Determine the average usage for a period of time.

        @param period: The number of seconds to calculate the average for.
        """
        if period > self.maxTime:
            raise ValueError("Period of %d is greater than maximum record age %d" % (period, self.maxTime))
        lastTime = int(self.time())
        if lastTime != self.lastTime:
            self.zeroBetween(self.lastTime, lastTime)
            self.lastTime = lastTime
        p = xrange(self.lastTime - period, self.lastTime)
        return sum([self.records[self.index(n)] for n in p]) / float(period)

class ResourceMeter:
    def __init__(self):
        self.resources = {}

    def resourceNames(self):
        return self.resources.keys()

    def use(self, name, amount):
        self.resources[name] = self.resources.get(name, 0) + amount

    def release(self, name, amount):
        self.resources[name] -= amount

    def used(self, name):
        return self.resources.get(name, 0)

def plural(n):
    return round(n, 2) != 1 and "s" or ""

def formatSize(size, suffix='byte', prefixes=("kilo", "mega", "giga", "tera", "peta"), pluralize=True, fractionalPart=True):
    pfxs = [''] + list(prefixes)
    step = float(1024)
    s = size
    while s >= step and pfxs:
        s /= step
        pfxs.pop(0)
    if pfxs:
        if fractionalPart:
            start = '%0.02f' % (s,)
        else:
            s = int(round(s))
            start = '%d' % (s,)
        end = "%s%s%s" % (pfxs[0], suffix, plural(s) * pluralize)
        return ' '.join((start, end))
    return "a whole lot of %s%s (%d)" % (suffix, 's' * pluralize, size)
