
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

"""
System Storeup for handling required tasks such as sending bounces,
password reset information, etc.
"""

import gc
import os.path

from atop import credup
from atop import powerup
from atop.sched import SubScheduler
from atop.service import Serviceup

from twisted.cred.checkers import ICredentialsChecker
from twisted.python import components
from twisted.python import log
from twisted.python import reflect
from twisted.application.service import IService, IServiceCollection

from quotient import stats
from quotient.prefup import Prefup

from quotient.stats import formatSize

class SignupDisallowed(Exception):
    pass

class IllegalUsername(SignupDisallowed):
    pass

class AdministrativeRestriction(SignupDisallowed):
    pass

def makeUser(store, username, password, powerups, free, domainName=None):
    from quotient import provisioner
    from quotient import application
    if not username:
        raise IllegalUsername("Please enter a string")

    if len(username) < 3 or '.' in username:
        raise IllegalUsername(username)

    # XXX This part sucks
    if free:
        tag = IControlPanel.USER_CREATION
    else:
        tag = IControlPanel.TICKET_REDEMPTION

    if not IControlPanel(store).actionAllowed(tag):
        raise AdministrativeRestriction("Signup is currently disabled")

    cc = ICredentialsChecker(store)
    a = application.Avatar(store, username)
    ps = powerup.IPowerStation(a)
    pv = ps.installPowerup(provisioner.Provisioner)
    provisioner.oneYearService.grantTo(pv)

    prefs = ps.installPowerup(Prefup)
    for p, visible in powerups:
        pup = ps.installPowerup(p)
        if visible:
            prefs.makeVisible(pup)
    a.setInsecurePassword(password)
    a.setSecurePassword(password)
    personalize(a, username)
    cc.addAvatar(a, domainName)
    return a

def personalize(store, username):
    from quotient.smarts.person import IPersonCollector, Person

    pc = IPersonCollector(store, None)
    if pc is None:
        return
    from quotient.addressbook import IAddressBook
    from quotient.prefup import IPrefs
    person = IPersonCollector(store).newPerson(username)
    IPrefs(store).setMe(person)
    # Add me to the address book
    IAddressBook(store).addPerson(person)

def makeSystemUsers(store):
    from quotient import provisioner
    from quotient import application
    from quotient.reset import PasswordResetter
    from quotient.proto.smtpout import OutgoingMailup

    systemUsers = [
        ("postmaster", (SubScheduler, OutgoingMailup)),
        ("janitor", (PasswordResetter,))
        ]

    c = ICredentialsChecker(store)
    for name, pups in systemUsers:
        a = application.Avatar(store, name)
        ps = powerup.IPowerStation(a)
        ps.installPowerup(provisioner.Provisioner)
        provisioner.oneYearService.grantTo(provisioner.IProvisioner(a))
        c.addAvatar(a)
        map(powerup.IPowerStation(a).installPowerup, pups)

class IControlPanel(components.Interface):
    USER_CREATION = 'user-creation'
    TICKET_REDEMPTION = 'ticket-redemption'

    MESSAGE_INDEXING = 'disable-indexing'
    EXTERNAL_GRABBING = 'disable-grabbing'

    SPAMBAYES_TRAINING = 'disable-spambayes'

    PROVISIONING = 'provisioning'

    def allowAction(self, action, yesno):
        pass

    def actionAllowed(self, action):
        pass

class ConvenientHackRemoveBeforeOnePointOh(dict):
    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            self[key] = stats.AveragingResourceMeter(key, 60 * 15)
            return self[key]

class ControlPanel(powerup.Storeup):
    __implements__ = (powerup.Storeup.__implements__, IControlPanel, IService)

    dependencies = [Serviceup]

    RESOURCE_METERS = ('transaction-commits', 'transaction-aborts',
                       'pickled', 'unpickled', 'cache-hits',
                       'page-renders', 'authorized-logins',
                       'unauthorized-logins', 'messages',
                       'bytes')

    SERIALIZED_LENGTH_WARNING = 8192

    running = False

    persistenceVersion = 1

    def upgradeToVersion1(self):
        self.touch()
        s = powerup.IPowerStation(self.store).installPowerup(Serviceup)
        s.addService(self)

    def setUpPools(self, store):
        from puplist import POWERUPS
        self.POWERUPS = POWERUPS

        self.meters = ConvenientHackRemoveBeforeOnePointOh()
        for n in self.RESOURCE_METERS:
            self.meters[n] = stats.AveragingResourceMeter(n, 60 * 15)

        self.actions = {IControlPanel.USER_CREATION: True,
                        IControlPanel.TICKET_REDEMPTION: True,
                        IControlPanel.MESSAGE_INDEXING: True,
                        IControlPanel.EXTERNAL_GRABBING: True,
                        IControlPanel.SPAMBAYES_TRAINING: True,
                        IControlPanel.PROVISIONING: False}
        store.setComponent(IControlPanel, self)
        IServiceCollection(self.store).addService(self)

    def privilegedStartService(self):
        raise AssertionError("privilegedStartService *cannot* be called!")

    def startService(self):
        self.touch()
        self.running = True
        log.addObserver(self.observe)

    def stopService(self):
        self.touch()
        self.running = False
        log.removeObserver(self.observe)

    def observe(self, events):
        # This might be optimized by adding a "channel" key
        # which could be dispatched upon in O(1) time

        if events['message']:
            return

        if 'db_put' in events:
            L = len(events['pickle'])
            if L > self.SERIALIZED_LENGTH_WARNING:
                log.msg("Warning: %-6d bytes pickled from %s" % (L, events['item']))
            self.meters['pickled'].record(L)
            return

        if 'db_get' in events:
            L = len(events['pickle'])
            if L > self.SERIALIZED_LENGTH_WARNING:
                log.msg("Warning: %-6d bytes unpickled from %s" % (L, events['item']))
            self.meters['unpickled'].record(L)
            return

        if 'db_txn' in events:
            if events['db_txn'] == 'commit':
                self.meters['transaction-commits'].record()
            elif events['db_txn'] == 'abort':
                self.meters['transaction-aborts'].record()
            return

        if 'db_cache' in events:
            if events['db_cache'] == 'hit':
                self.meters['cache-hits'].record()
            return

        if 'login' in events:
            if events['login'] is None:
                self.meters['unauthorized-logins'].record()
            else:
                self.meters['authorized-logins'].record()
            return

        if 'storage' in events:
            if events['storage'] == 'used':
                self.meters['bytes'].record(events['amount'])
                self.meters['messages'].record()

        if 'http_render' in events:
            self.meters['page-renders'].record()
            return

    def allowAction(self, action, yesno):
        self.touch()
        self.actions[action] = yesno

    def actionAllowed(self, action):
        return self.actions.setdefault(action, False)

class ISystemAdministrator(components.Interface):
    def archive(self):
        pass

    def shutdown(self):
        pass

class Adminup(powerup.Powerup):
    # Mostly, but not entirely, just a stub
    # Allows the ControlPanel for the Store to be rendered

    __implements__ = (powerup.Powerup.__implements__, ISystemAdministrator,)

    powerupName = "Administrative Controls"

    storeDependencies = [ControlPanel]
    dependencies = [SubScheduler]

    persistenceVersion = 1

    def upgradeToVersion1(self):
        self.touch()
        powerup.IPowerStation(self.store).installPowerup(SubScheduler)

    def setUpPools(self, avatar):
        avatar.setComponent(ISystemAdministrator, self)

    def archive(self):
        self.store.store.archive()

    def shutdown(self):
        from twisted.internet import reactor
        reactor.callLater(1, reactor.stop)


from nevow import tags
import formless
from formless import webform

from quotient import taglib
from quotient.web.root import Base

class ISystemSettings(formless.TypedInterface):
    allowUserCreation = formless.Boolean(label="New Users",
                                         description="Allow new users to register without claiming a ticket")
    allowTicketRedemption = formless.Boolean(label="Ticket Redemption",
                                             description="Allow new users to register by claiming a ticket")

    allowMessageIndexing = formless.Boolean(label="Fulltext Indexing",
                                            description="Allow newly received messages to be indexed.")
    allowExternalGrabbing = formless.Boolean(label="External Account Grabbing",
                                             description="Allow messages to be retrieved from external accounts.")
    enableProvisioner = formless.Boolean(label="Provisioning",
                                         description="Restrict activities based on provisioned limits.")

    allowSpambayes = formless.Boolean(label="Spambayes",
                                             description="Enable Bayesan filtering of mail.")

class IAdminDisplay(formless.TypedInterface):
    SystemSettings = ISystemSettings

    enableProtocolDebugging = formless.Boolean(label="Proto Debugging",
                                               description="Log POP3, IMAP4, and SMTP traffic")

#    def reloadPupList(self):
#        "Reload the master powerup list for new users from quotient.puplist."
#    reloadPupList = formless.autocallable(reloadPupList)


    def enablePowerup(self, name = formless.String()):
        """Add the named powerup to the list of those given to new users.
        """
    enablePowerup = formless.autocallable(enablePowerup)

    def disablePowerup(self, name = formless.String()):
        """Remove the named powerup from the lsit of those given to new users.
        """
    disablePowerup = formless.autocallable(disablePowerup)

    def installExtractor(self, extractPowerup = formless.String(),
                         user = formless.String()):
        """Install the given extractor on the given user, or all users if the user is not given.
        """
    installExtractor = formless.autocallable(installExtractor)

    def startProfiler(self):
        """Start profiling with the specified profiler.
        """
    startProfiler = formless.autocallable(startProfiler)

    def stopProfiler(self):
        """Deactivate the profiler.
        """
    stopProfiler = formless.autocallable(stopProfiler)

    def shutdown(self):
        """Shut the server down"""
    shutdown = formless.autocallable(shutdown)

    def archive(self):
        """Archive database logfiles"""
    archive = formless.autocallable(archive)

    def grantTickets(self,
                     username = formless.String(),
                     numberToGrant = formless.Integer()):
        """Grant tickets to a user
        """
    grantTickets = formless.autocallable(grantTickets)

    def toggleSuspension(self, username = formless.String()):
        """Toggle administrative suspension for a user.
        """
    toggleSuspension = formless.autocallable(toggleSuspension)

    def regenerateNavigation(self, username = formless.String()):
        """Clear and recalculate the navigation UI for the given user.
        """
    regenerateNavigation = formless.autocallable(regenerateNavigation)

    def resetPassword(self, username = formless.String(), password = formless.Password()):
        """Set the given user's password to the given string.
        """
    resetPassword = formless.autocallable(resetPassword)

def renderMeter(name, period, format):
    def render(self, context, data):
        m = IControlPanel(self.original.store.store).meters[name]
        return format(m.average(period))
    return render

def makeResourceRow(name, format='%4.2f'.__mod__):
    f = lambda i: tags.td[renderMeter(name, i, format)]
    return tags.tr[tags.td[name], map(f, (60, 60 * 5, 60 * 15))]

class AdminDisplay(Base):
    __implements__ = (IAdminDisplay, Base.__implements__)

    def render_garbage(self, context, data):
        return tags.p["Number of Garbage Objects:", len(gc.garbage)]

    def render_newUserPowerups(self, context, data):
        cp = IControlPanel(self.original.store.store)
        return tags.p[
            "Powerups installed on new users:",
            ', '.join([reflect.qual(p) for (p, v) in cp.POWERUPS])
            ]

    document = taglib.page(title="Administrative Controls")[
        render_garbage,
        render_newUserPowerups,
        webform.renderForms(),
        tags.table()[tags.tr[tags.td["Resource Usage Averages"],
                             tags.td["1 min"], tags.td["5 min"], tags.td["15 min"]
                             ],
                     makeResourceRow("transaction-commits"),
                     makeResourceRow("transaction-aborts"),
                     makeResourceRow("pickled", formatSize),
                     makeResourceRow("unpickled", formatSize),
                     makeResourceRow("cache-hits"),
                     makeResourceRow("page-renders"),
                     makeResourceRow("authorized-logins"),
                     makeResourceRow("unauthorized-logins"),
                     makeResourceRow("messages"),
                     makeResourceRow("bytes", formatSize),
                     ]
        ]

    def action(tag):
        def get(self):
            cp = IControlPanel(self.original.store.store)
            return cp.actionAllowed(tag)
        def set(self, value):
            cp = IControlPanel(self.original.store.store)
            return cp.allowAction(tag, value)
        return get, set

    allowUserCreation = property(*action(IControlPanel.USER_CREATION))
    allowTicketRedemption = property(*action(IControlPanel.TICKET_REDEMPTION))

    allowMessageIndexing = property(*action(IControlPanel.MESSAGE_INDEXING))
    allowExternalGrabbing = property(*action(IControlPanel.EXTERNAL_GRABBING))

    allowSpambayes = property(*action(IControlPanel.SPAMBAYES_TRAINING))
    enableProvisioner = property(*action(IControlPanel.PROVISIONING))

    def enableProtocolDebugging():
        def get(self):
            from quotient.proto import smtpin
            return smtpin.ESMTPInFactory.debug
        def set(self, value):
            from quotient.proto import smtpin, imapup, popup
            smtpin.ESMTPInFactory.debug = value
            imapup.IMAP4Factory.debug = value
            popup.POP3Factory.debug = value
        return get, set
    enableProtocolDebugging = property(*enableProtocolDebugging())

    def shutdown(self):
        return ISystemAdministrator(self.original).shutdown()

    def archive(self):
        return ISystemAdministrator(self.original).archive()

    def reloadPupList(self):
        from quotient.puplist import POWERUPS
        cp = IControlPanel(self.original.store.store)
        cp.POWERUPS = POWERUPS
        cp.touch()
        return "POWERUPS list reloaded."

    def enablePowerup(self, name):
        from quotient.puplist import VISIBLE

        cp = IControlPanel(self.original.store.store)
        try:
            pup = reflect.namedAny(name)
        except:
            return "Unknown Powerup!  Use the fully qualified class name."
        else:
            try:
                if not issubclass(pup, powerup.Powerup):
                    raise ValueError
            except:
                return "That's not a powerup.  You're the admin.  Stop trying to hack stuff!"
            else:
                if pup is Adminup:
                    return "Cannot specify /that/ powerup."

                if [p for (p, v) in cp.POWERUPS if p is pup]:
                    return "That powerup is already installed."
                cp.touch()
                cp.POWERUPS.insert(0, (pup, pup in VISIBLE))
                return "Added %s" % (pup,)

    def disablePowerup(self, name):
        cp = IControlPanel(self.original.store.store)
        try:
            pup = reflect.namedAny(name)
        except:
            return "Unknown Powerup!  Use the fully qualified class name."
        else:
            try:
                if not issubclass(pup, powerup.Powerup):
                    raise ValueError
            except:
                return "That's not a powerup!"
            else:
                for (p, v) in cp.POWERUPS:
                    if p is pup:
                        cp.touch()
                        cp.POWERUPS.remove((p, v))
                        return "Removed %s" % (p,)
                return "That powerup isn't being installed currently."

    def installExtractor(self, extractPowerup, user):
        from quotient.smartenup import retroactivelyExtract
        try:
            pup = reflect.namedAny(extractPowerup)
        except Exception, e:
            return "Error finding extract powerup: " + str(e)
        else:
            from twisted.internet import reactor
            cc = ICredentialsChecker(self.original.store.store)
            if user:
                try:
                    av = cc.getAvatar(user)
                except Exception, e:
                    return "Error finding user: " + str(e)
                else:
                    reactor.callLater(0, self.original.store.transact,
                                      retroactivelyExtract,
                                      av, pup)
                    return 'Extraction initiated'
            else:
                for av in cc.getAvatars():
                    reactor.callLater(0, self.original.store.transact,
                                      retroactivelyExtract,
                                      av, pup)
                return "Extraction initiated"

    def startProfiler(self):
        # Mmm, globals are bad, but profilers are global
        global profiler

        if profiler is not None:
            return "A profiler is already running."

        from atop.tpython import HotshotProfilerService as Profiler
        profiler = Profiler()
        profiler.start()
        return "Profiling."

    def stopProfiler(self):
        global profiler

        if profiler is None:
            return "No profiler is running."

        def pickFileName():
            i = 0
            while os.path.exists("profile.%d" % i):
                i += 1
            return "profile.%d" % i

        profiler.stop()
        profiler.save(file(pickFileName(), 'wb'))
        profiler = None

        return "Profiler stopped.  Profile written."

    def grantTickets(self, username, numberToGrant):
        from quotient.signup import ITicketIssuer
        from atop.credup import ICredentialsChecker
        cc = ICredentialsChecker(self.original.store.store)
        ti = ITicketIssuer(cc.getAvatar(username))
        ti.remainingTicketCount += numberToGrant
        return "Granted %r %s tickets, leaving them with a total of %s." % (
            username, numberToGrant, ti.remainingTicketCount)

    def toggleSuspension(self, username):
        from atop.credup import ICredentialsChecker
        from quotient.provisioner import IProvisioner
        cc = ICredentialsChecker(self.original.store.store)
        prov = IProvisioner(cc.getAvatar(username))
        perm = prov.permissioner
        prov.touch()
        perm.administrativeSuspension = not perm.administrativeSuspension
        return 'suspension ' + (perm.administrativeSuspension and 'enacted' or 'removed')

    def regenerateNavigation(self, username):
        from atop.credup import ICredentialsChecker
        from quotient.configure import IConfigure
        from quotient.prefup import IPrefs
        cc = ICredentialsChecker(self.original.store.store)
        av = cc.getAvatar(username)
        IPrefs(av).emptyPrimaryNav()
        IConfigure(av).setupPrimaryNav()

    def resetPassword(self, username, password):
        from atop.credup import ICredentialsChecker
        cc = ICredentialsChecker(self.original.store.store)
        av = cc.getAvatar(username)
        av.setInsecurePassword(password)
        av.setSecurePassword(password)
        return "Password reset"

profiler = None


class ISysup(components.Interface):
    pass

class Sysup(powerup.Storeup):
    __implements__ = (ISysup,)

    dependencies = [ControlPanel]

    def setUpPools(self, store):
        makeSystemUsers(store)
        store.setComponent(ISysup, self)
