# -*- test-case-name: "quotient.test.test_functional" -*-
import re, os
import os.path as osp
from os.path import join as opj

from twisted.trial import unittest
from twisted.trial.unittest import FailTest
from twisted.python import log
from twisted.web.microdom import parseString
from twisted.web.domhelpers import findElementsWithAttribute as findwa
from twisted.web.domhelpers import findNodes as findn
from twisted.web.domhelpers import findNodesNamed as findnn
from twisted.web.domhelpers import gatherTextNodes as gather

from quotient.functional import unittest as funit

from quotient.functional.client import Client, IPage 
from quotient.functional.navtags import *
from quotient.functional import navtags


linkdom = "<a id='test.Link' href='http://www.example.com'>Testing Link</a>"

noIdLinkDom = "<a href='http://www.example.com'>I don't have an id attribute!</a>"

formdom = """
<form action="/__login__/me/" method="POST" id="signin">
    <input type="text" name="username" id="username" />
    <input type="password" name="password" id="password" />
    <input type="submit" id="login" value="login" />
</form>
"""

__debug = False

def debugObserver(events):
    if __debug and 'functional' in events:
        print ' '.join(["functional: "] + list(events['functional']))
log.addObserver(debugObserver)

def _dmesg(self, msg):
    log.msg(functional=msg)


class BogusPage(object):
    __implements__ = IPage,
    followCalled = False
    client = None
    domstr_got, domstr_set = False, False

    _domstr = None

    def getLink(self, id):
        self.getLinkId = id

    def follow(self):
        followCalled = True
        return self
        
    def getForm(self, id):
        self.getFormId = id
        return self

    def submit(self, adict):
        self.submitDict = adict

    def _getDomstr(self):
        self.domstr_got = True
        return self._domstr

    def _setDomstr(self, val):
        self.domstr_set = True
        self._domstr = val
    
    domstr = property(_getDomstr, _setDomstr)

class BogusClient(object):
    got, page = None, None
    def get(self, *a):
        self.got = a

class TestFunctionalTestHarness(unittest.TestCase):
    def setUp(self):
        self.client, self.page = BogusClient(), BogusPage()
        self.client.page = self.page
        self.page.client = self.client

    def test_getIdAttr(self):
        class BogusContext:
            pass
        context = BogusContext()
        context.key = 'UseThis.ignore'
        name = 'a_name'
        a = funit.getIdAttr(context, name)
        self.failUnlessEqual(a, 'UseThis.a_name')

    def test_Link(self):
        from quotient.functional.client import Link, IdNotFoundError
        doc = parseString(linkdom)
        node = findnn(doc,'a')[0]

        for attr in ['id', 'href']:
            self.assert_(node.hasAttribute(attr))

        L = Link(node)
        fueq = self.failUnlessEqual

        fueq(L.id, 'test.Link')
        fueq(L.url, 'http://www.example.com')
        fueq(L.name, 'Testing Link')
        fueq(L.page, None)
        self.failUnlessRaises(RuntimeError, L.follow)

        L = Link(node, self.client.page)
        self.failUnlessIdentical(L.page, self.client.page)
        L.follow()
        fueq(L.page.client.got, ('http://www.example.com',))

        # now test for exceptional condition where link found doesn't 
        # come with an id attribute
               
        doc = parseString(noIdLinkDom)
        node = findnn(doc, 'a')[0]
        self.failUnlessRaises(IdNotFoundError, Link, node)

    def test_Form(self):
        doc = parseString(formdom)
        node = findnn(doc,'form')[0]

        for attr in ['action','method','id']:
            self.assert_(node.hasAttribute(attr))

        from quotient.functional.client import Form
        f = Form(node)
        fueq = self.failUnlessEqual

        fueq(f.id, 'signin')
        fueq(f.action, '/__login__/me/')
        fueq(f.page, None)
        self.failUnlessRaises(RuntimeError, f.submit)
        
        f = Form(node, self.page)
        args = {'divmod': 'k-lame', 'everything else': 'k-rad'}
        f.submit(args)
        fueq(f.page.client.got, (f.action, args))

    

class TestNavTags(unittest.TestCase):
    def setUp(self):
        global client
        client = BogusClient()
        
    def runSeq(self, seq, cb):
        self.seq = seq
        self.cb = cb
        navtags.runSeq(seq, cb)

class BogusUrlTrail(object):
    def __init__(self):
        import tempfile 
        ntmpFile = tempfile.NamedTemporaryFile
        self._t1, self._t2 = ntmpFile(), ntmpFile() 
        self.path, self.filename = self._t1.name, self._t2.name

class TestAssertTags(unittest.TestCase):
    def setUp(self):
        import quotient 
        quotient.functional.client._theClient = c = BogusClient()
        quotient.functional.client.page = p = BogusPage()
        self.failUnlessIdentical(quotient.functional.client.page, p)
        quotient.functional.client.page.domstr = formdom
        self.assert_(quotient.functional.client.page.domstr_set)
        quotient.functional.client.page.client = c
        quotient.functional.client.urlTrail = but = BogusUrlTrail()
        self.assert_(but.path)
        self.assert_(but.filename)

    def tearDown(self):
        import quotient 
        quotient.functional.client._theClient = quotient.functional.client.Client()
        quotient.functional.client.page = None

    def testAssertTags(self):

                          #([synonyms], ([failingCaseArgs], [passingCaseArgs]))
        for tags, args in [([failIf], ([True], [False])),
                           ([failUnless, assert_], ([False], [True])),
                           ([failUnlessEqual, assertEqual, assertEquals], ([1, 2], [1,1])),
                           ([failUnlessIdentical, assertIdentical], ([None, 'this is not identical'], [None, None])),
                           ([assertNotIdentical, failIfIdentical], ([None, None], [None, 'not identical'])),
                           ([failIfEqual, assertNotEqual, assertNotEquals], ([1, 1], [1,2])),
                           ([failUnlessIn, assertIn], (['it', 'you must not say the word!'], ['it', 'this is it'])),
                           ([failIfIn, assertNotIn], (['it', 'we were just looking for it in this forest,'], ['it', 'nope'])),
                           ]:
            try:
                for synonym in tags:
                    negCaseArgs, posCaseArgs = args
                    self.failUnlessRaises(FailTest, synonym(*negCaseArgs).doYourThing)
                    synonym(*posCaseArgs).doYourThing()
            except FailTest:
                print '%s did not raise FailTest' % synonym
                raise


    def testPageAssertTags(self):
        for tags, args in [([assertString, failUnlessString], (['applejuice'], ['password'])),
                           ([assertNotString, failIfString], (['password'], ['applejuice']))]:

            try:
                for synonym in tags:
                    negCaseArgs, posCaseArgs = args
                    self.failUnlessRaises(FailTest, synonym(*negCaseArgs).doYourThing)
                    synonym(*posCaseArgs).doYourThing()
            except FailTest:
                print '%s did not raise FailTest' % synonym
                import quotient
                if not quotient.functional.client.page.domstr_got:
                    print "domstr wasn't touched!!"
                raise


class FakeThing(object):
    __implements__ = IDoMyThing,
    didMyThing, value = None, None

    def __init__(self, value=0, parent=''):
        self.value, self.parent = value, parent

    def doYourThing(self):
        self.didMyThing = True

    def __repr__(self):
        return "<FakeThing at %s  value: %s  parent: %s>" % (hex(id(self)), self.value, self.parent)

dummyeb = lambda e: e

class TestRunSeq(unittest.TestCase):

    def setUp(self):
        from quotient.functional.navtags import runSeq
        self.runSeq = runSeq
        self.numCbHits = 0

    def cbCount(self, val):
        self.numCbHits += 1
##         print "cbCount hit!"
        return val


    def testPerIterCb(self):
        s = [FakeThing(), FakeThing(), FakeThing()]
        self.runSeq(s, (self.cbCount, dummyeb))
        self.failUnlessEqual(self.numCbHits, len(s), 'number of cb hits != len sequence')
        for x in s:
            self.assert_(x.didMyThing, "you didn't do your thing!+")

    def testGeneratorSimple(self):
        numYields = 5

        def gimmieIter():
            for n in xrange(numYields):
                yield FakeThing(n)

        self.runSeq([gimmieIter()], (self.cbCount, dummyeb))
        self.failUnlessEqual(self.numCbHits, numYields + 1)


    def testGeneratorInList(self):
        numIters, numYields = 3, 5
        fts = []
        self.currentNum = 0

        def gimmieIter():
            for n in xrange(numYields):
                fts.append(FakeThing(self.currentNum))
                self.currentNum += 1
                yield fts[-1]

        self.runSeq([gimmieIter() for n in xrange(numIters)], (self.cbCount, dummyeb))

        self.failUnlessEqual(self.numCbHits, numYields * numIters + numIters)
        for i, ft in enumerate(fts):
            self.failUnlessEqual(ft.value, i)


    def testNestedGenerators(self):
        outerYields, innerYields = 5, 3

        fts = []
        self.letter, self.subltr = 'A', ''

        def gimmieMetaIter():
            for x in xrange(outerYields):
                subltr = chr(ord(self.letter) + x)

                def gimmieIter(sl):
                    for y in xrange(innerYields):
                        fts.append(FakeThing(y, sl))
                        yield fts[-1]

                yield gimmieIter(subltr)

        self.runSeq([gimmieMetaIter()], (self.cbCount,dummyeb))
        log.msg(navtags='fts: %s' % str(fts))

        count = 0
        for a in xrange(outerYields):
            letter = chr(ord(self.letter) + a)
            for b in xrange(innerYields):
                self.failUnlessEqual(fts[count].parent, letter)
                self.failUnlessEqual(fts[count].value, b)
                count += 1


class TestUrlTrail(unittest.TestCase):
    def setUpClass(self):
        self.dir = self.mktemp()
        os.mkdir(self.dir)
        self.assert_(osp.exists(self.dir))


class TestChildProcessSpawnage(unittest.TestCase):
    def testSpawnageTimeout(self):
        """make sure that if we can't spawn a child process, the tests die correctly"""
        from twisted.python import failure
        from twisted.internet import reactor, defer
        from quotient.functional import unittest as funit
        funit._childTimeout_debug = False # print (or don't) child timeout debugging messages

        tc = funit.TestCase()

        tdir = self.mktemp()
        os.mkdir(tdir)
        tacfile = opj(tdir, 'quotient.tac')
        fp = file(tacfile, 'w')
        fp.write('o/` lalala I am an invalid tac file! lalala o/`'); fp.flush(); fp.close()
        
        tc.tdir, tc.tacfile = tdir, tacfile
        tc.childProcessSetupTimeout = 0

        def cb(val):
            log.msg(childTimeout='cb called with val: %s' % val)
            return val

        def eb(err):
            log.msg(childTimeout='eb called, err: %s' % err)
            err.trap(unittest.FailTest)

        d = tc._spawnChildProcess().addCallbacks(cb, eb)
        unittest.deferredResult(d)


