# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import datetime

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from atop import store
from atop import sched
from atop import credup
from atop.test import unittest

from twisted.protocols import imap4
from twisted.cred.portal import Portal, IRealm
from twisted.cred.checkers import ICredentialsChecker
from twisted.internet import defer

from quotient.test.emaildata import test_messages
from quotient.test import helper

from quotient import mimemessage
from quotient.proto import imapup
from quotient import grabbers
from quotient import searchup
from quotient import cleanup
from quotient import smartenup
from quotient.application import IMailApp

class MsgSet(tuple):
    pass

class Loopback(unittest.TestCase, helper.ProvisionerMixin):

    skip = "I have no idea why this still hangs."

    storePowerups = (sched.SystemScheduler, credup.StoreRealm,)

    def setUpClass(self):
        grabbers.IMAP4MessageGrabber.allowInsecureLogin = True

    def tearDownClass(self):
        grabbers.IMAP4MessageGrabber.allowInsecureLogin = False

    def dbSetUp(self):
        self.avatar = self.createAvatar([imapup.IMAPAccountUp, grabbers.GrabberPowerup, sched.SubScheduler], name="USAR")
        self.avatar.setInsecurePassword("passWORD")
        self.avatar.setSecurePassword("PASSword")

        self.msg_strings = test_messages

        for m in self.msg_strings:
            mr = mimemessage.IMIMEDelivery(self.avatar)
            mr.createMIMEReceiver(False).feedStringNow(m)

    def dbTearDown(self):
        # nc = sched.IScheduler(self.store).nextCall
        # nc and nc.cancel()
        nc = getattr(searchup.ISearch(self.avatar, default=None), 'checkpointDefer', None)
        nc and nc.cancel()

    def dbtestWholeThing(self):
        p = Portal(IRealm(self.store))
        cc = ICredentialsChecker(self.store)
        cc.addDomain("localhost")
        cc.addAvatar(self.avatar, "localhost")
        p.registerChecker(cc)

        sFactory = imapup.IMAP4Factory(p)
        sFactory.debug = True
        from twisted.internet import reactor
        p = reactor.listenTCP(0, sFactory, interface="127.0.0.1")
        gp = grabbers.IGrabberAdder(self.avatar)
        cFactory = gp.addIMAPGrabber("USAR", "passWORD", "127.0.0.1", (p.getHost()[2], False), "USAR@localhost", 0, False)

        val = unittest.deferredResult(cFactory.run())

        inbox = cleanup.IClean(self.avatar).cleanPool
        msgs = list(inbox.queryIndex('uid', result=store.ITEM))
        self.assertEquals(len(msgs), len(self.msg_strings) * 2)
        half = len(msgs) / 2

        begin, end = msgs[:half], msgs[half:]

        origs = []
        for x in begin:
            f = x.openFile()
            f.seek(x.bodyOffset)
            origs.append(f.read())
        news = []
        for x in end:
            f = x.openFile()
            f.seek(x.bodyOffset)
            news.append(f.read())
        for m in zip(origs, news):
            self.assertEquals(*m)


class Account(unittest.TestCase, helper.ProvisionerMixin):
    storePowerups = [sched.SystemScheduler]

    def dbSetUp(self):
        pups = (mimemessage.EmailPowerup, imapup.IMAPAccountUp)
        self.avatar = self.createAvatar(pups, name="Usar")

    def dbTearDown(self):
        # nc = sched.IScheduler(self.store).nextCall
        # nc and nc.cancel()
        nc = getattr(searchup.ISearch(self.avatar, default=None), 'checkpointDefer', None)
        nc and nc.cancel()

    def testLogout(self):
        acc = imap4.IAccount(self.avatar)
        acc.logout()

    def testCreate(self):
        acc = imap4.IAccount(self.avatar)

        self.assertRaises(imap4.MailboxCollision, acc.create, "inbox")
        self.failUnless(acc.create("foo"))

        def getfoo():
            return IMailApp(self.avatar).appRoot.getIndexItem("name", "foo")
        p = self.store.transact(getfoo)
        self.failUnless(isinstance(p, store.Pool))

    def testAddMailbox(self):
        acc = imap4.IAccount(self.avatar)
        self.assertRaises(RuntimeError, acc.addMailbox, "foo", "bar")

    def testSelect(self):
        acc = imap4.IAccount(self.avatar)
        arr = cleanup.IClean(self.avatar).cleanPool

        inboxDeferred = defer.maybeDeferred(acc.select, "inbox")
        inbox = unittest.deferredResult(inboxDeferred)

        self.assertIdentical(inbox.original, arr)

        spamDeferred = defer.maybeDeferred(acc.select, "spam")
        spam = unittest.deferredResult(spamDeferred)
        self.assertEquals(spam.original.name, "Spam")

        randomDeferred = defer.maybeDeferred(acc.select, "adoijwwodij")
        random = unittest.deferredResult(randomDeferred)
        self.assertEquals(random, None)

    def testDelete(self):
        acc = imap4.IAccount(self.avatar)
        self.assertRaises(imap4.MailboxException, acc.delete, "inbox")

    def testRename(self):
        acc = imap4.IAccount(self.avatar)
        self.assertRaises(imap4.MailboxException, acc.rename, "inbox", "foo")

    def testSubscriptions(self):
        acc = imap4.IAccount(self.avatar)

        self.failIf(acc.isSubscribed("foo"))

        self.avatar.transact(acc.subscribe, "foo")

        self.failUnless(acc.isSubscribed("Foo"))
        self.failUnless(acc.isSubscribed("foo"))

        self.avatar.transact(acc.unsubscribe, "Foo")

        self.failIf(acc.isSubscribed("Foo"))
        self.failIf(acc.isSubscribed("foo"))

        self.assertRaises(imap4.MailboxException, self.avatar.transact, acc.unsubscribe, "foo")

    def testListing(self):
        acc = imap4.IAccount(self.avatar)

        x = unittest.deferredResult(acc.listMailboxes("", "i%ox"))
        self.assertEquals([m[0] for m in x], ["Inbox"])
        x = unittest.deferredResult(acc.listMailboxes("", "in%y"))
        self.assertEquals(x, [])

    def testNamespaces(self):
        acc = imap4.IAccount(self.avatar)

        self.assertEquals(acc.getPersonalNamespaces(), [["", None]])
        self.assertEquals(acc.getSharedNamespaces(), None)
        self.assertEquals(acc.getUserNamespaces(), None)

class Mailbox(unittest.TestCase, helper.ProvisionerMixin):
    storePowerups = [sched.SystemScheduler]

    def dbSetUp(self):
        pups = (mimemessage.EmailPowerup, imapup.IMAPAccountUp)
        self.avatar = self.createAvatar(pups, name="Usar")

    def dbTearDown(self):
        # nc = sched.IScheduler(self.store).nextCall
        # nc and nc.cancel()
        nc = getattr(searchup.ISearch(self.avatar, default=None), 'checkpointDefer', None)
        nc and nc.cancel()

    def dbtestUIDValidity(self):
        p = imapup.deferredOpenPool(self.avatar, cleanup.IClean(self.avatar).cleanPool, False)
        p = unittest.deferredResult(p)
        u1 = p.getUIDValidity()

        p = imapup.deferredOpenPool(self.avatar, cleanup.IClean(self.avatar).cleanPool, False)
        p = unittest.deferredResult(p)
        u2 = p.getUIDValidity()

        self.assertEquals(u1, u2)

    def dbtestUIDNext(self):
        p = imapup.deferredOpenPool(self.avatar, cleanup.IClean(self.avatar).cleanPool, False)
        p = unittest.deferredResult(p)

        n = p.getUIDNext()
        self.assertEquals(n, p.getUIDNext())
        self.assertEquals(n, p.getUIDNext())
        self.assertEquals(n, p.getUIDNext())

        unittest.deferredResult(p.addMessage(StringIO("Hello!")))

        self.assertEquals(n + 1, p.getUIDNext())

    def dbtestAddMessage(self):
        p = imapup.deferredOpenPool(self.avatar, cleanup.IClean(self.avatar).cleanPool, True)
        p = unittest.deferredResult(p)

        d = p.addMessage(StringIO(test_messages[0]), ('\\Answered',))
        m = unittest.deferredResult(d)

        a = list(cleanup.IClean(self.avatar).cleanPool.queryIndex())
        self.assertEquals([m], [e.index_uid(p.original) + 1 for e in a])

    def dbtestStore(self):
        md = mimemessage.IMIMEDelivery(self.avatar)
        msg1 = md.createMIMEReceiver(True).feedStringNow(test_messages[0])
        msg2 = md.createMIMEReceiver(True).feedStringNow(test_messages[1])

        p = imapup.deferredOpenPool(self.avatar, cleanup.IClean(self.avatar).cleanPool, True)
        p = unittest.deferredResult(p)

        msg1uid = p.uidByItem(msg1)
        msg2uid = p.uidByItem(msg2)

        # This will set \\Seen on both messages
        x = p.store(MsgSet((msg1uid, msg2uid)), ('\\Seen',), 1, True)

        # Guh.  .store() returns {} no matter what now
        # self.assertEquals(x, {msg1uid: ['\\Seen'], msg2uid: ['\\Seen']})

        self.assertEquals(msg1.getIndexKey(p.original, 'flags'), msg1.SEEN)
        self.assertEquals(msg2.getIndexKey(p.original, 'flags'), msg2.SEEN)

        # This will unset \\Seen on just msg1
        x = p.store(MsgSet((msg1uid,)), ('\\Seen',), -1, True)

        # Guh.  .store() returns {} no matter what now
        # self.assertEquals(x, {msg1uid: []})

        self.assertEquals(msg1.getIndexKey(p.original, 'flags'), 0)
        self.assertEquals(msg2.getIndexKey(p.original, 'flags'), msg2.SEEN)

        # This will set flags to \\Answered on msg2
        x = p.store(MsgSet((msg2uid,)), ('\\Answered',), 0, True)

        # Guh.  .store() returns {} no matter what now
        # self.assertEquals(x, {msg2uid: ['\\Answered']})

        self.assertEquals(msg1.getIndexKey(p.original, 'flags'), 0)
        self.assertEquals(msg2.getIndexKey(p.original, 'flags'), msg2.ANSWERED)

        # This will set \\Seen on both messages
        x = p.store(MsgSet((1, 2)), ('\\Seen',), 1, False)

        # Guh.  .store() returns {} no matter what now
        # self.assertEquals(x, {msg1uid: ['\\Seen'], msg2uid: ['\\Seen', '\\Answered']})

        self.assertEquals(msg1.getIndexKey(p.original, 'flags'), msg1.SEEN)
        self.assertEquals(msg2.getIndexKey(p.original, 'flags'), msg2.SEEN | msg2.ANSWERED)

class LotsOfMessages(unittest.TestCase, helper.ProvisionerMixin):

    # Number of messages to deliver for this test case
    mper = 25
    mloop = 40
    mcount = mper * mloop

    def setUpClass(self):
        store.MAX_LOCKS = 100

    def tearDownClass(self):
        store.MAX_LOCKS = 1000

    def dbTearDown(self):
        nc = sched.IScheduler(self.store).nextCall
        nc and nc.cancel()
        nc = getattr(searchup.ISearch(self.avatar, default=None), 'checkpointDefer', None)
        nc and nc.cancel()

    def _createAvatar(self):
        self.avatar = self.createAvatar((mimemessage.EmailPowerup, imapup.IMAPAccountUp))

    def _deliverMessages(self):
        md = mimemessage.IMIMEDelivery(self.avatar)
        mr = md.createMIMEReceiver(True)
        for i in range(self.mper):
            mr.feedStringNow(test_messages[0])

    def _listMailboxes(self):
        imap = imap4.IAccount(self.avatar)
        boxes = unittest.deferredResult(defer.maybeDeferred(imap.listMailboxes, "", ""))

    def _selectInbox(self):
        imap = imap4.IAccount(self.avatar)
        inbox = unittest.deferredResult(defer.maybeDeferred(imap.select, "Inbox"))
        self.assertEquals(inbox.getMessageCount(), self.mcount)

    def testOperations(self):
        for f in [self._createAvatar] + ([self._deliverMessages] * self.mloop):
            self.store.transact(f)
        for f in [self._selectInbox, self._listMailboxes]:
            f()



class Protocol(unittest.TestCase):

    def dbSetUp(self):
        # Fake a response from the server, make sure only the right
        # UIDs are selected.

        self.f = grabbers.IMAP4MessageGrabber().filterOld

        self.results = {
            1: {'INTERNALDATE': '01-Jan-2000 SomejunkyTimeSpec GMT'},
            2: {'INTERNALDATE': '02-Jan-2000 AnotherJunkySpec EST'},
            3: {'INTERNALDATE': '03-Jan-2000 Whenever DST'},
            4: {'INTERNALDATE': '04-Jan-2000 HuhHuhHuh PST'},
            5: {'INTERNALDATE': '30-Jan-2000 TimeTimeTime UTC'},
            6: {'INTERNALDATE': '31-Jan-2000 MoreTime CST'},
            7: {'INTERNALDATE': '01-Feb-2000 HowAboutThis MST'}}

    def testDeletionFiltering(self):
        uids = self.f(self.results, 7, datetime.date(2000, 1, 4))
        self.assertEquals(uids, [])

        uids = self.f(self.results, 7, datetime.date(2000, 1, 7))
        self.assertEquals(uids, [])

        uids = self.f(self.results, 7, datetime.date(2000, 1, 8))
        self.assertEquals(uids, [])

        uids = self.f(self.results, 7, datetime.date(2000, 1, 9))
        self.assertEquals(uids, [1])

        uids = self.f(self.results, 7, datetime.date(2000, 1, 10))
        uids.sort()
        self.assertEquals(uids, [1, 2])

        uids = self.f(self.results, 7, datetime.date(2000, 1, 11))
        uids.sort()
        self.assertEquals(uids, [1, 2, 3])

        uids = self.f(self.results, 7, datetime.date(2000, 1, 12))
        uids.sort()
        self.assertEquals(uids, [1, 2, 3, 4])

        uids = self.f(self.results, 7, datetime.date(2000, 1, 13))
        uids.sort()
        self.assertEquals(uids, [1, 2, 3, 4])

        uids = self.f(self.results, 7, datetime.date(2000, 1, 31))
        uids.sort()
        self.assertEquals(uids, [1, 2, 3, 4])

        uids = self.f(self.results, 7, datetime.date(2000, 3, 1))
        uids.sort()
        self.assertEquals(uids, [1, 2, 3, 4, 5, 6, 7])

class FlagChangeTestCase(unittest.TestCase, helper.ProvisionerMixin):
    def dbSetUp(self):
        self.avatar = self.createAvatar((mimemessage.EmailPowerup, imapup.IMAPAccountUp))

        self.msg_strings = test_messages
        self.msg_objs = []

        for m in self.msg_strings:
            mr = mimemessage.IMIMEDelivery(self.avatar)
            self.msg_objs.append(mr.createMIMEReceiver(False).feedStringNow(m))

    def dbtestSetSomeFlags(self):
        p = imapup.deferredOpenPool(self.avatar, cleanup.IClean(self.avatar).cleanPool, False)
        p = unittest.deferredResult(p)
        self.assertEquals(p.getUnseenCount(), len(self.msg_objs))

        x = len(self.msg_objs)
        for m in self.msg_objs:
            x -= 1
            m.changeFlags(None, m.SEEN, True)
            self.assertEquals(p.getUnseenCount(), x)

        for m in self.msg_objs:
            x += 1
            m.changeFlags(None, m.SEEN, False)
            self.assertEquals(p.getUnseenCount(), x)
