# -*- test-case-name: "quotient.test.test_searchup" -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import os
from os.path import join as opj

from twisted.trial import util
from twisted.internet import reactor, base, defer
from twisted.python import failure

base.DelayedCall.debug = True
defer.Deferred.debug = True

from atop.store import Store, SubStore
from atop.powerup import IPowerStation
from atop.test import unittest
from atop.sched import SystemScheduler, IScheduler

from quotient.test import emaildata, helper, testdata
from quotient.mimemessage import IMIMEDelivery, EmailPowerup
from quotient.searchup import Searchup
from quotient.scripts.usercreate import importDirectories
from quotient.sysup import ControlPanel, IControlPanel

class TestException(Exception):
    pass


class TestSearchup(unittest.TestCase, helper.ProvisionerMixin):
    storePowerups = [SystemScheduler, ControlPanel]

    def dbSetUp(self):
        self.substore = self.createAvatar([EmailPowerup])
        self.searchup = IPowerStation(self.substore).installPowerup(Searchup)
        cp = IPowerStation(self.store).installPowerup(ControlPanel)
        cp.allowAction(IControlPanel.MESSAGE_INDEXING, 'yes')

    def tearDownClass(self):
        cpd = self.searchup.checkpointDefer
        if cpd and getattr(cpd, 'active', lambda: False)():
            cpd.cancel()
        self.searchup._setCheckpointTimer = lambda _:None
        reactor.crash()

    def loadMessages(self):
        self.md = IMIMEDelivery(self.substore)
        self.substore.transact(lambda :self.searchup.indexPool(self.md.getArrivalRef().getItem()))

        recv = self.md.createMIMEReceiver(True)
        for m in testdata.ascii.iterMessages():
            recv.feedStringNow(m)

    def doIndexing(self):
        from quotient import sysup
        perm = sysup.IControlPanel.MESSAGE_INDEXING
        self.assert_(sysup.IControlPanel(self.store).actionAllowed(perm))
        self.assert_(not self.searchup.paused)
        self.store.transact(self.searchup.run)
        return self.searchup.finishedIndexing()

    def test_quickAsciiTest(self):
        self.loadMessages()
        
        if not reactor.running:
            reactor.startRunning()

        d = self.doIndexing()

        def doSearch(ignore):
            return self.searchup.search('addresses')
        d.addCallback(doSearch)

        def cleanup(result):
            cd = self.searchup.checkpointDefer
            if cd and cd.active():
                cd.cancel()
                return result
        d.addBoth(cleanup)

        result = util.deferredResult(d)
        self.assert_(result)

#    test_quickAsciiTest.skip = 'skip this for right now'


    def queryAll(self):
        from lupy.search import hits

        results, deferreds  = [], []
        
        for query in testdata.ascii.iterWords():
            d = self.searchup.search(query)
            def progressMeter(result):
                import sys
                sys.stdout.write('.'); sys.stdout.flush()
                return result
            d.addCallback(progressMeter)
            d.addCallback(lambda r: results.append(isinstance(r, hits.Hits)))
            deferreds.append(d)

        dlist = defer.DeferredList(deferreds)

        d = defer.Deferred()
        def checkResult(ignore):
            for r in results:
                if r != True:
                    d.errback(failure.DefaultException('query failed'))
            #print 'hoorj! passed!'
            d.callback("hoorj!")

        dlist.addCallback(checkResult)
        return d


    def test_AllYourTerm(self):
        self.loadMessages()

        self.assert_(reactor.running == False)

        if not reactor.running:
            reactor.startRunning()
        d = self.doIndexing()

        d.addCallback(lambda _:self.queryAll())

        def eb(reason):
            reason.trap(failure.DefaultException)
            self.fail(reason)

        d.addErrback(eb)

        result = util.deferredResult(d)
        cd = self.searchup.checkpointDefer
        if cd.active():
            cd.cancel()

    def testTransactionReversionPreservesIndexer(self):
        indexer = self.searchup.indexer
        def _1():
            self.searchup.touch()
            raise TestException("Revert this transaction")
        self.assertRaises(TestException, self.store.transact, _1)
    
        self.assertIdentical(indexer, self.searchup.indexer)
        


