# -*- test-case-name: quotient.test.test_zones -*-
# This module was brought together from a number of Python-CVS-related sources.
# Copyright attribution should mention BSD license, probably.
# Specifically:
#
#  http://cvs.sourceforge.net/viewcvs.py/python/python/nondist/sandbox/datetime/
#  http://www.python.org/doc/current/lib/datetime-tzinfo.html

from datetime import time, datetime, date
from datetime import tzinfo
from datetime import timedelta

import itertools
day = itertools.count().next

MONDAY     = day()
TUESDAY    = day()
WEDNESDAY  = day()
THURSDAY   = day()
FRIDAY     = day()
SATURDAY   = day()
SUNDAY     = day()

month = itertools.count(1).next

JANUARY    = month()
FEBRUARY   = month()
MARCH      = month()
APRIL      = month()
MAY        = month()
JUNE       = month()
JULY       = month()
AUGUST     = month()
SEPTEMBER  = month()
OCTOBER    = month()
NOVEMBER   = month()
DECEMBER   = month()

_DAY = timedelta(1)

HOUR = timedelta(hours=1)
ZERO = timedelta()

def is_leap_year(dt):
    """True if date in a leap year, False if not.
    """
    return (date(dt.year, 2, 28) + _DAY).month == 2

def days_in_month(dt):
    """Total number of days in date's month.
    """

    if dt.month == 12:
        return 31
    else:
        next = date(dt.year, dt.month+1, 1)
        return next.toordinal() - dt.replace(day=1).toordinal()

def first_weekday_on_or_after(weekday, dt):
    """First day of kind MONDAY .. SUNDAY on or after date.

    The time and tzinfo members (if any) aren't changed.
    """

    days_to_go = (weekday - dt.weekday()) % 7
    if days_to_go:
        dt += timedelta(days_to_go)
    return dt

def first_weekday_on_or_before(weekday, dt):
    """First day of kind MONDAY .. SUNDAY on or before date.

    The time and tzinfo members (if any) aren't changed.
    """

    days_to_go = (dt.weekday() - weekday) % 7
    if days_to_go:
        dt -= timedelta(days_to_go)
    return dt

def weekday_of_month(weekday, dt, index):
    if index >= 0:
        base = first_weekday_on_or_after(weekday, dt.replace(day=1))
        return base + timedelta(weeks=index)
    else:
        base = first_weekday_on_or_before(weekday,
                                          dt.replace(day=days_in_month(dt)))
        return base + timedelta(weeks=1+index)


zonemap = {}

def getzone(name):
    return zonemap[name]


class TimeZone(tzinfo):
    """Registered tzinfo subclass.
    """

    def __init__(self, zonename):
        self.zonename = zonename
        zonemap[zonename] = self

    def __str__(self):
        now = datetime.now(self)
        return "%s (%s)" % (self.zonename, self.tzname(now))

    def __reduce__(self):
        return getzone, (self.zonename,)


class USTimeZone(TimeZone):
    dstoff = timedelta(hours=1)
    zero = timedelta(0)
    # DST starts at 2am (standard time) on the first Sunday in April.
    start = datetime(1, APRIL, 1, 2)
    # and ends at 2am (DST time; 1am standard time) on the last Sunday of Oct.
    end = datetime(1, OCTOBER, 1, 1)

    def __init__(self, cityname, stdhours, stdname, dstname):
        TimeZone.__init__(self, cityname)
        self.stdoff = timedelta(hours=stdhours)
        self.stdname = stdname
        self.dstname = dstname

    def utcoffset(self, dt):
        return self.stdoff + self.dst(dt)

    def tzname(self, dt):
        if self.dst(dt):
            return self.dstname
        else:
            return self.stdname

    def dst(self, dt):
        if dt is None or dt.tzinfo is None:
            # An exception instead may be sensible here, in one or more of
            # the cases.
            return self.zero

        assert dt.tzinfo is self

        # Find first Sunday in April.
        start = weekday_of_month(SUNDAY, self.start.replace(year=dt.year), 0)
        assert start.weekday() == 6 and start.month == 4 and start.day <= 7

        # Find last Sunday in October.
        end = weekday_of_month(SUNDAY, self.end.replace(year=dt.year), -1)
        assert end.weekday() == 6 and end.month == 10 and end.day >= 25

        # Can't compare naive to aware objects, so strip the timezone from
        # dt first.
        if start <= dt.replace(tzinfo=None) < end:
            return self.dstoff
        else:
            return self.zero

# The switches are at 1AM UTC on the last Sundays in March and
# October.
_dston = datetime(1, MARCH, 1, 1)
_dstoff = datetime(1, OCTOBER, 1, 1)

class Fixed(TimeZone):

    def __init__(self, offset, name):
        TimeZone.__init__(self, name)
        self.offset = offset
        self.name = name

    def tzname(self, dt):
        return self.name

    def utcoffset(self, dt):
        return self.offset

    def dst(self, dt):
        return ZERO

class Europe(TimeZone):
    def __init__(self, cityname, offset, stdname, dstname):
        TimeZone.__init__(self, cityname)
        self.offset = offset
        self.stdname = stdname
        self.dstname = dstname

    def tzname(self, dt):
        if self.dst(dt):
            return self.dstname
        else:
            return self.stdname

    def utcoffset(self, dt):
        return self.offset + self.dst(dt)

    def dst(self, dt):
        if dt is None or dt.tzinfo is None:
            return ZERO
        assert dt.tzinfo is self
        dston = _dston.replace(year=dt.year)
        dstoff = _dstoff.replace(year=dt.year)
        dston = weekday_of_month(SUNDAY, dston,  -1)
        dstoff = weekday_of_month(SUNDAY, dstoff, -1)
        # Convert dt to a naive UTC too (we have to strip the tzinfo member
        # in order to compare to the naive dston and dstoff).
        dt -= self.offset
        if dston <= dt.replace(tzinfo=None) < dstoff:
            return HOUR
        else:
            return ZERO

import time as _time

STDOFFSET = timedelta(seconds = -_time.timezone)
if _time.daylight:
    DSTOFFSET = timedelta(seconds = -_time.altzone)
else:
    DSTOFFSET = STDOFFSET

DSTDIFF = DSTOFFSET - STDOFFSET

class LocalTimezone(TimeZone):

    def __init__(self):
        TimeZone.__init__(self, "local")

    def utcoffset(self, dt):
        if self._isdst(dt):
            return DSTOFFSET
        else:
            return STDOFFSET

    def dst(self, dt):
        if self._isdst(dt):
            return DSTDIFF
        else:
            return ZERO

    def tzname(self, dt):
        return _time.tzname[self._isdst(dt)]

    def _isdst(self, dt):
        tt = (dt.year, dt.month, dt.day,
              dt.hour, dt.minute, dt.second,
              dt.weekday(), 0, -1)
        stamp = _time.mktime(tt)
        tt = _time.localtime(stamp)
        return tt.tm_isdst > 0

UTC              = Fixed(ZERO, "UTC")
Local            = LocalTimezone()
Eastern          = USTimeZone("Eastern", -5, "EST", "EDT")
Central          = USTimeZone("Central", -6, "CST", "CDT")
Mountain         = USTimeZone("Mountain", -7, "MST", "MDT")
Pacific          = USTimeZone("Pacific", -8, "PST", "PDT")
London           = Europe("London", ZERO, "WET", "WDT")
Amsterdam        = Europe("Amsterdam", HOUR, "MET", "MDT")

_z = zonemap.items()
_z.sort()
zones = [zrec[1] for zrec in _z]

__all__ = ['London', 'Amsterdam', 'UTC', 'Eastern', 'Central', 'Mountain',
           'Pacific', 'zones']

